//
//	plasticfs - the plastic file system
//	Copyright (C) 2003, 2004 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the execve class
//

#include <plasticfs.h>
#include <cerrno>
#include <cstdarg>
#include <cstring>
#include <unistd.h> // for environ


using namespace libplasticfs;


extern "C" int
execve(const char *path, char *const argv[], char *const envp[])
{
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }
    return plasticfs::get_master()->execve(path, argv, envp);
}


extern "C" int
__execve(const char *path, char *const argv[], char *const envp[])
{
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }
    return plasticfs::get_master()->execve(path, argv, envp);
}


/**
  * The get_argv function is used to get this argv[] value for a
  * function in the exec familty whichg accepts a NULL terminated list
  * of arguments.
  *
  * \returns the array of arguments in a new[]ed array.  The caller is
  * reposnsible for delete[]ing it.  If NULL is returned it means that
  * an ENOMEM error occurred.
  */
static char *const *
get_argv(const char *arg0, va_list ap)
{
    size_t argv_max = 1024;
    const char **argv = new const char *[argv_max];
    size_t argc = 0;
    argv[argc++] = arg0;
    for (;;)
    {
	//
	// Fetch the next function argument.  We have to fetch all of
	// the arguments even if the ::new failed, so that va_end()
	// is called from the correct position in the argument list.
	//
	const char *arg = va_arg(ap, const char *);

	//
	// Make sure we have enough room in the argv array.
	//
	if (argv && argc >= argv_max)
	{
	    size_t new_argv_max = argv_max * 2;
	    const char **new_argv = new const char *[new_argv_max];
	    if (!new_argv)
	    {
		delete [] argv;
		argv = 0;
	    }
	    else
	    {
		memcpy(new_argv, argv, sizeof(argv[0]) * argv_max);
		delete [] argv;
		argv = new_argv;
		argv_max = new_argv_max;
	    }
	}

	//
	// Add the argument to the array (including the NULL).
	//
	if (argv)
	    argv[argc++] = arg;

	//
	// If the argument was NULL we are done.
	//
	if (!arg)
	    break;
    }
    return (char *const *)argv;
}


int
execl(const char *path, const char *arg0, ...)
{
    va_list ap;
    va_start(ap, arg0);
    char *const *argv = get_argv(arg0, ap);
    va_end(ap);
    if (!argv)
	return -1;
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }

    //
    // Now pass it to the file system for execution.
    //
    int result = plasticfs::get_master()->execve(path, argv, environ);
    delete [] argv;
    return result;
}


int
execle(const char *path, const char *arg0, ...)
{
    va_list ap;
    va_start(ap, arg0);
    char *const *argv = get_argv(arg0, ap);
    va_end(ap);

    //
    // After all the command line arguments comes the environment pointer.
    //
    char *const *envp = va_arg(ap, char *const *);
    va_end(ap);

    if (!argv)
	return -1;
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }

    //
    // Now pass it to the file system for execution.
    //
    int result = plasticfs::get_master()->execve(path, argv, envp);
    delete [] argv;
    return result;
}


int
execlp(const char *path, const char *arg0, ...)
{
    va_list ap;
    va_start(ap, arg0);
    char *const *argv = get_argv(arg0, ap);
    va_end(ap);
    if (!argv)
	return -1;
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }

    //
    // Now pass it to the file system for execution.
    //
    int result = plasticfs::get_master()->execvp(path, argv);
    delete [] argv;
    return result;
}


int
execv(const char *path, char *const argv[])
{
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }
    return plasticfs::get_master()->execve(path, argv, environ);
}


int
execvp(const char *path, char *const argv[])
{
    if (!path)
    {
	errno = EINVAL;
	return -1;
    }
    return plasticfs::get_master()->execvp(path, argv);
}


extern "C" int
fexecve(int fd, char *const argv[], char *const envp[])
{
    return plasticfs::get_master()->fexecve(fd, argv, envp);
}


extern "C" int
__fexecve(int fd, char *const argv[], char *const envp[])
{
    return plasticfs::get_master()->fexecve(fd, argv, envp);
}
