//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/file_tracker.cc
//

#ifndef LIB_FILE_TRACKER_H
#define LIB_FILE_TRACKER_H

#pragma interface "libplasticfs::file_tracker"

#include <plasticfs.h>
#include <symtab.h>
#include <itab.h>
#include <sys/types.h>

namespace libplasticfs {

/**
  * The file_tracker class is used to represent
  */
class file_tracker
{
public:
    /**
      * The destructor.
      */
    virtual ~file_tracker();

    /**
      * The default constructor.
      */
    file_tracker();

    struct active_t
    {
	enum { MAGIC=0xDDEEDDDD };
	unsigned magic;
	int reference_count;
	int sub_fd;
	size_t sub_idx;
	rcstring relpath;
	symtab direntries;
	int flags;
	off_t filepos;
	off_t sub_filepos;
	plasticfs *fs;

	active_t() : magic(MAGIC), reference_count(1), sub_fd(-1), sub_idx(0),
	    flags(0), filepos(0), sub_filepos(0) {}
	~active_t() { magic = 0; }
	void close();
    };

    /**
      * The find method is used to locate the data attached to a specific
      * file descriptor.
      */
    active_t *
    find(int fd)
    {
	active_t *p = (active_t *)active.query(fd);
	if (p && p->magic != active_t::MAGIC)
	    p = 0;
	return p;
    }

    /**
      * The open method is used to create new data to associate with a
      * file descriptor.
      */
    active_t *open(int fd, int flags, const rcstring &relpath, plasticfs *fs);

    /**
      * The remove method is used to remove a reference to the associated
      * data structure.  When the last reference is removed, the data
      * is deleted.
      */
    void close(int fd);

    /**
      * The dup method is used to increates the number of references to
      * the data assiciated with old_fd to include new_fd.
      */
    void dup(int old_fd, int new_fd);

private:
    /**
      * The active table is indexed by the file descriptor.  The O(1)
      * lookup means that there isn't a significant performance penalty
      * for doing this.
      */
    itab active;

    /**
      * The copy constructor.  Do not use.
      */
    file_tracker(const file_tracker &);

    /**
      * The assignment operator.  Do not use.
      */
    file_tracker &operator=(const file_tracker &);
};

} // namespace libplasticfs

#endif // LIB_FILE_TRACKER_H
