//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/input.cc
//

#ifndef LIB_INPUT_H
#define LIB_INPUT_H

#pragma interface "libplasticfs::input"

#include <extra_symbol_definitions.h>
#include <sys/types.h>

#include <rcstring.h>

namespace libplasticfs {

/**
  * The input class is used to represent an arbitrary input stream.
  * This class hierarchy present because we can't use stdio or iostreams.
  */
class input
{
public:
    /**
      * The destructor.
      */
    virtual ~input();

    enum { EOF = -1 };

    /**
      * The getc method may be used to get a single character from an
      * input stream.  Returns EOF at end of file.
      */
    int
    getc()
    {
	if (bufpos >= bufmax)
        {
	    ssize_t n = underflow(buffer, sizeof(buffer));
	    if (n <= 0)
		return EOF;
	    bufmax = buffer + n;
	    bufpos = buffer;
	}
	return *bufpos++;
    }
	    
    /**
      * The readline method may be used to read a whole line from an
      * input stream.  Returns false at end of file.
      */
    bool readline(rcstring &);

    /**
      * The readlink_c method may be used to read a whole line from an
      * input stream, using C escapes.  Returns false at end of file.
      */
    bool readline_c(rcstring &);

protected:
    /**
      * The default constructor.
      * Only derived classes may use this.
      */
    input();

    /**
      * The underflow method is used to fill the input buffer when more
      * data is required.  Returns 0 for end of file, and -1 on error.
      */
    virtual ssize_t underflow(void *, size_t) = 0;

private:
    /**
      * The buffer instance variable is used to remember the remaining
      * buffered input characters.
      */
    unsigned char buffer[4096];

    /**
      * The buffer instance variable is used to remember the position
      * within the input buffer beyond which there is no further valid
      * data.
      */
    unsigned char *bufmax;

    /**
      * The buffer instance variable is used to remember the position
      * within the input buffer containing the next input character,
      * unless it is beyond bufmax.
      */
    unsigned char *bufpos;

    /**
      * The copy constructor.  Do not use.
      */
    input(const input &);

    /**
      * The assignment operator.  Do not use.
      */
    input &operator=(const input &);
};

} // namespace libplasticfs

#endif // LIB_INPUT_H
