//
//	plasticfs - the plastic file system
//	Copyright (C) 2003, 2004 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the itab class
//

#pragma implementation "libplasticfs::itab"

#include <cstring>
#include <itab.h>

using namespace libplasticfs;


itab::~itab()
{
    for (hash_ty j = 0; j < hash_modulus; ++j)
    {
	row_ty **rpp = &hash_table[j];
	while (*rpp)
	{
	    row_ty *rp = *rpp;
	    *rpp = rp->overflow;
	    if (reap)
		reap(rp->data);
	    delete rp;
	}
    }
    delete [] hash_table;
}


itab::itab()
{
    hash_ty size = 5;
    reap = 0;
    hash_modulus = 1 << 2; // MUST be a power of 2
    while (hash_modulus < size)
	hash_modulus <<= 1;
    hash_cutover = hash_modulus;
    hash_split = hash_modulus - hash_cutover;
    hash_cutover_mask = hash_cutover - 1;
    hash_cutover_split_mask = (hash_cutover * 2) - 1;
    hash_load = 0;
    hash_table = new row_ty * [hash_modulus];
    for (hash_ty j = 0; j < hash_modulus; ++j)
	hash_table[j] = 0;
}


void
itab::split()
{
    //
    // get the list to be split across buckets
    //
    row_ty *p = hash_table[hash_split];
    hash_table[hash_split] = 0;

    //
    // increase the modulus by one
    //
    hash_ty new_hash_modulus = hash_modulus + 1;
    row_ty **new_hash_table = new row_ty * [new_hash_modulus];
    memcpy(new_hash_table, hash_table, hash_modulus * sizeof(row_ty *));
    delete [] hash_table;
    hash_table = new_hash_table;
    hash_table[hash_modulus] = 0;
    hash_modulus = new_hash_modulus;

    hash_split = hash_modulus - hash_cutover;
    if (hash_split >= hash_cutover)
    {
	hash_cutover = hash_modulus;
	hash_split = 0;
	hash_cutover_mask = hash_cutover - 1;
	hash_cutover_split_mask = (hash_cutover * 2) - 1;
    }

    //
    // now redistribute the list elements
    //
    // It is important to preserve the order of the links because
    // they can be push-down stacks, and to simply add them to the
    // head of the list will reverse the order of the stack!
    //
    while (p)
    {
	row_ty *p2 = p;
	p = p2->overflow;
	p2->overflow = 0;

	hash_ty index = p2->key & hash_cutover_mask;
	if (index < hash_split)
	    index = (p2->key & hash_cutover_split_mask);
	row_ty **ipp;
	for (ipp = &hash_table[index]; *ipp; ipp = &(*ipp)->overflow)
	    ;
	*ipp = p2;
    }
}


void *
itab::query(key_t key)
    const
{
    hash_ty index = key & hash_cutover_mask;
    if (index < hash_split)
	index = key & hash_cutover_split_mask;
    for (row_ty *p = hash_table[index]; p; p = p->overflow)
    {
	if (key == p->key)
    	    return p->data;
    }
    return 0;
}


void
itab::assign(key_t key, void *data)
{
    hash_ty index = key & hash_cutover_mask;
    if (index < hash_split)
	index = key & hash_cutover_split_mask;
    for (row_ty *p = hash_table[index]; p; p = p->overflow)
    {
	if (key == p->key)
	{
    	    if (reap)
       		reap(p->data);
    	    p->data = data;
    	    return;
	}
    }

    row_ty *np = new row_ty;
    np->key = key;
    np->overflow = hash_table[index];
    np->data = data;
    hash_table[index] = np;

    hash_load++;
    while (hash_load * 10 >= hash_modulus * 8)
	split();
}


void
itab::remove(key_t key)
{
    hash_ty index = key & hash_cutover_mask;
    if (index < hash_split)
	index = key & hash_cutover_split_mask;

    row_ty **pp = &hash_table[index];
    for (;;)
    {
	row_ty *p = *pp;
	if (!p)
	    break;
	if (key == p->key)
	{
	    if (reap)
	       	reap(p->data);
	    *pp = p->overflow;
	    delete p;
	    hash_load--;
	    break;
	}
	pp = &p->overflow;
    }
}


#if 0

void
itab::dump(const char *caption)
    const
{
    error("symbol table %s = {", caption);
    for (hash_ty j = 0; j < hash_modulus; ++j)
	for (row_ty *p = hash_table[j]; p; p = p->overflow)
	    error("key = \"%s\", data = %08lX", p->key.C_str(), (long)p->data);
    error("}");
}

#endif


void
itab::walk(walk_t &arg)
{
    for (hash_ty j = 0; j < hash_modulus; ++j)
       	for (row_ty *p = hash_table[j]; p; p = p->overflow)
	    arg.action(p->key, p->data);
}


void
itab::set_reaper(void (*fp)(void *))
{
    reap = fp;
}
