//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the output class
//

#pragma implementation "libplasticfs::output"

#include <output.h>
#include <cstring>
#include <cctype>

using namespace libplasticfs;


output::~output()
{
    // it's too late to flush() here
    // the derived class' destructor has already been run
}


output::output() :
    bufpos(buffer)
{
}


bool
output::write(const void *p, size_t nbytes)
{
    while (nbytes)
    {
	// it it all fits within the buffer, stash it and return
	if (bufpos + nbytes < buffer + sizeof(buffer))
	{
	    memcpy(bufpos, p, nbytes);
	    bufpos += nbytes;
	    break;
	}

	// figure how much will fit into the buffer
	size_t partial = buffer + sizeof(buffer) - bufpos;
	if (partial == sizeof(buffer))
	{
	    // if completely aligned, do as much as possible 
	    partial = nbytes & ~(sizeof(buffer) - 1);
	    if (!overflow(p, partial))
		return false;
	}
	else
	{
	    // copy some into buffer, and then write whole buffer
	    memcpy(bufpos, p, partial);
	    if (!overflow(buffer, sizeof(buffer)))
		return false;
	    bufpos = buffer;
	}

	// advance past the amount written so far
	nbytes -= partial;
	p = (void *)((char *)p + partial);
    }
    return true;
}


bool
output::puts(const char *s)
{
    return write(s, strlen(s));
}


bool
output::puts(const rcstring &s)
{
    return write(s.c_str(), s.length());
}


bool
output::write_c(const char *s, size_t nbytes)
{
    while (nbytes--)
    {
	unsigned char c = *s++;
	switch (c)
	{
	case '\\':
	case '"':
	case '\'':
	    escaped:
	    if (!putc('\\'))
		return false;
	    break;

	case '\a': c = 'a'; goto escaped;
	case '\b': c = 'b'; goto escaped;
	case '\f': c = 'f'; goto escaped;
	case '\n': c = 'n'; goto escaped;
	case '\r': c = 'r'; goto escaped;
	case '\t': c = 't'; goto escaped;
	case '\v': c = 'v'; goto escaped;

	default:
	    if (!isprint(c))
	    {
		if (!printf("\\%3.3o", c))
		    return false;
		continue;
	    }
	    break;
	}
	if (!putc(c))
	    return false;
    }
    return true;
}


bool
output::puts_c(const char *s)
{
    return write_c(s, strlen(s));
}


bool
output::puts_c(const rcstring &s)
{
    return write_c(s.c_str(), s.length());
}


#include <stdio.h>
#include <stdarg.h>


bool
output::printf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    char buffer[2000];
    vsnprintf(buffer, sizeof(buffer), fmt, ap);
    va_end(ap);
    return puts(buffer);
}
