//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/output.cc
//

#ifndef LIB_OUTPUT_H
#define LIB_OUTPUT_H

#pragma interface "libplasticfs::output"

#include <extra_symbol_definitions.h>
#include <sys/types.h>

#include <rcstring.h>

namespace libplasticfs {

/**
  * The output class is used to represent an arbitrary output stream.
  * This exists because we can't use stdio or iostreams.
  */
class output
{
public:
    /**
      * The destructor.
      */
    virtual ~output();

    /**
      * The putc method is used to emit a single character to the
      * output stream.
      */
    bool
    putc(int c)
    {
	if (bufpos >= buffer + sizeof(buffer))
	{
	    if (!overflow(buffer, bufpos - buffer))
		return false;
	    bufpos = buffer;
	}
	*bufpos++ = c;
	return true;
    }

    /**
      * The flush method is used to flush the contents of the buffer,
      * if there is anything in it.
      */
    bool
    flush()
    {
	if (bufpos > buffer)
	{
	    if (!overflow(buffer, bufpos - buffer))
		return false;
	    bufpos = buffer;
	}
	return true;
    }

    /**
      * The puts method is used to write a C string (not including the
      * terminating NUL) to the output stream.
      */
    bool puts(const char *);

    /**
      * The puts method is used to write a C string (not including the
      * terminating NUL) to the output stream.
      */
    bool puts(const rcstring &);

    /**
      * The puts_c method is used to write a C string (not including the
      * terminating NUL) to the output stream.	It will use C escapes
      * for all unprintable characters.
      */
    bool puts_c(const char *);

    /**
      * The puts_c method is used to write a C string (not including the
      * terminating NUL) to the output stream.	It will use C escapes
      * for all unprintable characters.
      */
    bool puts_c(const rcstring &);

    /**
      * The write method is used to write arbitrary data to the output
      * stream.
      */
    bool write(const void *, size_t);

    /**
      * The write method is used to write arbitrary data to the output
      * stream. It will use C escapes for all unprintable characters.
      */
    bool write_c(const char *, size_t);

    /**
      * The printf method is used to write formatted text to the output
      * stream.  See printf(3) for format specifications.
      */
    bool printf(const char *fmt, ...)
					  __attribute__((format(printf, 2, 3)));

protected:
    /**
      * The default constructor.
      * May only be used by derived classes.
      */
    output();

    /**
      * The overflow method is used to write out a full buffer to the
      * output stream.  It does not manipulate the buffer itself.
      */
    virtual bool overflow(const void *, size_t) = 0;

private:
    /**
      * The buffer instance variable is used to remember the characters
      * buffered for writing.
      *
      * The length of the buffer MUST be a power of two.
      */
    unsigned char buffer[1 << 12];

    /**
      * The bufpos instance variable is used to remember the next point
      * within the buffer that a character may be stored.
      */
    unsigned char *bufpos;

    /**
      * The copy constructor.  Do not use.
      */
    output(const output &);

    /**
      * The assignment operator.  Do not use.
      */
    output &operator=(const output &);
};

} // namespace libplasticfs

#endif // LIB_OUTPUT_H
