//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/plasticfs.cc
//

#ifndef LIB_PLASTICFS_H
#define LIB_PLASTICFS_H

#pragma interface "libplasticfs::plasticfs"

struct stat; // forward
struct stat64; // forward
struct utimbuf; // forward
struct statfs; // forward
struct statfs64; // forward

#include <extra_symbol_definitions.h>

// DO NOT #include <unistd.h> here
#include <sys/types.h>
#include <dirent.h>
#include <config.h>

namespace libplasticfs {

/**
  * The plasticfs class is used to represent a pure virtual interface
  * to a plastic file system.  Each of the public methods are a file
  * operation such as open, close, read, etc.
  */
class plasticfs
{
public:
    /**
      * The get_master method is used to obtain the pointer to the master
      * filter pointer.  If necessary the PLASTICFS environment variable
      * will be parsed to determine the appropriate value.
      */
    static plasticfs *get_master(void);

    /**
      * The destructor.
      */
    virtual ~plasticfs();

    /**
      * The open method is used to open files.  It is used to intercept
      * the usual open(2) system call.
      */
    virtual int open(const char *path, int flags, int permissions) = 0;

    /**
      * The close method is used to close files.  It is used to intercept
      * the usual close(2) system call.
      */
    virtual int close(int filedesc) = 0;

    /**
      * The read method attempts to read up to nbytes bytes from file
      * descriptor filedesc into the buffer starting at data.  The read()
      * method is used to intercept the usual read(2) system call.
      */
    virtual ssize_t read(int filedesc, void *data, size_t nbytes) = 0;

    /**
      * The write method writes up to nbytes bytes to the file referenced
      * by the file descriptor filedesc from the buffer starting at data.
      * POSIX requires that a read() which can be proved to occur after
      * a write() has returned returns the new	data.  Note that not
      * all file systems are POSIX conforming.  The write method is used
      * to intercept the usual write(2) system call.
      */
    virtual ssize_t write(int filedesc, const void *data, size_t nbytes) = 0;

    /**
      * The lseek method repositions the offset of the file descriptor
      * fildes to the argument offset according to the directive whence.
      * The lseek method is used to intercept the usual lseek(2)
      * system call.
      */
    virtual off_t lseek(int fildes, off_t offset, int whence) = 0;

    /**
      * The lseek64 method repositions the offset of the file descriptor
      * fildes to the argument offset according to the directive whence.
      * The lseek64 method is used to intercept the usual lseek64(2)
      * system call.
      */
    virtual off64_t lseek64(int fildes, off64_t offset, int whence) = 0;

    /**
      * The readlink method places the contents of the symbolic link path
      * in the buffer data, which has size nbytes.  The readlink method
      * does not append a NUL character to data.  It will truncate the
      * contents (to a length of nbytes characters), in case the buffer
      * is too small to hold all of the contents.  The readlink method
      * is used to intercept the usual readlink(2) call.
      */
    virtual int readlink(const char *path, char *data, size_t nbytes) = 0;

    /**
      * The stat method is used to attach the appropriate version and
      * then call xstat.  It is used to intercept the usual stat(2) call.
      * Note that it is not usually necessary to overload this method,
      * as it immediately calls the virtual xstat method.
      */
    virtual int stat(const char *path, struct stat *statbuf);

    /**
      * The xstat method is used to obtain information about the
      * specified file.  It is used to intercept the usual inlining of
      * the stat(2) call.
      */
    virtual int xstat(int version, const char *path, struct stat *statbuf) = 0;

    /**
      * The lstat method is used to attach the appropriate version and
      * then call lxstat.  It is used to intercept the usual lstat(2) call.
      * Note that it is not usually necessary to overload this method,
      * as it immediately calls the virtual lxstat method.
      */
    virtual int lstat(const char *path, struct stat *statbuf);

    /**
      * The lxstat method is used to obtain information about the
      * specified file.  It is used to intercept the usual inlining of
      * the lstat(2) call.
      */
    virtual int lxstat(int version, const char *path, struct stat *statbuf) = 0;

    /**
      * The fstat method is used to attach the appropriate version and
      * then call fxstat.  It is used to intercept the usual fstat(2) call.
      * Note that it is not usually necessary to overload this method,
      * as it immediately calls the virtual fxstat method.
      */
    virtual int fstat(int filedesc, struct stat *statbuf);

    /**
      * The fxstat method is used to obtain information about the
      * specified file.  It is used to intercept the usual inlining of
      * the fstat(2) call.
      */
    virtual int fxstat(int version, int filedesc, struct stat *statbuf) = 0;

    /**
      * The stat64 method is used to attach the appropriate version and
      * then call xstat64.  It is used to intercept the usual stat64(2)
      * call.  Note that it is not usually necessary to overload this
      * method, as it immediately calls the virtual xstat64 method.
      */
    virtual int stat64(const char *path, struct stat64 *statbuf);

    /**
      * The xstat64 method is used to obtain information about the
      * specified file.  It is used to intercept the usual inlining of
      * the stat64(2) call.
      */
    virtual int xstat64(int version, const char *path, struct stat64 *buf) = 0;

    /**
      * The lstat64 method is used to attach the appropriate version and
      * then call lxstat64.  It is used to intercept the usual lstat64(2)
      * call.  Note that it is not usually necessary to overload this
      * method, as it immediately calls the virtual lxstat64 method.
      */
    virtual int lstat64(const char *path, struct stat64 *buf);

    /**
      * The lxstat64 method is used to obtain information about the
      * specified file.  It is used to intercept the usual inlining of
      * the lstat64(2) call.
      */
    virtual int lxstat64(int version, const char *path, struct stat64 *buf) = 0;

    /**
      * The fstat64 method is used to attach the appropriate version and
      * then call fxstat64.  It is used to intercept the usual fstat64(2)
      * call.  Note that it is not usually necessary to overload this
      * method, as it immediately calls the virtual fxstat64 method.
      */
    virtual int fstat64(int filedesc, struct stat64 *buf);

    /**
      * The fxstat64 method is used to obtain information about the
      * specified file.  It is used to intercept the usual inlining of
      * the fstat64(2) call.
      */
    virtual int fxstat64(int version, int filedesc, struct stat64 *buf) = 0;

#ifndef HAVE_GETDENTS
    //
    // It would have been infinitely preferable to intercept getdents
    // and getdents64, and have glibc's opendir family call us through it.
    // Alas, these symbols are not exported to libc.so.6, and so glibc's
    // readdir does not call out to us, so we are forced to intercept
    // far too much.  However, the similar functions getdirentries and
    // getdirentries64 are available.
    //
#endif

    /**
      * The opendir() method opens a directory stream corresponding to
      * the directory name, and returns a pointer to the directory stream.
      * The stream is positioned at the first entry in the directory.
      */
    virtual DIR *opendir(const char *path);

    /**
      * The closedir() method closes the directory stream associated
      * with dir.  The directory stream descriptor dir is not available
      * after this call.
      */
    virtual int closedir(DIR *);

    /**
      * The readdir() method returns a pointer to a dirent structure
      * representing the next directory entry in the directory stream
      * pointed to by dir.  It returns NULL on reaching the end-of-file
      * or if an error occurred.
      */
    virtual dirent *readdir(DIR *);

    /**
      * The readdir64() method returns a pointer to a dirent64 structure
      * representing the next directory entry in the directory stream
      * pointed to by dir.  It returns NULL on reaching the end-of-file
      * or if an error occurred.
      */
    virtual dirent64 *readdir64(DIR *);

    /**
      * The rewinddir() method resets the position of the directory
      * stream dir to the beginning of the directory.
      */
    virtual void rewinddir(DIR *);

    /**
      * The seekdir() method sets the location in the directory stream
      * from which the next readdir() call will start.	seekdir() should
      * be used with an offset returned by telldir().
      */
    virtual void seekdir(DIR *, off_t);

    /**
      * The telldir() method returns the current location associated
      * with the directory stream dir.
      */
    virtual off_t telldir(DIR *);

    /**
      * The dirfd() method returns the file descriptor associated
      * with the directory stream dir.
      */
    virtual int dirfd(DIR *);

    /**
      * The getdirentries method is used to read directory entries from
      * the directory specified by fd into buf.  At most nbytes are read.
      * Reading starts at offset *basep, and *basep is updated with the
      * new position after reading.
      */
    virtual ssize_t getdirentries(int fd, char *buf, size_t nbytes,
	off_t *basep) = 0;

    /**
      * The getdents method reads several dirent structures from the
      * directory pointed at by fd into the memory area pointed to
      * by dirp.  The parameter count is the size of the memory area.
      *
      * This is not actually exported by glibc.  We fake it by calling
      * the getdirentries method.
      */
    virtual int getdents(int fd, dirent *buf, unsigned nbytes) = 0;

    /**
      * The getdirentries64 method is used to read directory entries from
      * the directory specified by fd into buf.  At most nbytes are read.
      * Reading starts at offset *basep, and *basep is updated with the
      * new position after reading.
      */
    virtual ssize_t getdirentries64(int fd, char *buf, size_t nbytes,
	off64_t *basep) = 0;

    /**
      * The getdents64 method reads several dirent structures from the
      * directory pointed at by fd into the memory area pointed to
      * by dirp.  The parameter count is the size of the memory area.
      *
      * This is not actually exported by glibc.  We fake it by calling
      * the getdirentries64 method.
      */
    virtual int getdents64(int fd, dirent64 *buf, unsigned nbytes) = 0;

    /**
      * Get the pathname of the current working directory, and put it in
      * SIZE bytes of BUF.  Returns NULL if the directory couldn't be
      * determined or SIZE was too small.  If successful, returns BUF.
      * In GNU, if BUF is NULL, an array is allocated with `malloc';
      * the array is SIZE bytes long, unless SIZE == 0, in which case
      * it is as big as necessary.
      */
    virtual char *getcwd(char *buf, long size);

    /**
      * The realpath mathod expands all symbolic links and resolves
      * references to '/./', '/../' and extra '/' characters in the
      * null terminated string named by path and stores the canon
      * icalized absolute pathname in the buffer of size PATH_MAX named
      * by resolved_path.  The resulting path will have no symbolic link,
      * '/./' or '/../' components.
      */
    virtual char *realpath(const char *, char *resolved_path);

    /**
      * The realpath mathod expands all symbolic links and resolves
      * references to '/./', '/../' and extra '/' characters in the
      * null terminated string named by path and stores the canon
      * icalized absolute pathname in the buffer of size nbytes named
      * by resolved_path.  The resulting path will have no symbolic link,
      * '/./' or '/../' components.
      */
    virtual bool realpath(const char *, char *resolved_path, size_t nbytes,
	bool file_must_exist);

    /**
      * The chmod method is used to change the permission modes of the
      * given file.
      */
    virtual int chmod(const char *path, mode_t mode) = 0;

    /**
      * The fchmod method is used to change the permission modes of the
      * given file.
      */
    virtual int fchmod(int filedesc, mode_t mode) = 0;

    /**
      * The mkdir method is used to create directories.
      */
    virtual int mkdir(const char *path, mode_t mode) = 0;

    /**
      * The mkfifo method is used to create FIFO special files.
      */
    virtual int mkfifo(const char *path, mode_t mode) = 0;

    /**
      * The link method is used to create a hard link between file
      * entries.
      */
    virtual int link(const char *src, const char *dest) = 0;

    /**
      * The unlink method is used to remove file entries.
      */
    virtual int unlink(const char *path) = 0;

    /**
      * The rename method is used to change the name of file entries.
      */
    virtual int rename(const char *oldpath, const char *newpath) = 0;

    /**
      * The pathconf method is used to get a value for configuration
      * option 'name' for the file name 'path'.  See pathconf(3) for more
      * information.
      */
    virtual long pathconf(const char *path, int name) = 0;

    /**
      * The fpathconf method is used to get a value for configuration
      * option 'name' for the file descriptor 'fd'.  See fpathconf(3)
      * for more information.
      */
    virtual long fpathconf(int fd, int name) = 0;

    /**
      * The mknod method is used to create a filesystem node (file,
      * device special file or named pipe) named pathname, specified
      * by mode and dev.
      */
    virtual int mknod(const char *path, mode_t mode, dev_t dev);

    /**
      * The xmknod method is used to create a filesystem node (file,
      * device special file or named pipe) named pathname, specified
      * by mode and dev.
      */
    virtual int xmknod(int vers, const char *path, mode_t mode, dev_t dev) = 0;

    /**
      * The fcntl method performs one of various miscellaneous operations
      * on the file descriptor.  The operation in question is determined
      * by cmd.
      */
    virtual int fcntl(int filedesc, int cmd, void *arg) = 0;

    /**
      * The fcntl64 method performs one of various miscellaneous
      * operations on the file descriptor.  The operation in question
      * is determined by cmd.
      */
    virtual int fcntl64(int filedesc, int cmd, void *arg) = 0;

    /**
      * The dup method is used to create a copy of the file descriptor
      * oldfd.  After successful return of dup, the old and new
      * descriptors may be used interchangeably. They share locks,
      * file position pointers and flags; for example, if the file
      * position is modified by using lseek on one of the descriptors,
      * the position is also changed for the other.
      */
    virtual int dup(int oldfd) = 0;

    /**
      * The dup2 method is used to create a copy of the file descriptor
      * oldfd.  After successful return of dup2, the old and new
      * descriptors may be used interchangeably. They share locks,
      * file position pointers and flags; for example, if the file
      * position is modified by using lseek on one of the descriptors,
      * the position is also changed for the other.
      */
    virtual int dup2(int oldfd, int newfd) = 0;

    /**
      * The symlink method is sued to create a symbolic link named
      * `newpath' which contains the string `oldpath'.
      */
    virtual int symlink(const char *oldpath, const char *newpath) = 0;

    /**
      * The chdir method is used to change the current directory to that
      * specified in path.
      */
    virtual int chdir(const char *path) = 0;

    /**
      * The fchdir method ise used to changes  the current directory to
      * that specified by the open file descriptor.
      */
    virtual int fchdir(int fd) = 0;

    /**
      * The rmdir method is used to delete a directory, which must
      * be empty.
      */
    virtual int rmdir(const char *path) = 0;

    /**
      * The utime method is used to change the access and modification
      * times of the inode specified by 'path' to the actime and modtime
      * fields of 'buf' respectively.  If buf is NULL, then the access and
      * modification times of the file are set to the current time.
      */
    virtual int utime(const char *path, utimbuf *buf) = 0;

    /**
      * The lutime method is used to change the access and modification
      * times of the inode specified by 'path' to the actime and modtime
      * fields of 'buf' respectively.  In the case of a symbolic link,
      * where the link itself is utime()ed, not the file that it refers
      * to.  If buf is NULL, then the access and modification times of
      * the file are set to the current time.
      */
    virtual int lutime(const char *path, utimbuf *buf) = 0;

    /**
      * The futime method is used to change the access and modification
      * times of the inode specified by open file descriptor 'fd' to the
      * actime and modtime fields of 'buf' respectively.  If buf is NULL,
      * then the access and modification times of the file are set to
      * the current time.
      */
    virtual int futime(int fd, utimbuf *buf) = 0;

    /**
      * The execve method is used to execute the programs.
      *
      * \param filename
      *     a binary executable, or a script starting with a line of
      *     the form "#!  interpreter  [arg]".
      * \param argv
      *     an  array  of argument strings passed to the new program.
      *     Must be terminated by NULL.
      * \param envp
      *     an array of strings of the form key=value, which are passed as
      *     environment to the new program.  Must be terminated by NULL.
      */
    virtual int execve(const char *filename, char *const *argv,
	char *const *envp) = 0;

    /**
      * The execvp method is used to execute the programs, searching
      * down $PATH for all filenames not containing a slash (/).
      */
    virtual int execvp(const char *filename, char *const *argv);

    /**
      * The fexecve method is used to execute the programs.
      */
    virtual int fexecve(int fd, char *const *argv, char *const *envp) = 0;

    /**
      * The access method may be used to check whether the process would
      * be allowed to read, write or test for existence of the file (or
      * other file system object) whose name is pathname. If pathname
      * is a symbolic link permissions of the file referred to by this
      * symbolic link are tested.
      */
    virtual int access(const char *path, int type) = 0;

    /**
      * The statfs method is used to retrieve information about a mounted
      * file system.
      *
      * \param path
      *     the path name of any file within the mounted filesystem.
      */
    virtual int statfs(const char *path, struct statfs *buf) = 0;

    /**
      * The fstatfs method is used to retrieve information about a mounted
      * file system.
      */
    virtual int fstatfs(int fd, struct statfs *buf) = 0;

    /**
      * The statfs64 method is used to retrieve information about a mounted
      * file system.
      *
      * \param path
      *     the path name of any file within the mounted filesystem.
      */
    virtual int statfs64(const char *path, struct statfs64 *buf) = 0;

    /**
      * The fstatfs64 method is used to retrieve information about a mounted
      * file system.
      */
    virtual int fstatfs64(int fd, struct statfs64 *buf) = 0;

    /**
      * The truncate method may be used to to truncate the file named by
      * `path' to at most `length' bytes in size.
      */
    virtual int truncate(const char *path, off_t length) = 0;

    /**
      * The truncate64 method may be used to to truncate the file named by
      * `path' to at most `length' bytes in size.
      */
    virtual int truncate64(const char *path, off64_t length) = 0;

    /**
      * The ftruncate method may be used to to truncate the file referred
      * to by `fd' to at most `length' bytes in size.
      */
    virtual int ftruncate(int fd, off_t length) = 0;

    /**
      * The ftruncate64 method may be used to to truncate the file referred
      * to by `fd' to at most `length' bytes in size.
      */
    virtual int ftruncate64(int fd, off64_t length) = 0;

    // Yet to intercept...
    // ssize_t pread(int fd, void *buf, size_t nbytes, off_t offset);
    // ssize_t pread64(int fd, void *buf, size_t nbytes, off64_t offset);
    // ssize_t pwrite(int fd, const void *buf, size_t nbytes, off_t offset);
    // ssize_t pwrite64(int fd, const void *buf, size_t nbytes, off64_t offset);
    // int chown(const char *path, uid_t owner, gid_t group);
    // int lchown(const char *path, uid_t owner, gid_t group);
    // int fchown(int fd, uid_t owner, gid_t group);
    // int revoke(const char *path);
    // int chroot(const char *path);

    /**
      * The error_message method is used to print error messages on
      * the standard error stream.
      */
    static void error_message(const char *fmt, ...)
					  __attribute__((format(printf, 1, 2)));

protected:
    /**
      * The default constructor.
      * Only derived classes may use the constructor.
      */
    plasticfs();

    /**
      * The get_path_max() method is used to get the value of PATH_MAX
      * as accurately as possible.
      */
    static size_t get_path_max();

    /**
      * The maxsymlinks() method is used to get the value of MAXSYMLINKS
      * as accurately as possible.
      */
    static int maxsymlinks();

private:
    /**
      * The master class variable is used to remember the base of the
      * list of file system filters to be applied.
      */
    static plasticfs *master;

    /**
      * The parse method is used to parse the PLASTICFS environmentr
      * variable, to determine which file system filters are to be
      * applied.
      */
    static plasticfs *parse();

    /**
      * The script_execute method is used by the execvp method when
      * invoking a shell script.
      */
    int script_execute(const char *filename, char *const *argv);

    /**
      * The copy constructor.  Do not use.
      */
    plasticfs(const plasticfs &);

    /**
      * The assignment operator.  Do not use.
      */
    plasticfs &operator=(const plasticfs &);
};

} // namespace libplasticfs

#endif // LIB_PLASTICFS_H
