//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_dlsym_rtld_next class
//

#pragma implementation "libplasticfs::plasticfs_dlsym_rtld_next"

#include <extra_symbol_definitions.h>
#include <cstring>
#include <cstdarg>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <plasticfs/dlsym_rtld_next.h>
#include <rcstring.h>

//
// Define the DLSYM_DEBUG symbol when you need to see what methods
// are being called.
//
//#define DLSYM_DEBUG

using namespace libplasticfs;


plasticfs_dlsym_rtld_next::~plasticfs_dlsym_rtld_next()
{
}


plasticfs_dlsym_rtld_next::plasticfs_dlsym_rtld_next() :
    plasticfs(),
    open_intercepted(0),
    close_intercepted(0),
    read_intercepted(0),
    write_intercepted(0),
    lseek_intercepted(0),
    lseek64_intercepted(0),
    readlink_intercepted(0),
    xstat_intercepted(0),
    xstat64_intercepted(0),
    lxstat_intercepted(0),
    lxstat64_intercepted(0),
    fxstat_intercepted(0),
    fxstat64_intercepted(0),
#ifdef DLSYM_OPENDIR
    opendir_intercepted(0),
    closedir_intercepted(0),
    readdir_intercepted(0),
    readdir64_intercepted(0),
    rewinddir_intercepted(0),
    seekdir_intercepted(0),
    telldir_intercepted(0),
    dirfd_intercepted(0),
    getcwd_intercepted(0),
#endif // DLSYM_OPENDIR
    getdents_intercepted(0),
    getdents64_intercepted(0),
    getdirentries_intercepted(0),
    getdirentries64_intercepted(0),
    chmod_intercepted(0),
    fchmod_intercepted(0),
    mkdir_intercepted(0),
    mkfifo_intercepted(0),
    unlink_intercepted(0),
    rename_intercepted(0),
    link_intercepted(0),
    symlink_intercepted(0),
    pathconf_intercepted(0),
    fpathconf_intercepted(0),
    xmknod_intercepted(0),
    fcntl_intercepted(0),
    fcntl64_intercepted(0),
    dup_intercepted(0),
    dup2_intercepted(0),
    chdir_intercepted(0),
    fchdir_intercepted(0),
    rmdir_intercepted(0),
    utime_intercepted(0),
    lutime_intercepted(0),
    futime_intercepted(0),
    execve_intercepted(0),
    fexecve_intercepted(0),
    access_intercepted(0),
    statfs_intercepted(0),
    fstatfs_intercepted(0),
    statfs64_intercepted(0),
    fstatfs64_intercepted(0),
    truncate_intercepted(0),
    ftruncate_intercepted(0),
    truncate64_intercepted(0),
    ftruncate64_intercepted(0)
{
}


//
// Use this macro to make sure the assignments are still type safe.
// An argument with a default value for find_symbol would sacrifice
// type safety.
//
#define SYM(name) #name
#define FIND_SYMBOL(name)						\
	FIND_SYMBOL_LIKE(name, name)
#define FIND_SYMBOL_LIKE(name, like)						\
	if (!name##_intercepted)					\
	{								\
	    name##_intercepted = (like##_t)find_symbol("__" SYM(name));	\
	    if (!name##_intercepted)					\
		name##_intercepted = like##_dummy;			\
	}


void *
plasticfs_dlsym_rtld_next::find_symbol(const char *name)
{
    //
    // dlsym takes a "handle" of a dynamic library returned by dlopen
    // and the null terminated symbol name, returning the address where
    // that symbol is loaded. If the symbol is not found, dlsym returns
    // NULL; however, the correct way to test for an error from dlsym is
    // to save the result of dlerror into a variable, and then check if
    // saved value is not NULL. This is because the value of the symbol
    // could actually be NULL. It is also necessary to save the results
    // of dlerror into a variable because if dlerror is called again,
    // it will return NULL.
    //
    void *result = 0;
    if (name[0] == '_' && name[1] == '_')
    {
	result = dlsym(RTLD_NEXT, name);
	const char *err = dlerror();
	if (err)
	{
	    name += 2;
	    result = dlsym(RTLD_NEXT, name);
	    err = dlerror();
	    if (err)
	    {
		error_message
		(
		    "plasticfs: dlsym(RTLD_NEXT, \"%s\") failed: %s\n",
		    name,
		    err
		);
		return 0;
	    }
	}
	if (result == 0)
	{
	    // That's weird, the symbol's value is NULL.
	    // I think this means it's an alias or something.
	    #ifdef DLSYM_DEBUG
	    error_message("plasticfs: found \"%s\" as NULL, trying again\n",
		name);
	    #endif

	    name += 2;
	    result = dlsym(RTLD_NEXT, name);
	    err = dlerror();
	    if (err)
	    {
		error_message
		(
		    "plasticfs: dlsym(RTLD_NEXT, \"%s\") failed: %s\n",
		    name,
		    err
		);
		return 0;
	    }
	}
    }
    else
    {
	result = dlsym(RTLD_NEXT, name);
	const char *err = dlerror();
	if (err)
	{
	    error_message
	    (
		"plasticfs: dlsym(RTLD_NEXT, \"%s\") failed: %s\n",
		name,
		err
	    );
	    return 0;
	}
    }

    if (result == 0)
    {
	error_message("plasticfs: found \"%s\" as NULL\n", name);
	return 0;
    }
#ifdef DLSYM_DEBUG
    error_message("plasticfs: found \"%s\" as %p\n", name, result);
#endif
    return result;
}


static int
open_dummy(const char *, int flags, ...)
{
    mode_t mode = 0;
    va_list ap;
    va_start(ap, flags);
    if (flags & O_CREAT)
	mode = va_arg(ap, mode_t);
    va_end(ap);

    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::open(const char *path, int flags, int mode)
{
    FIND_SYMBOL(open)
    if (flags & O_CREAT)
	return open_intercepted(path, flags, mode);
    return open_intercepted(path, flags);
}


static int
close_dummy(int)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::close(int filedesc)
{
    FIND_SYMBOL(close)
    return close_intercepted(filedesc);
}


static ssize_t
read_dummy(int, void *, size_t)
{
    errno = ENOSYS;
    return -1;
}


ssize_t
plasticfs_dlsym_rtld_next::read(int filedesc, void *data, size_t nbytes)
{
    FIND_SYMBOL(read)
    return read_intercepted(filedesc, data, nbytes);
}


static ssize_t
write_dummy(int, const void *, size_t)
{
    errno = ENOSYS;
    return -1;
}


ssize_t
plasticfs_dlsym_rtld_next::write(int filedesc, const void *data, size_t nbytes)
{
    FIND_SYMBOL(write)
    return write_intercepted(filedesc, data, nbytes);
}


static int
readlink_dummy(const char *, char *, size_t)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::readlink(const char *path, char *buffer, size_t nbytes)
{
    FIND_SYMBOL(readlink)
    return readlink_intercepted(path, buffer, nbytes);
}


static int
xstat_dummy(int, const char *, struct stat *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::xstat(int vers, const char *path, struct stat *buf)
{
    FIND_SYMBOL(xstat)
    return xstat_intercepted(vers, path, buf);
}


int
plasticfs_dlsym_rtld_next::lxstat(int vers, const char *path, struct stat *buf)
{
    FIND_SYMBOL_LIKE(lxstat, xstat)
    return lxstat_intercepted(vers, path, buf);
}


static int
xstat64_dummy(int, const char *, struct stat64 *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::xstat64(int version, const char *path,
    struct stat64 *buf)
{
    FIND_SYMBOL(xstat64)
    return xstat64_intercepted(version, path, buf);
}


int
plasticfs_dlsym_rtld_next::lxstat64(int version, const char *path,
    struct stat64 *buf)
{
    FIND_SYMBOL_LIKE(lxstat64, xstat64)
    return lxstat64_intercepted(version, path, buf);
}


static int
fxstat_dummy(int, int, struct stat *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::fxstat(int version, int filedesc, struct stat *buf)
{
    FIND_SYMBOL(fxstat)
    return fxstat_intercepted(version, filedesc, buf);
}


static int
fxstat64_dummy(int, int, struct stat64 *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::fxstat64(int vers, int filedesc, struct stat64 *buf)
{
    FIND_SYMBOL(fxstat64)
    return fxstat64_intercepted(vers, filedesc, buf);
}


static off_t
lseek_dummy(int, off_t, int)
{
    errno = ENOSYS;
    return -1;
}


off_t
plasticfs_dlsym_rtld_next::lseek(int fd, off_t offset, int whence)
{
    FIND_SYMBOL(lseek)
    return lseek_intercepted(fd, offset, whence);
}


static off64_t
lseek64_dummy(int, off64_t, int)
{
    errno = ENOSYS;
    return -1;
}


off64_t
plasticfs_dlsym_rtld_next::lseek64(int fd, off64_t offset, int whence)
{
    FIND_SYMBOL(lseek64)
    return lseek64_intercepted(fd, offset, whence);
}


#ifdef DLSYM_OPENDIR


static DIR *
opendir_dummy(const char *)
{
    errno = ENOSYS;
    return 0;
}


DIR *
plasticfs_dlsym_rtld_next::opendir(const char *path)
{
    FIND_SYMBOL(opendir)
    return opendir_intercepted(path);
}


static int 
closedir_dummy(DIR *)
{
    errno = ENOSYS;
    return -1;
}


int 
plasticfs_dlsym_rtld_next::closedir(DIR *dirp)
{
    FIND_SYMBOL(closedir)
    return closedir_intercepted(dirp);
}


static dirent *
readdir_dummy(DIR *)
{
    errno = ENOSYS;
    return 0;
}


dirent *
plasticfs_dlsym_rtld_next::readdir(DIR *dirp)
{
    FIND_SYMBOL(readdir)
    return readdir_intercepted(dirp);
}


static dirent64 *
readdir64_dummy(DIR *)
{
    errno = ENOSYS;
    return 0;
}


dirent64 *
plasticfs_dlsym_rtld_next::readdir64(DIR *dirp)
{
    FIND_SYMBOL(readdir64)
    return readdir64_intercepted(dirp);
}


static void 
rewinddir_dummy(DIR *)
{
    errno = ENOSYS;
}


void 
plasticfs_dlsym_rtld_next::rewinddir(DIR *dirp)
{
    FIND_SYMBOL(rewinddir)
    rewinddir_intercepted(dirp);
}


static void 
seekdir_dummy(DIR *, off_t)
{
    errno = ENOSYS;
}


void 
plasticfs_dlsym_rtld_next::seekdir(DIR *dirp, off_t offset)
{
    FIND_SYMBOL(seekdir)
    return seekdir_intercepted(dirp, offset);
}


static off_t 
telldir_dummy(DIR *)
{
    errno = ENOSYS;
    return -1;
}


off_t 
plasticfs_dlsym_rtld_next::telldir(DIR *dirp)
{
    FIND_SYMBOL(telldir)
    return telldir_intercepted(dirp);
}


static int 
dirfd_dummy(DIR *)
{
    errno = ENOSYS;
    return -1;
}


int 
plasticfs_dlsym_rtld_next::dirfd(DIR *dirp)
{
    FIND_SYMBOL(dirfd)
    return dirfd_intercepted(dirp);
}


static char *
getcwd_dummy(char *, size_t)
{
    errno = ENOSYS;
    return 0;
}


char *
plasticfs_dlsym_rtld_next::getcwd(char *buf, long nbytes)
{
    FIND_SYMBOL(getcwd)
    return getcwd_intercepted(buf, nbytes);
}


#endif // DLSYM_OPENDIR


#ifdef HAVE_GETDENTS

static int 
getdents_dummy(int, dirent *, unsigned)
{
    errno = ENOSYS;
    return -1;
}

#endif


int 
plasticfs_dlsym_rtld_next::getdents(int fd, dirent *buffer, unsigned nbytes)
{
    #ifdef DLSYM_DEBUG
    error_message("%s: %d: getdents(fd = %d, buffer = %p, nbytes = %u)\n",
	__FILE__, __LINE__, fd, buffer, nbytes);
    #endif
#ifdef HAVE_GETDENTS
    FIND_SYMBOL(getdents)
    return getdents_intercepted(fd, buffer, nbytes);
#else
    off_t base = lseek(fd, 0, SEEK_CUR);
    if (base == (off_t)(-1))
	return -1;
    return getdirentries(fd, (char *)buffer, nbytes, &base);
#endif
}


#ifdef HAVE_GETDENTS

static int 
getdents64_dummy(int, dirent64 *, unsigned)
{
    errno = ENOSYS;
    return -1;
}

#endif


int 
plasticfs_dlsym_rtld_next::getdents64(int fd, dirent64 *buffer, unsigned nbytes)
{
    #ifdef DLSYM_DEBUG
    error_message("%s: %d: getdents64(fd = %d, buffer = %p, nbytes = %u)\n",
	__FILE__, __LINE__, fd, buffer, nbytes);
    #endif
#ifdef HAVE_GETDENTS
    FIND_SYMBOL(getdents64)
    return getdents64_intercepted(fd, buffer, nbytes);
#else
    off64_t base = lseek64(fd, 0, SEEK_CUR);
    if (base == (off64_t)(-1))
	return -1;
    return getdirentries64(fd, (char *)buffer, nbytes, &base);
#endif
}


static ssize_t 
getdirentries_dummy(int, char *, size_t, off_t *)
{
    errno = ENOSYS;
    return -1;
}


ssize_t 
plasticfs_dlsym_rtld_next::getdirentries(int fd, char *buffer, size_t nbytes,
    off_t *base)
{
    #ifdef DLSYM_DEBUG
    error_message("%s: %d: getdirentriess(fd = %d, buffer = %p, nbytes = %ld, "
	"*base = %ld)\n", __FILE__, __LINE__, fd, buffer, (long)nbytes,
	(long)*base);
    #endif
    FIND_SYMBOL(getdirentries)
    return getdirentries_intercepted(fd, buffer, nbytes, base);
}


#ifdef HAVE_GETDIRENTRIES64

static ssize_t 
getdirentries64_dummy(int, char *, size_t, off64_t *)
{
    errno = ENOSYS;
    return -1;
}

#endif


ssize_t 
plasticfs_dlsym_rtld_next::getdirentries64(int fd, char *buffer, size_t nbytes,
    off64_t *base64)
{
    #ifdef DLSYM_DEBUG
    error_message("%s: %d: getdirentriess64(fd = %d, buffer = %p, "
	"nbytes = %ld, *base = %ld) {\n", __FILE__, __LINE__, fd, buffer,
	(long)nbytes, (long)*base);
    #endif
#ifdef HAVE_GETDIRENTRIES64
    FIND_SYMBOL(getdirentries64)
    ssize_t result = getdirentries64_intercepted(fd, buffer, nbytes, base64);
    #ifdef DLSYM_DEBUG
    error_message("%s: %d: *base = %ld; return %ld }\n", __FILE__, __LINE__,
	(long)*base, (long)result);
    #endif
    return result;
#else
    //
    // Debian *has* getdirentries64, it just doesn't export it from libc.so
    //
    // So we use getdirentries, and convert the result
    // dirent64 are larger than dirent, so scale everything
    // worst case 24 bytes vs 16 bytes
    //
    char temp[2048];
    size_t temp_size = sizeof(temp);
    if (temp_size > nbytes/2)
	temp_size = nbytes/2;
    off_t base = *base64;
    ssize_t result = getdirentries(fd, temp, temp_size, &base);
    if (result <= 0)
    	return result;
    *base64 = base;

    dirent64 *op = (dirent64 *)buffer;
    ssize_t pos = 0;
    while (pos < result)
    {
	dirent *ip = (dirent *)(temp + pos);
	op->d_ino = ip->d_ino;
	op->d_off = ip->d_off;
	op->d_type = ip->d_type;
	strcpy(op->d_name, ip->d_name);
	op->d_reclen = ip->d_reclen + sizeof(*op) - sizeof(*ip);
	op = (dirent64 *)((char *)op + op->d_reclen);
	pos += ip->d_reclen;
    }
    return ((char *)op - buffer);
#endif
}


static int 
chmod_dummy(const char *, mode_t)
{
    errno = ENOSYS;
    return -1;
}


int 
plasticfs_dlsym_rtld_next::chmod(const char *path, mode_t mode)
{
    FIND_SYMBOL(chmod)
    return chmod_intercepted(path, mode);
}


static int 
fchmod_dummy(int, mode_t)
{
    errno = ENOSYS;
    return -1;
}


int 
plasticfs_dlsym_rtld_next::fchmod(int fd, mode_t mode)
{
    FIND_SYMBOL(fchmod)
    return fchmod_intercepted(fd, mode);
}


int 
plasticfs_dlsym_rtld_next::mkdir(const char *path, mode_t mode)
{
    FIND_SYMBOL_LIKE(mkdir, chmod)
    return mkdir_intercepted(path, mode);
}


int 
plasticfs_dlsym_rtld_next::mkfifo(const char *path, mode_t mode)
{
    FIND_SYMBOL_LIKE(mkfifo, chmod)
    return mkfifo_intercepted(path, mode);
}


static int 
unlink_dummy(const char *)
{
    errno = ENOSYS;
    return -1;
}


int 
plasticfs_dlsym_rtld_next::unlink(const char *path)
{
    FIND_SYMBOL(unlink)
    return unlink_intercepted(path);
}


static int 
rename_dummy(const char *, const char *)
{
    errno = ENOSYS;
    return -1;
}


int 
plasticfs_dlsym_rtld_next::rename(const char *oldpath, const char *newpath)
{
    FIND_SYMBOL(rename)
    return rename_intercepted(oldpath, newpath);
}


int 
plasticfs_dlsym_rtld_next::link(const char *oldpath, const char *newpath)
{
    FIND_SYMBOL_LIKE(link, rename)
    return link_intercepted(oldpath, newpath);
}


int 
plasticfs_dlsym_rtld_next::symlink(const char *oldpath, const char *newpath)
{
    FIND_SYMBOL_LIKE(symlink, rename)
    return symlink_intercepted(oldpath, newpath);
}


static long
pathconf_dummy(const char *, int)
{
    errno = ENOSYS;
    return -1;
}


long 
plasticfs_dlsym_rtld_next::pathconf(const char *path, int name)
{
    FIND_SYMBOL(pathconf)
    return pathconf_intercepted(path, name);
}


static long
fpathconf_dummy(int, int)
{
    errno = ENOSYS;
    return -1;
}


long 
plasticfs_dlsym_rtld_next::fpathconf(int fd, int name)
{
    FIND_SYMBOL(fpathconf)
    return fpathconf_intercepted(fd, name);
}


static int
xmknod_dummy(int, const char *, mode_t, dev_t)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::xmknod(int ver, const char *path, mode_t mode,
    dev_t dev)
{
    FIND_SYMBOL(xmknod)
    return xmknod_intercepted(ver, path, mode, dev);
}


static int
fcntl_dummy(int, int cmd, ...)
{
    va_list ap;
    va_start(ap, cmd);
    void *arg = va_arg(ap, void *);
    va_end(ap);
    (void)arg;

    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::fcntl(int filedesc, int cmd, void *arg)
{
    FIND_SYMBOL(fcntl)

    //
    // How the function is called depends on the cmd value.
    //
#if 0
    // This is how glibc does it.
    fcntl_intercepted(filedesc, cmd, arg);
#else
    // This is how you may have expected it to be done.
    switch (cmd)
    {
    case F_DUPFD:
    case F_SETFD:
    case F_SETFL:
    case F_SETOWN:
    case F_SETSIG:
	return fcntl_intercepted(filedesc, cmd, (long)arg);
 
    case F_GETLK:
    case F_SETLK:
    case F_SETLKW:
	return fcntl_intercepted(filedesc, cmd, (flock *)arg);
 
#if defined(F_GETLK64) && (F_GETLK64 != F_GETLK)
    case F_GETLK64:
    case F_SETLK64:
    case F_SETLKW64:
	return fcntl_intercepted(filedesc, cmd, (flock64 *)arg);
#endif

    default:
	return fcntl_intercepted(filedesc, cmd);
    }
#endif
}



int
plasticfs_dlsym_rtld_next::fcntl64(int filedesc, int cmd, void *arg)
{
    FIND_SYMBOL_LIKE(fcntl64, fcntl)

    //
    // How the function is called depends on the cmd value.
    //
#if 1
    // This is how glibc does it
    return fcntl64_intercepted(filedesc, cmd, arg);
#else
    // This is how you may have expected it to be done.
    switch (cmd)
    {
    case F_DUPFD:
    case F_SETFD:
    case F_SETFL:
    case F_SETOWN:
    case F_SETSIG:
	return fcntl64_intercepted(filedesc, cmd, (long)arg);

    case F_GETLK:
    case F_SETLK:
    case F_SETLKW:
	return fcntl64_intercepted(filedesc, cmd, (flock *)arg);

    case F_GETLK64:
    case F_SETLK64:
    case F_SETLKW64:
	return fcntl64_intercepted(filedesc, cmd, (flock64 *)arg);

    default:
	return fcntl64_intercepted(filedesc, cmd);
    }
#endif
}


int
plasticfs_dlsym_rtld_next::dup(int oldfd)
{
    FIND_SYMBOL_LIKE(dup, close)
    return dup_intercepted(oldfd);
}


static int
dup2_dummy(int, int)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::dup2(int oldfd, int newfd)
{
    FIND_SYMBOL(dup2)
    return dup2_intercepted(oldfd, newfd);
}


int
plasticfs_dlsym_rtld_next::chdir(const char *path)
{
    FIND_SYMBOL_LIKE(chdir, unlink)
    return chdir_intercepted(path);
}


int
plasticfs_dlsym_rtld_next::fchdir(int fd)
{
    FIND_SYMBOL_LIKE(fchdir, close)
    return fchdir_intercepted(fd);
}


int
plasticfs_dlsym_rtld_next::rmdir(const char *path)
{
    FIND_SYMBOL_LIKE(rmdir, unlink)
    return rmdir_intercepted(path);
}


static int
utime_dummy(const char *, utimbuf *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::utime(const char *path, utimbuf *buf)
{
    FIND_SYMBOL(utime)
    return utime_intercepted(path, buf);
}


int
plasticfs_dlsym_rtld_next::lutime(const char *path, utimbuf *buf)
{
    FIND_SYMBOL_LIKE(lutime, utime)
    return lutime_intercepted(path, buf);
}


static int
futime_dummy(int, utimbuf *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::futime(int fd, utimbuf *buf)
{
    FIND_SYMBOL(futime)
    return futime_intercepted(fd, buf);
}


static int
execve_dummy(const char *, char *const *, char *const *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::execve(const char *filename, char *const *argv,
    char *const *envp)
{
    FIND_SYMBOL(execve)
    return execve_intercepted(filename, argv, envp);
}


static int
fexecve_dummy(int, char *const *, char *const *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::fexecve(int fd, char *const *argv, char *const *envp)
{
    FIND_SYMBOL(fexecve)
    return fexecve_intercepted(fd, argv, envp);
}


static int
access_dummy(const char *, int)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::access(const char *path, int type)
{
    FIND_SYMBOL(access)
    return access_intercepted(path, type);
}


static int
statfs_dummy(const char *, struct statfs *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::statfs(const char *path, struct statfs *buf)
{
    FIND_SYMBOL(statfs)
    return statfs_intercepted(path, buf);
}


static int
fstatfs_dummy(int, struct statfs *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::fstatfs(int fd, struct statfs *buf)
{
    FIND_SYMBOL(fstatfs)
    return fstatfs_intercepted(fd, buf);
}


static int
statfs64_dummy(const char *, struct statfs64 *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::statfs64(const char *path, struct statfs64 *buf)
{
    FIND_SYMBOL(statfs64)
    return statfs64_intercepted(path, buf);
}


static int
fstatfs64_dummy(int, struct statfs64 *)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::fstatfs64(int fd, struct statfs64 *buf)
{
    FIND_SYMBOL(fstatfs64)
    return fstatfs64_intercepted(fd, buf);
}


static int
truncate_dummy(const char *, off_t)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::truncate(const char *path, off_t length)
{
    FIND_SYMBOL(truncate)
    return truncate_intercepted(path, length);
}


static int
ftruncate_dummy(int, off_t)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::ftruncate(int fd, off_t length)
{
    FIND_SYMBOL(ftruncate)
    return ftruncate_intercepted(fd, length);
}


static int
truncate64_dummy(const char *, off64_t)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::truncate64(const char *path, off64_t length)
{
    FIND_SYMBOL(truncate64)
    return truncate64_intercepted(path, length);
}


static int
ftruncate64_dummy(int, off64_t)
{
    errno = ENOSYS;
    return -1;
}


int
plasticfs_dlsym_rtld_next::ftruncate64(int fd, off64_t length)
{
    FIND_SYMBOL(ftruncate64)
    return ftruncate64_intercepted(fd, length);
}
