//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_execvp class
//

#include <plasticfs.h>
#include <cerrno>
#include <cstring>
#include <cstdlib> // for getenv
#include <paths.h>
#include <unistd.h> // for environ

using namespace libplasticfs;

// Portions of this file are derived from GNU Libc 2.3 posix/exevvp.c
// Copyright (C) 1991,92,95,96,97,98,99,2002 Free Software Foundation, Inc.


//
// The file is accessible but it is not an executable file.  Invoke the
// shell to interpret it as a script.
//
int
plasticfs::script_execute(const char *file, char *const *argv)
{
    // Count the arguments.
    size_t argc = 0;
    while (argv[argc])
	++argc;

    //
    // Construct an argument list for the shell.
    //
    char **new_argv = new char *[argc + 2];
    new_argv[0] = _PATH_BSHELL;
    new_argv[1] = (char *)file;
    for (size_t j = 1; j < argc; ++j)
	new_argv[j + 1] = (char *)argv[j];
    new_argv[argc + 1] = 0;

    //
    // Execute the shell.
    //
    int result = execve(new_argv[0], new_argv, environ);
    delete [] new_argv;
    return result;
}


//
// Execute FILE, searching in the `PATH' environment variable if
// it contains no slashes, with arguments ARGV and environment from
// `environ'.
//
int
plasticfs::execvp(const char *filename, char *const *argv)
{
    if (!*filename)
    {
	// We check the simple case first.
	errno = ENOENT;
	return -1;
    }

    if (strchr(filename, '/'))
    {
	// Don't search when the filename contains a slash.
	int result = execve(filename, argv, environ);
	if (errno == ENOEXEC)
	    result = script_execute(filename, argv);
	return result;
    }

    //
    // Get the PATH environment variable.  This contains a colon-separated
    // list for directories to search for commands.
    //
    char *path = getenv("PATH");
    if (!path)
    {
	//
	// There is no `PATH' in the environment.
	// Use the default, with "." at the start.
	//
	path = ".:" _PATH_DEFPATH;
    }

    size_t len = strlen(filename) + 1;
    size_t pathlen = strlen(path);
    char *name_base = new char [pathlen + len + 1];

    // Copy the file name at the top
    // and add the slash.
    char *name = (char *)memcpy(name_base + pathlen + 1, filename, len);
    *--name = '/';

    bool got_eacces = false;
    for (;;)
    {
	//
	// Find the end of the next path element
	//
	char *path_begin = path;
	char *path_end = path;
	for (;;)
	{
	    if (!*path)
	    {
		path_end = path;
		break;
	    }
	    if (*path == ':')
	    {
		path_end = path;
		++path;
		break;
	    }
	    ++path;
	}

	//
	// Two adjacent colons, or a colon at the beginning or the end
	// of `PATH' means to search the current directory.
	//
	if (path_begin == path_end)
	{
	    path_begin = ".";
	    path_end = path_begin + 1;
	}

	//
	// Copy this path element to the start of the name.
	//
	size_t path_len = path_end - path_begin;
	char *startp = name - path_len;
	memcpy(startp, path_begin, path_len);

	// Try to execute this name.  If it works, execv will not return.
	int result = execve(startp, argv, environ);
	if (errno == ENOEXEC)
	    result = script_execute(startp, argv);

	switch (errno)
	{
	case EACCES:
	    //
	    // Record the we got a `Permission denied' error.  If we end
	    // up finding no executable we can use, we want to diagnose
	    // that we did find one but were denied access.
	    //
	    got_eacces = true;
	    break;

	case ENOENT:
	case ESTALE:
	case ENOTDIR:
	    //
	    // Those errors indicate the file is missing or not executable
	    // by us, in which case we want to just try the next path
	    // directory.
	    //
	    break;

	default:
	    //
	    // Some other error means we found an executable file, but
	    // something went wrong executing it; return the error to
	    // our caller.
	    //
	    return result;
	}

	//
	// If we are at the end of the path, we are done.
	//
	if (!*path)
	    break;
    }

    //
    // We tried every path element and none of them worked.
    //
    if (got_eacces)
    {
	// At least one failure was due to permissions, so report
	// that error.
	errno = EACCES;
    }

    //
    // Free the buffer we were using to build the filename.
    //
    delete [] name_base;

    //
    // Report failure.
    //
    return -1;
}
