//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter class
//

#pragma implementation "libplasticfs::plasticfs_filter"

#include <plasticfs/filter.h>
#include <cerrno>
#include <cstring>
#include <errno_insulator.h>
#include <sys/stat.h>

using namespace libplasticfs;

//
// Define the FILTER_DEBUG symbol when you need to see what methods
// are being called.
//
//#define FILTER_DEBUG


plasticfs_filter::~plasticfs_filter()
{
    delete_deeper();
}


plasticfs_filter::plasticfs_filter(plasticfs *arg) :
    deeper(arg)
{
}


rcstring
plasticfs_filter::cwd()
{
    errno_insulator guard;
    if (!cwd_cache.length())
    {
	char buffer[4000];
	if (!get_deeper_cwd(buffer, sizeof(buffer)))
	{
	    errno_insulator g2;
	    error_message("getcwd: %s", g2.gets());
	    cwd_cache = rcstring("/unable/to/determine/current/directory");
	}
	else
	    cwd_cache = rcstring(buffer);
    }
    return cwd_cache;
}


rcstring
plasticfs_filter::path_join(const rcstring &s1, const rcstring &s2)
{
    if (s2.length() == 0 || s2 == rcstring("."))
	return s1;
    size_t len1 = s1.length();
    while (len1 && s1[len1 - 1] == '/')
	--len1;
    return (rcstring(s1.c_str(), len1) + rcstring("/") + s2);
}


rcstring
plasticfs_filter::make_absolute(const rcstring &path)
{
    if (path.length() && path[0] == '/')
	return path;
    return path_join(cwd(), path);
}


bool
plasticfs_filter::relative_name_mapping(const char *, rcstring &)
{
    return false;
}


static bool
is_end_of_path(const char *s)
{
    while (*s == '/')
	++s;
    return !*s;
}


bool
plasticfs_filter::is_below_directory(const rcstring &mount_point,
    const char *dest_buf, size_t &nbytes)
{
    size_t mtptlen = mount_point.length();
    if (mtptlen == 1)
    {
	// root is a special case.
	nbytes = 0;
	return true;
    }
    if (mtptlen == nbytes)
	return (0 == memcmp(mount_point.c_str(), dest_buf, nbytes));
    if
    (
	mtptlen < nbytes
    &&
	0 == memcmp(mount_point.c_str(), dest_buf, mtptlen)
    &&
	dest_buf[mtptlen] == '/'
    )
    {
	nbytes = mtptlen + 1;
	return true;
    }
    return false;
}


bool
plasticfs_filter::relative_realpath(const rcstring &name, rcstring &abspath,
    rcstring &relpath, bool file_must_exist, bool do_last_element)
{
    #ifdef FILTER_DEBUG
    error_message("%s: %d: relative_realpath(name=\"%s\", file_must_exist=%s, "
	"do_last_element=%s) {\n", __FILE__, __LINE__, name.c_str(),
	(file_must_exist ? "true" : "false"),
	(do_last_element ? "true" : "false"));
    #endif
    // The delete_string_guard_t class may be used to ensure that the
    // subject string is guaranteed to be deleted when it goes out
    // of scope.  It's even exception safe.
    class delete_string_guard_t
    {
    private:
	char *&subject;
    public:
	delete_string_guard_t(char *&arg): subject(arg) { }
	~delete_string_guard_t() { if (subject) delete [] subject; }
    };

    //
    // We make the name absolute as the first step.
    //
    rcstring absname = make_absolute(name);

    size_t path_max = get_path_max();
    char *dest_buf = new char[path_max];
    delete_string_guard_t dest_guard(dest_buf);
    char *dest = dest_buf;
    char *dest_buf_limit = dest_buf + path_max;
    *dest++ = '/';

    char *name_buf = 0;
    delete_string_guard_t name_buf_guard(name_buf);

    char *readlink_buf = 0;
    delete_string_guard_t readlink_buf_guard(readlink_buf);

    //
    // The link counter is to count the number of symlinks we have
    //followed, in case we get caught in a symbolic link loop.
    //
    int num_links = maxsymlinks();

    //
    // We have to track whether or not the resulting path (in dest_buf)
    // is below the mount point.  This is when we switch from using
    // deeper->lstat to relative_lstat.
    //
    bool below_mount_point = false;

    //
    // This tracks where in dest_buf the relative portion starts,
    // if below_mount_point is true.
    //
    size_t relative_start_index = 0;

    //
    // The 'start' points at the start of the next path element, and
    // 'end' eventually points at the end of the next path element.
    // While we initially use 'absname', we could redirect it into
    // 'name_buf' when we chase symbolic links, so avoid assumptions
    // about which string we are pointing into.
    //
    const char *start;
    const char *end;
    for (start = end = absname.c_str(); ; start = end)
    {
	#ifdef FILTER_DEBUG
	error_message("%s: %d: dest=\"%.*s\"\n", __FILE__, __LINE__,
	    (int)(dest - dest_buf), dest_buf);
	#endif
	//
	// Keep track of whether or not we have passed below the mount
	// point, because it tells us we need to whitch which lstat
	// method we use.
	//
	if (!below_mount_point)
	{
	    size_t nbytes = dest - dest_buf;
	    if (is_below_mount_point(dest_buf, nbytes))
	    {
		#ifdef FILTER_DEBUG
		error_message("%s: %d: is below mount point\n", __FILE__,
		    __LINE__);
		#endif
		below_mount_point = true;
		relative_start_index = nbytes;
	    }
	}

	// Skip all path separator characters.
	while (*start == '/')
	    ++start;
	#ifdef FILTER_DEBUG
	error_message("%s: %d: start=\"%s\"\n", __FILE__, __LINE__, start);
	#endif

	// Find the end of the next path element.
	for (end = start; *end && *end != '/'; ++end)
	    ;
	size_t element_length = end - start;
	#ifdef FILTER_DEBUG
	error_message("%s: %d: element=\"%.*s\"\n", __FILE__, __LINE__,
	    (int)(end - start), start);
	#endif
	if (element_length == 0)
	    break;

	if (element_length == 1 && start[0] == '.')
	{
	    // Ignore "." in the path.
	}
	else if (element_length == 2 && start[0] == '.' && start[1] == '.')
	{
	    // Back up to previous path element, ignore if at root already.
	    while (dest > dest_buf + 1)
	    {
		--dest;
		if (*dest == '/')
		    break;
	    }

	    //
	    // Have to watch out for backing out previous to the
	    // mount point.  The easiest way to do this is to rescan
	    // at the top of the loop.
	    //
	    below_mount_point = false;
	    relative_start_index = 0;
	}
	else
	{
	    //
	    // Make sure there is enough room in dest_buf for the next path
	    // element (including the '/' and the NUL).
	    //
	    if (dest + element_length + 2 > dest_buf_limit)
	    {
		errno = ENAMETOOLONG;
		#ifdef FILTER_DEBUG
		error_message("%s: %d: return ENAMETOOLONG }\n", __FILE__, __LINE__);
		#endif
		return false;
	    }

	    //
	    // Copy the next element to the end of the result buffer.
	    //
	    if (dest[-1] != '/')
	    {
		*dest++ = '/';
		if (below_mount_point)
		{
		    const char *rel_dest_buf = dest_buf + relative_start_index;
		    if (*rel_dest_buf == '/')
			++relative_start_index;
		}
	    }
	    memcpy(dest, start, element_length);
	    dest += element_length;

	    //
	    // Map the file name, if necessary.
	    //
	    if (below_mount_point)
	    {
		rcstring new_element;
		*dest = '\0';
		if (relative_name_mapping(dest_buf, new_element))
		{
		    char *newp =
			dest - element_length + new_element.length() + 2;
		    if (newp > dest_buf_limit)
		    {
			errno = ENAMETOOLONG;
			#ifdef FILTER_DEBUG
			error_message("%s: %d: return ENAMETOOLONG }\n",
			    __FILE__, __LINE__);
			#endif
			return false;
		    }
		    dest -= element_length;
		    element_length = new_element.length();
		    memcpy(dest, new_element.c_str(), element_length);
		    dest += element_length;
		}
	    }

	    //
	    // We need to know if this is the last path element.
	    //
	    // Some file actions (e.g. readlink) must not follow the last
	    // path element, but it is allowed to follow all the others.
	    //
	    if (!do_last_element && is_end_of_path(end))
		continue;

	    //
	    // See what the file is.
	    //
	    *dest = '\0';
	    struct stat st;
	    int statretval;
	    rcstring relpath;
	    if (below_mount_point)
	    {
		const char *rel_dest_buf = dest_buf + relative_start_index;
		statretval = relative_lstat(rel_dest_buf, &st);
	    }
	    else
	    {
		statretval = deeper->lxstat(_STAT_VER, dest_buf, &st);
	    }
	    if (statretval < 0)
	    {
		if (!file_must_exist && is_end_of_path(end) && errno == ENOENT)
		{
		    //
		    // We've walked off the end of reality,
		    // pretend it wasn't a symlink,
		    // move on to the next path element.
		    //
		    continue;
		}
		#ifdef FILTER_DEBUG
		error_message("%s: %d: return -1 (stat failed) }\n", __FILE__, __LINE__);
		#endif
		return false;
	    }

	    //
	    // If it's not a symbolic link,
	    // move on to the next path element.
	    //
	    if (!S_ISLNK(st.st_mode))
	    {
		if (S_ISDIR(st.st_mode))
		    continue;
		if (is_end_of_path(end))
		    continue;
		errno = ENOTDIR;
		#ifdef FILTER_DEBUG
		error_message("%s: %d: return ENOTDIR }\n", __FILE__, __LINE__);
		#endif
		return false;
	    }

	    //
	    // Make sure we aren't caught in a symbolic link loop.
	    //
	    if (--num_links <= 0)
	    {
		errno = ELOOP;
		#ifdef FILTER_DEBUG
		error_message("%s: %d: return ELOOP }\n", __FILE__, __LINE__);
		#endif
		return false;
	    }

	    //
	    // Allocate a buffer to read the symbolic link into.
	    //
	    if (!readlink_buf)
	    {
		readlink_buf = new char [path_max];
		if (!readlink_buf)
		{
		    #ifdef FILTER_DEBUG
		    error_message("%s: %d: return ENOMEM }\n", __FILE__, __LINE__);
		    #endif
		    return false;
		}
	    }

	    //
	    // Read the symbolic link.
	    //
	    rcstring readlink_value;
	    int n = -1;
	    if (below_mount_point)
	    {
		const char *rel_dest_buf = dest_buf + relative_start_index;
		n = relative_readlink(rel_dest_buf, readlink_buf, path_max);
	    }
	    else
	    {
		n = deeper->readlink(dest_buf, readlink_buf, path_max);
	    }
	    if (n < 0)
	    {
		#ifdef FILTER_DEBUG
		error_message("%s: %d: return -1 (readlink failed) }\n", __FILE__, __LINE__);
		#endif
		return false;
	    }
	    if (n == 0)
	    {
		readlink_buf[0] = '.';
		n = 1;
	    }

	    //
	    // Look at what's left to scan.  It's either empty, or starts
	    // with a slash.  If it isn't empty, append it to the end
	    // of the symlink value.
	    //
	    // Make sure it will fit on the
	    // end of 'readlink_buf', including the NUL.
	    //
	    size_t len = strlen(end);
	    if (n + len + 1 > path_max)
	    {
		errno = ENAMETOOLONG;
		#ifdef FILTER_DEBUG
		error_message("%s: %d: return ENAMETOOLONG }\n", __FILE__, __LINE__);
		#endif
		return false;
	    }

	    //
	    // Copy it to the end of 'readlink_buf' and pretend it was
	    // always that long.  This end part could be coming out
	    // of name_buf, so we can't simply copy it straight there.
	    //
	    if (len)
	    {
		memcpy(readlink_buf + n, end, len);
		n += len;
	    }

	    //
	    // Allocate a buffer to hold the diverted name
	    // constructed from the symbolic link.
	    //
	    if (!name_buf)
	    {
		name_buf = new char [path_max + 1];
		if (!name_buf)
		{
		    #ifdef FILTER_DEBUG
		    error_message("%s: %d: return ENOMEM }\n", __FILE__, __LINE__);
		    #endif
		    return false;
		}
	    }

	    //
	    // Copy the symlink into name_buf.  Don't worry about the
	    // leading path, we just continue from the point we were at.
	    //
	    memcpy(name_buf, readlink_buf, n);
	    name_buf[n] = 0;
	    end = name_buf;

	    //
	    // What we do with the result buffer depends on whether it's
	    // an absolute or relative symbolic link.
	    //
	    if (readlink_buf[0] == '/')
	    {
		// It's an absolute symlink, so restart the result.
		// And watch out for the mount point indicator.
		dest = dest_buf + 1;
		below_mount_point = false;
	    }
	    else
	    {
		// It's a relative symlink, back up one path element.
		dest -= element_length;
		if (dest > dest_buf + 1)
		    --dest;
	    }
	}
    }
    if (dest > dest_buf + 1 && dest[-1] == '/')
    {
	--dest;
	if (relative_start_index > (size_t)(dest - dest_buf))
	    relative_start_index = dest - dest_buf;
    }

    //
    // Write out the two results, absolute and relative.
    //
    abspath = rcstring(dest_buf, dest - dest_buf);
    if (below_mount_point)
    {
	relpath =
	    rcstring
	    (
		dest_buf + relative_start_index,
		dest - dest_buf - relative_start_index
	    );
    }
    else
	relpath = abspath;

    //
    // Report success.
    //
    #ifdef FILTER_DEBUG
    error_message("%s: %d: return iabspath=\"%s\", relpath=\"%s\" }\n",
	__FILE__, __LINE__, abspath.c_str(), relpath.c_str());
    #endif
    return true;
}


int
plasticfs_filter::open(const char *path, int flags, int mode)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->open(path, flags, mode);
}


int
plasticfs_filter::close(int filedesc)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->close(filedesc);
}


ssize_t
plasticfs_filter::read(int filedesc, void *data, size_t nbytes)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->read(filedesc, data, nbytes);
}


ssize_t
plasticfs_filter::write(int filedesc, const void *data, size_t nbytes)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->write(filedesc, data, nbytes);
}


int
plasticfs_filter::readlink(const char *path, char *buffer, size_t nbytes)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->readlink(path, buffer, nbytes);
}


#ifdef FILTER_STAT


int
plasticfs_filter::stat(const char *path, struct stat *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->stat(path, buf);
}


int
plasticfs_filter::stat64(const char *path, struct stat64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->stat64(path, buf);
}


int
plasticfs_filter::lstat(const char *path, struct stat *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lstat(path, buf);
}


int
plasticfs_filter::lstat64(const char *path, struct stat64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lstat64(path, buf);
}


int
plasticfs_filter::fstat(int filedesc, struct stat *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fstat(filedesc, buf);
}


int
plasticfs_filter::fstat64(int filedesc, struct stat64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fstat64(filedesc, buf);
}


#endif // FILTER_STAT


int
plasticfs_filter::xstat(int version, const char *path, struct stat *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->xstat(version, path, buf);
}


int
plasticfs_filter::xstat64(int version, const char *path, struct stat64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->xstat64(version, path, buf);
}


int
plasticfs_filter::lxstat(int version, const char *path, struct stat *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lxstat(version, path, buf);
}


int
plasticfs_filter::lxstat64(int version, const char *path, struct stat64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lxstat64(version, path, buf);
}


int
plasticfs_filter::fxstat(int version, int filedesc, struct stat *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fxstat(version, filedesc, buf);
}


int
plasticfs_filter::fxstat64(int version, int filedesc, struct stat64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fxstat64(version, filedesc, buf);
}


off_t
plasticfs_filter::lseek(int fd, off_t offset, int whence)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lseek(fd, offset, whence);
}


off64_t
plasticfs_filter::lseek64(int fd, off64_t offset, int whence)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lseek64(fd, offset, whence);
}


#ifdef FILTER_OPENDIR


DIR *
plasticfs_filter::opendir(const char *path)
{
    if (!deeper)
    {
	errno = EINVAL;
	return 0;
    }
    return deeper->opendir(path);
}


int 
plasticfs_filter::closedir(DIR *dirp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->closedir(dirp);
}


dirent *
plasticfs_filter::readdir(DIR *dirp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return 0;
    }
    return deeper->readdir(dirp);
}


dirent64 *
plasticfs_filter::readdir64(DIR *dirp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return 0;
    }
    return deeper->readdir64(dirp);
}


void 
plasticfs_filter::rewinddir(DIR *dirp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return;
    }
    deeper->rewinddir(dirp);
}


void 
plasticfs_filter::seekdir(DIR *dirp, off_t offset)
{
    if (!deeper)
    {
	errno = EINVAL;
	return;
    }
    deeper->seekdir(dirp, offset);
}


off_t 
plasticfs_filter::telldir(DIR *dirp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->telldir(dirp);
}


int 
plasticfs_filter::dirfd(DIR *dirp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->dirfd(dirp);
}


char *
plasticfs_filter::getcwd(char *buf, long size)
{
    if (!deeper)
    {
	errno = EINVAL;
	return 0;
    }
    return deeper->getcwd(buf, size);
}

#endif // FILTER_OPENDIR


int 
plasticfs_filter::getdents(int fd, dirent *buf, unsigned len)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->getdents(fd, buf, len);
}


int 
plasticfs_filter::getdents64(int fd, dirent64 *buf, unsigned len)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->getdents64(fd, buf, len);
}


ssize_t 
plasticfs_filter::getdirentries(int fd, char *buf, size_t nbytes, off_t *base)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->getdirentries(fd, buf, nbytes, base);
}


ssize_t 
plasticfs_filter::getdirentries64(int fd, char *buf, size_t len, off64_t *base)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->getdirentries64(fd, buf, len, base);
}


char *
plasticfs_filter::get_deeper_cwd(char *buf, long size)
{
    if (!deeper)
    {
	errno = EINVAL;
	return 0;
    }
    return deeper->getcwd(buf, size);
}


int
plasticfs_filter::chmod(const char *path, mode_t mode)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->chmod(path, mode);
}


int
plasticfs_filter::fchmod(int fd, mode_t mode)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fchmod(fd, mode);
}


int
plasticfs_filter::mkdir(const char *path, mode_t mode)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->mkdir(path, mode);
}


int
plasticfs_filter::mkfifo(const char *path, mode_t mode)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->mkfifo(path, mode);
}


int
plasticfs_filter::unlink(const char *path)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->unlink(path);
}


int
plasticfs_filter::rmdir(const char *path)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->rmdir(path);
}


int
plasticfs_filter::rename(const char *oldpath, const char *newpath)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->rename(oldpath, newpath);
}


int
plasticfs_filter::link(const char *oldpath, const char *newpath)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->link(oldpath, newpath);
}


int
plasticfs_filter::symlink(const char *oldpath, const char *newpath)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->symlink(oldpath, newpath);
}


long
plasticfs_filter::pathconf(const char *path, int name)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->pathconf(path, name);
}


long
plasticfs_filter::fpathconf(int fd, int name)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fpathconf(fd, name);
}


int
plasticfs_filter::xmknod(int ver, const char *path, mode_t mode, dev_t dev)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->xmknod(ver, path, mode, dev);
}


int
plasticfs_filter::fcntl(int filedesc, int cmd, void *arg)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fcntl(filedesc, cmd, arg);
}


int
plasticfs_filter::fcntl64(int filedesc, int cmd, void *arg)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fcntl64(filedesc, cmd, arg);
}


int
plasticfs_filter::dup(int oldfd)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->dup(oldfd);
}


int
plasticfs_filter::dup2(int oldfd, int newfd)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->dup2(oldfd, newfd);
}


int
plasticfs_filter::chdir(const char *path)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    int result = deeper->chdir(path);
    if (result >= 0)
	cwd_cache = "";
    return result;
}


int
plasticfs_filter::fchdir(int fd)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    int result = deeper->fchdir(fd);
    if (result >= 0)
	cwd_cache = "";
    return result;
}


int
plasticfs_filter::utime(const char *path, utimbuf *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->utime(path, buf);
}


int
plasticfs_filter::lutime(const char *path, utimbuf *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->lutime(path, buf);
}


int
plasticfs_filter::futime(int fd, utimbuf *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->futime(fd, buf);
}


int
plasticfs_filter::execve(const char *filename, char *const *argv,
    char *const *envp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->execve(filename, argv, envp);
}


int
plasticfs_filter::fexecve(int fd, char *const *argv, char *const *envp)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fexecve(fd, argv, envp);
}


int
plasticfs_filter::access(const char *path, int type)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->access(path, type);
}


int
plasticfs_filter::statfs(const char *path, struct statfs *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->statfs(path, buf);
}


int
plasticfs_filter::fstatfs(int fd, struct statfs *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fstatfs(fd, buf);
}


int
plasticfs_filter::statfs64(const char *path, struct statfs64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->statfs64(path, buf);
}


int
plasticfs_filter::fstatfs64(int fd, struct statfs64 *buf)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->fstatfs64(fd, buf);
}


int
plasticfs_filter::truncate(const char *path, off_t length)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->truncate(path, length);
}


int
plasticfs_filter::ftruncate(int fd, off_t length)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->ftruncate(fd, length);
}


int
plasticfs_filter::truncate64(const char *path, off64_t length)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->truncate64(path, length);
}


int
plasticfs_filter::ftruncate64(int fd, off64_t length)
{
    if (!deeper)
    {
	errno = EINVAL;
	return -1;
    }
    return deeper->ftruncate64(fd, length);
}
