//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/plasticfs/filter.cc
//

#ifndef LIB_PLASTICFS_FILTER_H
#define LIB_PLASTICFS_FILTER_H

#pragma interface "libplasticfs::plasticfs_filter"

#include <plasticfs.h>
#include <rcstring.h>

//
// Define the FILTER_OPENDIR symbol if you want to pass
// through to the deeper class' opendir family, rather than use
// plasticfs's implementation.  If you do this, you don't see the
// open/getdirentries/close calls, only the opendir family of calls.
//
// #define FILTER_OPENDIR

//
// Define the FILTER_STAT symbol if you want to pass through to the
// deeper class's version-less stat family, rather than use plasticfs's
// implementation.  If you do this, you don't see the the xstat family
// of calls, only the version-less stat family of calls.
//
// #define FILTER_STAT

namespace libplasticfs {

/**
  * The plasticfs_filter class is used to represent a generic filter file
  * system.  The base class does nothing, except pass methods through.
  * Derived classes override those methods necessary foe their specific
  * behaviour, with the rest undisturbed.
  */
class plasticfs_filter:
    public plasticfs
{
public:
    /**
      * The destructor.
      */
    virtual ~plasticfs_filter();

    /**
      * The constructor.
      */
    plasticfs_filter(plasticfs *deeper);

    // See base class for documentation.
    int open(const char *path, int flags, int mode);

    // See base class for documentation.
    int close(int filedesc);

    // See base class for documentation.
    ssize_t read(int, void *, size_t);

    // See base class for documentation.
    ssize_t write(int, const void *, size_t);

    // See base class for documentation.
    off_t lseek(int, off_t, int);

    // See base class for documentation.
    off64_t lseek64(int, off64_t, int);

    // See base class for documentation.
    int readlink(const char *, char *, size_t);

#ifdef FILTER_STAT
    // See base class for documentation.
    int stat(const char *, struct stat *);

    // See base class for documentation.
    int stat64(const char *, struct stat64 *);

    // See base class for documentation.
    int lstat(const char *, struct stat *);

    // See base class for documentation.
    int lstat64(const char *, struct stat64 *);

    // See base class for documentation.
    int fstat(int filedesc, struct stat *);

    // See base class for documentation.
    int fstat64(int filedesc, struct stat64 *);
#endif // FILTER_STAT

    // See base class for documentation.
    int xstat(int, const char *, struct stat *);

    // See base class for documentation.
    int xstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int lxstat(int, const char *, struct stat *);

    // See base class for documentation.
    int lxstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int fxstat(int, int filedesc, struct stat *);

    // See base class for documentation.
    int fxstat64(int, int filedesc, struct stat64 *);

#ifdef FILTER_OPENDIR
    // See base class for documentation.
    DIR *opendir(const char *);

    // See base class for documentation.
    int closedir(DIR *);

    // See base class for documentation.
    dirent *readdir(DIR *);

    // See base class for documentation.
    dirent64 *readdir64(DIR *);

    // See base class for documentation.
    void rewinddir(DIR *);

    // See base class for documentation.
    void seekdir(DIR *, off_t);

    // See base class for documentation.
    off_t telldir(DIR *);

    // See base class for documentation.
    int dirfd(DIR *);

    // See base class for documentation.
    char *getcwd(char *, long);
#endif // FILTER_OPENDIR

    // See base class for documentation.
    int getdents(int, dirent *, unsigned);

    // See base class for documentation.
    int getdents64(int, dirent64 *, unsigned);

    // See base class for documentation.
    ssize_t getdirentries(int, char *, size_t, off_t *);

    // See base class for documentation.
    ssize_t getdirentries64(int, char *, size_t, off64_t *);

    // See base class for documentation.
    int chmod(const char *, mode_t);

    // See base class for documentation.
    int fchmod(int, mode_t);

    // See base class for documentation.
    int mkdir(const char *, mode_t);

    // See base class for documentation.
    int mkfifo(const char *, mode_t);

    // See base class for documentation.
    int link(const char *, const char *);

    // See base class for documentation.
    int unlink(const char *);

    // See base class for documentation.
    int rename(const char *, const char *);

    // See base class for documentation.
    long pathconf(const char *, int);

    // See base class for documentation.
    long fpathconf(int, int);

    // See base class for documentation.
    int xmknod(int ver, const char *path, mode_t mode, dev_t dev);

    // See base class for documentation.
    int fcntl(int, int, void *);

    // See base class for documentation.
    int fcntl64(int, int, void *);

    // See base class for documentation.
    int dup(int);

    // See base class for documentation.
    int dup2(int, int);

    // See base class for documentation.
    int symlink(const char *, const char *);

    // See base class for documentation.
    int chdir(const char *);

    // See base class for documentation.
    int fchdir(int);

    // See base class for documentation.
    int rmdir(const char *);

    // See base class for documentation.
    int utime(const char *, utimbuf *);

    // See base class for documentation.
    int lutime(const char *, utimbuf *);

    // See base class for documentation.
    int futime(int, utimbuf *);

    // See base class for documentation.
    int execve(const char *, char *const *, char *const *);

    // See base class for documentation.
    int fexecve(int, char *const *, char *const *);

    // See base class for documentation.
    int access(const char *, int);

    // See base class for documentation.
    int statfs(const char *, struct statfs *);

    // See base class for documentation.
    int fstatfs(int, struct statfs *);

    // See base class for documentation.
    int statfs64(const char *, struct statfs64 *);

    // See base class for documentation.
    int fstatfs64(int, struct statfs64 *);

    // See base class for documentation.
    int truncate(const char *, off_t);

    // See base class for documentation.
    int truncate64(const char *, off64_t);

    // See base class for documentation.
    int ftruncate(int, off_t);

    // See base class for documentation.
    int ftruncate64(int, off64_t);

protected:
    /**
      * The get_deeper_cwd method is used to get the current working
      * directory of the deeper file system.  This is a helper function
      * needed by many of the filters.
      *
      * Normally, the individual filters would use inherited::some_method
      * to access the method of a deeper filter, but in this case getcwd
      * is not always filtered.
      */
    char *get_deeper_cwd(char *, long);

    /**
      * The get_deeper method is used to get the file system.
      * Ideally, you should not use this method.
      */
    plasticfs *get_deeper() { return deeper; }

    /**
      * The delete_deeper method is used to delete the 'deeper'
      * file system.  Ideally, you should not use this method.
      * Only the plasticfs_filter_module destructor may call this method.
      */
    void delete_deeper() { if (deeper) { delete deeper; deeper = 0; } }

    /**
      * The cwd method is used to get the current working directory.
      * This method uses the get_deeper_cwd() method to determine the
      * value, and forms an rcstring from it.
      */
    virtual rcstring cwd();

    /**
      * The make_absolute method is used to add the current working
      * directory (as returned by the cwd() method) to the start of
      * relative paths.  It leaves absolute paths alone.
      */
    virtual rcstring make_absolute(const rcstring &);

    /**
      * The path_join method is used to join two paths together, taking
      * care of various boundary conditions.
      */
    static rcstring path_join(const rcstring &, const rcstring &);

    /**
      * The is_below_mount_point method is used to determine if a path
      * being resolved by relative_realpath is below the filter's mount
      * point.
      */
    virtual bool is_below_mount_point(const char *path, size_t &pathlen) = 0;

    /**
      * The is_below_directory method is used to determine if a
      * path being resolved by relative_realpath is below the filter's
      * mount point.  This is a helper method for use by derived classes.
      *
      * The mount_point argument is assumed to be an absolute path
      * (i.e. that is starts with a slash (/) character).
      */
    static bool is_below_directory(const rcstring &mount_point,
	const char *path, size_t &pathlen);

    /**
      * The relative_realpath method is used to resolve te symlinks
      * in a path.  It differs from other realpath methods in that it
      * returns both the absolute path and the relative path.
      *
      * \returns
      *     Returns true of success, and false (setting errno) on failure.
      * \param path
      *     The path being resolved.
      * \param abspath
      *     The absolute path of the resolved path.
      * \param relpath
      *     The relative path of the resolved path, if it is below the
      *     mount point.  Otherwise is ts the same as the absolure path.
      * \param file_must_exist
      *     Return errors if the file does not exist if true.
      * \param do_last_element
      *     If true, process last element.  If false, do not check the
      *     last element to see if it is a symlink.
      */
    bool relative_realpath(const rcstring &path, rcstring &abspath,
	rcstring &relpath, bool file_must_exist, bool do_last_element);

    /**
      * The relative_readlink is just like readlink, but it takes a
      * relative path name and puts a filter-specific spin on the semantics.
      */
    virtual int relative_readlink(const char *path, char *buffer,
	size_t nbytes) = 0;

    /**
      * The relative_lstat is just like lstat, but it takes a relative
      * path name and puts a filter-specific spin on the semantics.
      */
    virtual int relative_lstat(const char *, struct stat *) = 0;

    /**
      * The relative_name_mapping is used to map a filename from the
      * interface perception, to the reality of the deeper filesystem.
      * A return value of false indicates that nothing need be done.
      * A return value of true indicates that result has been set to
      * the remapped name of the last element of the given path.
      * The default implementation always returns false.
      */
    virtual bool relative_name_mapping(const char *path, rcstring &result);

private:
    /**
      * The cwd_cache instance variable is used to remember tha last
      * answer retruned by the cwd() method, until invalidated by the
      * chdir() method.
      */
    rcstring cwd_cache;

    /**
      * The deeper instance variable is used to remember the file system
      * being filtered,
      */
    plasticfs *deeper;

    /**
      * The default constructor.  Do not use.
      */
    plasticfs_filter();

    /**
      * The copy constructor.  Do not use.
      */
    plasticfs_filter(const plasticfs_filter &);

    /**
      * The assignment operator.  Do not use.
      */
    plasticfs_filter &operator=(const plasticfs_filter &);
};

} // namespace libplasticfs

#endif // LIB_PLASTICFS_FILTER_H
