//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_chroot class
//

#pragma implementation "libplasticfs::plasticfs_filter_chroot"

#include <plasticfs/filter/chroot.h>
#include <cstring>
#include <fcntl.h>

//
// Define the CHROOT_DEBUG symbol when you need to see what methods
// are being called.
//
// #define CHROOT_DEBUG

using namespace libplasticfs;


plasticfs_filter_chroot::~plasticfs_filter_chroot()
{
}


plasticfs_filter_chroot::plasticfs_filter_chroot(plasticfs *deeper,
       	const rcstring_list &args) :
    plasticfs_filter(deeper)
{
    if (args.size() != 2)
    {
	error_message
	(
	    "plastivfs: chroot filter: execatly one argument required"
	);
    }
    if (args.size() < 2)
	root = "/";
    else
	root = args[1];
}


plasticfs *
plasticfs_filter_chroot::factory(plasticfs *deeper, const rcstring_list &args)
{
    return new plasticfs_filter_chroot(deeper, args);
}


rcstring
plasticfs_filter_chroot::cwd()
{
    rcstring tmp = inherited::cwd();
    if (root == rcstring("/"))
	return tmp;
    if (root == tmp)
	return rcstring("/");
    if
    (
	root.length() < tmp.length()
    &&
	0 == memcmp(tmp.c_str(), root.c_str(), root.length())
    &&
	tmp[root.length()] == '/'
    )
    {
	return
	    rcstring(tmp.c_str() + root.length(), tmp.length() - root.length());
    }
    return tmp;
}


bool
plasticfs_filter_chroot::is_below_mount_point(const char *, size_t &)
{
    return true;
}


int
plasticfs_filter_chroot::relative_readlink(const char *path, char *buffer,
    size_t nbytes)
{
    rcstring where(path_join(root, rcstring(path)));
    return inherited::readlink(where.c_str(), buffer, nbytes);
}


int
plasticfs_filter_chroot::relative_lstat(const char *path, struct stat *st)
{
    rcstring where(path_join(root, rcstring(path)));
    return inherited::lxstat(_STAT_VER, where.c_str(), st);
}


int
plasticfs_filter_chroot::open(const char *path, int flags, int mode)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: open(\"%s\", %#o, %#o)\n", __FILE__, __LINE__, path,
	flags, mode);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    bool last_file_must_exist = ((flags & O_CREAT) == 0);
    if (!relative_realpath(path, abspath, relpath, last_file_must_exist, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::open(where.c_str(), flags, mode);
}


int
plasticfs_filter_chroot::readlink(const char *path, char *buf, size_t nbytes)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: readlink(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::readlink(where.c_str(), buf, nbytes);
}


int
plasticfs_filter_chroot::lxstat(int vers, const char *path, struct stat *st)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: lxstat(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, false))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::lxstat(vers, where.c_str(), st);
}


int
plasticfs_filter_chroot::lxstat64(int vers, const char *path,
    struct stat64 *st)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: lxstat64(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, false))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::lxstat64(vers, where.c_str(), st);
}


int
plasticfs_filter_chroot::xstat(int vers, const char *path, struct stat *st)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: xstat(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::xstat(vers, where.c_str(), st);
}


int
plasticfs_filter_chroot::xstat64(int vers, const char *path,
    struct stat64 *st)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: xstat64(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::xstat64(vers, where.c_str(), st);
}


int
plasticfs_filter_chroot::chmod(const char *path, mode_t mode)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: chmod(\"%s\", %#o)\n", __FILE__, __LINE__, path,
	mode);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::chmod(where.c_str(), mode);
}


int
plasticfs_filter_chroot::mkdir(const char *path, mode_t mode)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: mkdir(\"%s\", %#o)\n", __FILE__, __LINE__, path,
	mode);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::mkdir(where.c_str(), mode);
}


int
plasticfs_filter_chroot::mkfifo(const char *path, mode_t mode)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: mkfifo(\"%s\", %#o)\n", __FILE__, __LINE__, path,
	mode);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::mkfifo(where.c_str(), mode);
}


int
plasticfs_filter_chroot::link(const char *src_path, const char *dst_path)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: link(\"%s\", \"%s\")\n", __FILE__, __LINE__,
	src_path, dst_path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring src_abspath;
    rcstring src_relpath;
    if (!relative_realpath(src_path, src_abspath, src_relpath, true, true))
	return -1;
    rcstring dst_abspath;
    rcstring dst_relpath;
    if (!relative_realpath(dst_path, dst_abspath, dst_relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring src_where(path_join(root, src_relpath));
    rcstring dst_where(path_join(root, dst_relpath));
    return inherited::link(src_where.c_str(), dst_where.c_str());
}


int
plasticfs_filter_chroot::symlink(const char *text, const char *dst_path)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: symlink(\"%s\", \"%s\")\n", __FILE__, __LINE__, text,
	dst_path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring dst_abspath;
    rcstring dst_relpath;
    if (!relative_realpath(dst_path, dst_abspath, dst_relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring dst_where(path_join(root, dst_relpath));
    return inherited::symlink(text, dst_where.c_str());
}


int
plasticfs_filter_chroot::unlink(const char *path)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: unlink(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::unlink(where.c_str());
}


int
plasticfs_filter_chroot::rmdir(const char *path)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: rmdir(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::rmdir(where.c_str());
}


int
plasticfs_filter_chroot::rename(const char *old_path, const char *new_path)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: rename(\"%s\", \"%s\")\n", __FILE__, __LINE__,
	old_path, new_path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring old_abspath;
    rcstring old_relpath;
    if (!relative_realpath(old_path, old_abspath, old_relpath, true, true))
	return -1;
    rcstring new_abspath;
    rcstring new_relpath;
    if (!relative_realpath(new_path, new_abspath, new_relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring old_where(path_join(root, old_relpath));
    rcstring new_where(path_join(root, new_relpath));
    return inherited::rename(old_where.c_str(), new_where.c_str());
}


long
plasticfs_filter_chroot::pathconf(const char *path, int name)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: pathconf(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::pathconf(where.c_str(), name);
}


int
plasticfs_filter_chroot::xmknod(int ver, const char *path, mode_t mode,
    dev_t dev)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: xmknod(\"%s\", %#o, %#x)\n", __FILE__, __LINE__,
	path, mode, (int)dev);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::xmknod(ver, where.c_str(), mode, dev);
}


int
plasticfs_filter_chroot::chdir(const char *path)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: chdir(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::chdir(where.c_str());
}


int
plasticfs_filter_chroot::utime(const char *path, utimbuf *buf)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: utime(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::utime(where.c_str(), buf);
}


int
plasticfs_filter_chroot::lutime(const char *path, utimbuf *buf)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: lutime(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::lutime(where.c_str(), buf);
}


int
plasticfs_filter_chroot::execve(const char *path, char *const *argv,
    char *const *envp)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: execve(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::execve(where.c_str(), argv, envp);
}


int
plasticfs_filter_chroot::access(const char *path, int type)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: access(\"%s\")\n", __FILE__, __LINE__, path);
#endif
    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::access(where.c_str(), type);
}


int
plasticfs_filter_chroot::statfs(const char *path, struct statfs *buf)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: statfs(\"%s\")\n", __FILE__, __LINE__, path);
#endif

    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::statfs(where.c_str(), buf);
}


int
plasticfs_filter_chroot::statfs64(const char *path, struct statfs64 *buf)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: statfs64(\"%s\")\n", __FILE__, __LINE__, path);
#endif

    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::statfs64(where.c_str(), buf);
}


int
plasticfs_filter_chroot::truncate(const char *path, off_t length)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: truncate(\"%s\", %ld)\n", __FILE__, __LINE__, path,
	(long)length);
#endif

    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::truncate(where.c_str(), length);
}


int
plasticfs_filter_chroot::truncate64(const char *path, off64_t length)
{
#ifdef CHROOT_DEBUG
    error_message("%s: %d: truncate64(\"%s\", %ld)\n", __FILE__, __LINE__, path,
	(long)length);
#endif

    //
    // Resolve all of the symlinks in the path.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Add the root to the relative path and pass it on.
    //
    rcstring where(path_join(root, relpath));
    return inherited::truncate64(where.c_str(), length);
}
