//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_log class
//

#pragma implementation "libplasticfs::plasticfs_filter_log"

#include <plasticfs/filter/log.h>
#include <cstdarg>
#include <cstdio>
#include <cstring>
#include <errno_insulator.h>
#include <fcntl.h>
#include <utils.h>
#include <unistd.h>
#include <sys/vfs.h>

using namespace libplasticfs;


plasticfs_filter_log::~plasticfs_filter_log()
{
    if (fd >= 0)
    {
	inherited::close(fd);
	fd = -1;
    }
}


plasticfs_filter_log::plasticfs_filter_log(plasticfs *a_deeper,
	const rcstring_list &args) :
    plasticfs_filter(a_deeper),
    fd(-1)
{
    if (args.size() != 2)
    {
	error_message
	(
	    "plasticfs: the log filter requires exactly one argument\n"
	);
    }
    if (args.size() < 2)
	return;
    filename = args[1];
}


plasticfs *
plasticfs_filter_log::factory(plasticfs *deeper, const rcstring_list &args)
{
    return new plasticfs_filter_log(deeper, args);
}


bool
plasticfs_filter_log::is_below_mount_point(const char *, size_t &)
{
    // This is called by relative_realpath, available only to derived
    // classes of plasticfs_filter.  And since we (the derived class in
    // question) never call relative_realpath is isn't relevant what we
    // return here.
    return false;
}


int 
plasticfs_filter_log::relative_readlink(const char *, char *, size_t)
{
    // because is_below_mount_point is always false, this method never called.
    return -1;
}


int 
plasticfs_filter_log::relative_lstat(const char *, struct stat *)
{
    // because is_below_mount_point is always false, this method never called.
    return -1;
}


void
plasticfs_filter_log::deferred_open()
{
    if (fd < 0 && filename.length())
    {
	errno_insulator guard;
	int flags = O_WRONLY | O_CREAT | O_TRUNC | O_APPEND;
	fd = inherited::open(filename.c_str(), flags, 0666);
	if (fd < 0)
	{
	    errno_insulator tmp;
	    error_message
	    (
		"plasticfs: log filter: open \"%s\" failed: %s\n",
		filename.c_str(),
		tmp.gets()
	    );

	    // don't try again
	    filename = "";
	}

	// Try to set the close-on-exec flag, ignore an error.
	inherited::fcntl(fd, F_SETFD, (void *)(long)FD_CLOEXEC);
    }
}

void
plasticfs_filter_log::stash(const char *fmt, ...)
{
    //
    // Build the string to be logged.
    //
    va_list ap;
    va_start(ap, fmt);
    char buffer[2000];
    vsnprintf(buffer, sizeof(buffer), fmt, ap);
    va_end(ap);

    //
    // Write the string to the log.
    //
    if (fd < 0)
	deferred_open();
    if (fd >= 0)
	inherited::write(fd, buffer, strlen(buffer));
}


static const char *
open_flags_representation(int x)
{
    static char buffer[150];
    switch (x & O_ACCMODE)
    {
    case O_RDONLY:
	strlcpy(buffer, "O_RDONLY", sizeof(buffer));
	break;

    case O_WRONLY:
	strlcpy(buffer, "O_WRONLY", sizeof(buffer));
	break;

    case O_RDWR:
	strlcpy(buffer, "O_RDWR", sizeof(buffer));
	break;

    default:
	snprintf(buffer, sizeof(buffer), "%#X", x & O_ACCMODE);
	break;
    }
    x &= ~O_ACCMODE;

    int garbage = 0;
    while (x)
    {
	int bit = x & -x;
	x -= bit;
	char *s = 0;
	switch (bit)
	{
	case O_CREAT:     s = "+O_CREAT";     break;
	case O_EXCL:      s = "+O_EXCL";      break;
	case O_NOCTTY:    s = "+O_NOCTTY";    break;
	case O_TRUNC:     s = "+O_TRUNC";     break;
	case O_APPEND:    s = "+O_APPEND";    break;
	case O_NONBLOCK:  s = "+O_NONBLOCK";  break;
	case O_SYNC:      s = "+O_SYNC";      break;
	case O_ASYNC:     s = "+O_ASYNC";     break;
#if defined(O_DIRECT) && (O_DIRECT != O_SYNC)
	case O_DIRECT:    s = "+O_DIRECT";    break;
#endif
	case O_DIRECTORY: s = "+O_DIRECTORY"; break;
	case O_NOFOLLOW:  s = "+O_NOFOLLOW";  break;
	case O_LARGEFILE: s = "+O_LARGEFILE"; break;

	default:
	    garbage += bit;
	    continue;
	}
	strlcat(buffer, s, sizeof(buffer));
    }
    if (garbage)
    {
	char tmp[20];
	snprintf(tmp, sizeof(tmp), "+%#x", garbage);
	strlcat(buffer, tmp, sizeof(buffer));
    }
    return buffer;
}


int
plasticfs_filter_log::open(const char *path, int flags, int mode)
{
    int filedesc = inherited::open(path, flags, mode);
    errno_insulator guard;
    if (filedesc < 0)
    {
	stash
	(
	    "open(path = \"%s\", flags = %s, mode = %#o) = %d; errno = %d %s\n",
	    path,
	    open_flags_representation(flags),
	    mode,
	    filedesc,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "open(path = \"%s\", flags = %s, mode = %#o) = %d\n",
	    path,
	    open_flags_representation(flags),
	    mode,
	    filedesc
	);
    }
    return filedesc;
}


int
plasticfs_filter_log::close(int filedesc)
{
    int result = inherited::close(filedesc);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "close(filedesc = %d) = %d; errno = %d %s\n",
	    filedesc,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("close(filedesc = %d) = %d\n", filedesc, result);
    }
    return result;
}


ssize_t
plasticfs_filter_log::read(int filedesc, void *data, size_t nbytes)
{
    ssize_t result = inherited::read(filedesc, data, nbytes);
    errno_insulator guard;
    if (result == -1)
    {
	stash
	(
	    "read(filedesc = %d, data = %p, nbytes = %lu) "
		"= %ld; errno = %d %s\n",
	    filedesc,
	    data,
	    (unsigned long)nbytes,
	    (long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "read(filedesc = %d, data = %p, nbytes = %lu) = %ld\n",
	    filedesc,
	    data,
	    (unsigned long)nbytes,
	    (long)result
	);
    }
    return result;
}


ssize_t
plasticfs_filter_log::write(int filedesc, const void *data, size_t nbytes)
{
    ssize_t result = inherited::write(filedesc, data, nbytes);
    errno_insulator guard;
    if (result == -1)
    {
	stash
	(
	    "write(filedesc = %d, data = %p, nbytes = %lu) "
		"= %ld; errno = %d %s\n",
	    filedesc,
	    data,
	    (unsigned long)nbytes,
	    (long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "write(filedesc = %d, data = %p, nbytes = %lu) = %ld\n",
	    filedesc,
	    data,
	    (unsigned long)nbytes,
	    (long)result
	);
    }
    return result;
}


int
plasticfs_filter_log::readlink(const char *path, char *buffer, size_t nbytes)
{
    int result = inherited::readlink(path, buffer, nbytes);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "readlink(path = \"%s\", buffer = %p, nbytes = %lu) "
		"= %d; errno = %d %s\n",
	    path,
	    buffer,
	    (unsigned long)nbytes,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "readlink(path = \"%s\", buffer = %p \"%.*s\", nbytes = %lu) "
		"= %d\n",
	    path,
	    buffer,
	    result,
	    buffer,
	    (unsigned long)nbytes,
	    result
	);
    }
    return result;
}


static const char *
stat_mode_representation(mode_t x)
{
    char *s = 0;
    switch (x & S_IFMT)
    {
    case S_IFDIR: s = "S_IFDIR"; break;
    case S_IFCHR: s = "S_IFCHR"; break;
    case S_IFBLK: s = "S_IFBLK"; break;
    case S_IFREG: s = "S_IFREG"; break;
    case S_IFIFO: s = "S_IFIFO"; break;
    case S_IFLNK: s = "S_IFLNK"; break;
    case S_IFSOCK: s = "S_IFSOCK"; break;
    }
    static char buffer[50];
    if (s)
	snprintf(buffer, sizeof(buffer), "%s+%#o", s, (int)(x & ~S_IFMT));
    else
	snprintf(buffer, sizeof(buffer), "%#o", (int)x);
    return buffer;
}


static const char *
struct_stat_representation(const struct stat *st)
{
    if (!st)
	return "NULL";
    static char buffer[1000];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ st_dev = %#x, st_ino = %d, st_mode = %s, st_nlink = %d, "
	    "st_uid = %d, st_gid = %d, st_rdev = %#x, st_size = %d, "
	    "st_blksize = %d, st_blocks = %d, st_atime = %d, "
	    "st_mtime = %d, st_ctime = %d }",
	(int)st->st_dev,
	(int)st->st_ino,
	stat_mode_representation(st->st_mode),
	(int)st->st_nlink,
	(int)st->st_uid,
	(int)st->st_gid,
	(int)st->st_rdev,
	(int)st->st_size,
	(int)st->st_blksize,
	(int)st->st_blocks,
	(int)st->st_atime,
	(int)st->st_mtime,
	(int)st->st_ctime
    );
    return buffer;
}


int
plasticfs_filter_log::xstat(int version, const char *path, struct stat *statbuf)
{
    int result = inherited::xstat(version, path, statbuf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "stat(path = \"%s\", statbuf = %p) = %d; errno = %d %s\n",
	    path,
	    statbuf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "stat(path = \"%s\", statbuf = %s) = %d\n",
	    path,
	    struct_stat_representation(statbuf),
	    result
	);
    }
    return result;
}


static const char *
struct_stat64_representation(const struct stat64 *st)
{
    if (!st)
	return "NULL";
    static char buffer[1000];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ st_dev = %#x, st_ino = %d, st_mode = %s, st_nlink = %d, "
	    "st_uid = %d, st_gid = %d, st_rdev = %#x, st_size = %d, "
	    "st_blksize = %d, st_blocks = %d, st_atime = %d, "
	    "st_mtime = %d, st_ctime = %d }",
	(int)st->st_dev,
	(int)st->st_ino,
	stat_mode_representation(st->st_mode),
	(int)st->st_nlink,
	(int)st->st_uid,
	(int)st->st_gid,
	(int)st->st_rdev,
	(int)st->st_size,
	(int)st->st_blksize,
	(int)st->st_blocks,
	(int)st->st_atime,
	(int)st->st_mtime,
	(int)st->st_ctime
    );
    return buffer;
}


int
plasticfs_filter_log::xstat64(int version, const char *path, struct stat64 *buf)
{
    int result = inherited::xstat64(version, path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "stat64(path = \"%s\", statbuf = %p) = %d; errno = %d %s\n",
	    path,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "stat64(path = \"%s\", statbuf = %s) = %d\n",
	    path,
	    struct_stat64_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::lxstat(int version, const char *path, struct stat *buf)
{
    int result = inherited::lxstat(version, path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "lstat(path = \"%s\", statbuf = %p) = %d; errno = %d %s\n",
	    path,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "lstat(path = \"%s\", statbuf = %s) = %d\n",
	    path,
	    struct_stat_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::lxstat64(int vers, const char *path, struct stat64 *buf)
{
    int result = inherited::lxstat64(vers, path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "lstat64(path = \"%s\", statbuf = %p) = %d; errno = %d %s\n",
	    path,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "lstat64(path = \"%s\", statbuf = %s) = %d\n",
	    path,
	    struct_stat64_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::fxstat(int version, int filedesc, struct stat *buf)
{
    int result = inherited::fxstat(version, filedesc, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fstat(filedesc = %d, statbuf = %p) = %d; errno = %d %s\n",
	    filedesc,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "fstat(filedesc = %d, statbuf = %s) = %d\n",
	    filedesc,
	    struct_stat_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::fxstat64(int vers, int filedesc, struct stat64 *buf)
{
    int result = inherited::fxstat64(vers, filedesc, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fstat64(filedesc = %d, statbuf = %p) = %d; errno = %d %s\n",
	    filedesc,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "fstat64(filedesc = %d, statbuf = %s) = %d\n",
	    filedesc,
	    struct_stat64_representation(buf),
	    result
	);
    }
    return result;
}


static char *
whence_representation(int x)
{
    switch (x)
    {
    case SEEK_SET: return "SEEK_SET";
    case SEEK_CUR: return "SEEK_CUR";
    case SEEK_END: return "SEEK_END";
    }
    static char buffer[20];
    snprintf(buffer, sizeof(buffer), "%d", x);
    return buffer;
}


off_t
plasticfs_filter_log::lseek(int fd, off_t offset, int whence)
{
    off_t result = inherited::lseek(fd, offset, whence);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "lseek(filedesc = %d, offset = %ld, whence = %s) "
		"= %ld; errno = %d %s\n",
	    fd,
	    (long)offset,
	    whence_representation(whence),
	    (long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "lseek(filedesc = %d, offset = %ld, whence = %s) = %ld\n",
	    fd,
	    (long)offset,
	    whence_representation(whence),
	    (long)result
	);
    }
    return result;
}


off64_t
plasticfs_filter_log::lseek64(int fd, off64_t offset, int whence)
{
    off64_t result = inherited::lseek64(fd, offset, whence);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "lseek64(filedesc = %d, offset = %lld, whence = %s) "
		"= %lld; errno = %d %s\n",
	    fd,
	    (long long)offset,
	    whence_representation(whence),
	    (long long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "lseek64(filedesc = %d, offset = %lld, whence = %s) = %lld\n",
	    fd,
	    (long long)offset,
	    whence_representation(whence),
	    (long long)result
	);
    }
    return result;
}


#ifdef FILTER_OPENDIR


DIR *
plasticfs_filter_log::opendir(const char *path)
{
    DIR *result = inherited::opendir(path);
    errno_insulator guard;
    if (!result)
    {
	stash
	(
	    "opendir(path = \"%s\") = %p; errno = %d %s\n",
	    path,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "opendir(path = \"%s\") = %p\n",
	    path,
	    result
	);
    }
    return result;
}


int 
plasticfs_filter_log::closedir(DIR *dirp)
{
    int result = inherited::closedir(dirp);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "closedir(dirp = %p) = %d; errno = %d %s\n",
	    dirp,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "closedir(dirp = %p) = %d\n",
	    dirp,
	    result
	);
    }
    return result;
}


dirent *
plasticfs_filter_log::readdir(DIR *dirp)
{
    dirent *result = inherited::readdir(dirp);
    errno_insulator guard;
    if (!result)
    {
	stash
	(
	    "readdir(dirp = %p) = %p; errno = %d %s\n",
	    dirp,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "readdir(dirp = %p) = %p { d_name = \"%s\" }\n",
	    dirp,
	    result,
	    result->d_name
	);
    }
    return result;
}


dirent64 *
plasticfs_filter_log::readdir64(DIR *dirp)
{
    dirent64 *result = inherited::readdir64(dirp);
    errno_insulator guard;
    if (!result)
    {
	stash
	(
	    "readdir64(dirp = %p) = %p; errno = %d %s\n",
	    dirp,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "readdir64(dirp = %p) = %p { d_name = \"%s\" }\n",
	    dirp,
	    result,
	    result->d_name
	);
    }
    return result;
}


void 
plasticfs_filter_log::rewinddir(DIR *dirp)
{
    inherited::rewinddir(dirp);
    errno_insulator guard;
    stash("rewinddir(dirp = %p)\n", dirp);
}


void 
plasticfs_filter_log::seekdir(DIR *dirp, off_t offset)
{
    inherited::seekdir(dirp, offset);
    errno_insulator guard;
    stash("seekdir(dirp = %p, offset = %ld)\n", dirp, (long)offset);
}


off_t 
plasticfs_filter_log::telldir(DIR *dirp)
{
    off_t result = inherited::telldir(dirp);
    errno_insulator guard;
    if (result == -1)
    {
	stash
	(
	    "telldir(dirp = %p) = %ld; errno = %d %s\n",
	    dirp,
	    (long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "telldir(dirp = %p) = %ld\n",
	    dirp,
	    (long)result
	);
    }
    return result;
}


int 
plasticfs_filter_log::dirfd(DIR *dirp)
{
    int result = inherited::dirfd(dirp);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "dirfd(dirp = %p) = %d; errno = %d %s\n",
	    dirp,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "dirfd(dirp = %p) = %d\n",
	    dirp,
	    result
	);
    }
    return result;
}


char *
plasticfs_filter_log::getcwd(char *buf, long nbytes)
{
    char *result = inherited::getcwd(buf, nbytes);
    errno_insulator guard;
    if (!result)
    {
	stash
	(
	    "getcwd(buf = %p, size = %ld) = %p; errno = %d %s\n",
	    buf,
	    nbytes,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "getcwd(buf = %p, size = %ld) = %p \"%s\"\n",
	    buf,
	    nbytes,
	    result,
	    result
	);
    }
    return result;
}


#endif // FILTER_OPENDIR


int 
plasticfs_filter_log::getdents(int fd, dirent *buf, unsigned len)
{
    int result = inherited::getdents(fd, buf, len);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "getdents(fd = %d, buf = %p, nbytes = %u) = %d; errno = %d %s\n",
	    fd,
	    buf,
	    len,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "getdents(fd = %d, buf = %p, nbytes = %u) = %d\n",
	    fd,
	    buf,
	    len,
	    result
	);
    }
    return result;
}


int 
plasticfs_filter_log::getdents64(int fd, dirent64 *buf, unsigned len)
{
    int result = inherited::getdents64(fd, buf, len);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "getdents64(fd = %d, buf = %p, nbytes = %u) = %d; errno = %d %s\n",
	    fd,
	    buf,
	    len,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "getdents64(fd = %d, buf = %p, nbytes = %u) = %ld\n",
	    fd,
	    buf,
	    len,
	    (long)result
	);
    }
    return result;
}


ssize_t 
plasticfs_filter_log::getdirentries(int fd, char *buf, size_t len, off_t *base)
{
    ssize_t result = inherited::getdirentries(fd, buf, len, base);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "getdirentries(fd = %d, buf = %p, nbytes = %ld, basep = %p) "
		"= %ld; errno = %d %s\n",
	    fd,
	    buf,
	    (long)len,
	    base,
	    (long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "getdirentries(fd = %d, buf = %p, nbytes = %ld, basep = %p) "
		"= %ld\n",
	    fd,
	    buf,
	    (long)len,
	    base,
	    (long)result
	);
    }
    return result;
}


ssize_t 
plasticfs_filter_log::getdirentries64(int fd, char *buf, size_t len,
    off64_t *base)
{
    ssize_t result = inherited::getdirentries64(fd, buf, len, base);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "getdirentries64(fd = %d, buf = %p, nbytes = %ld, basep = %p) "
		"= %ld; errno = %d %s\n",
	    fd,
	    buf,
	    (long)len,
	    base,
	    (long)result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "getdirentries64(fd = %d, buf = %p, nbytes = %ld, basep = %p) "
		"= %ld\n",
	    fd,
	    buf,
	    (long)len,
	    base,
	    (long)result
	);
    }
    return result;
}


int
plasticfs_filter_log::chmod(const char *path, mode_t mode)
{
    int result = inherited::chmod(path, mode);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "chmod(path = \"%s\", mode = %#o) = %d; errno = %d %s\n",
	    path,
	    mode,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("chmod(path = \"%s\", mode = %#o) = %d\n", path, mode, result);
    }
    return result;
}


int
plasticfs_filter_log::fchmod(int fd, mode_t mode)
{
    int result = inherited::fchmod(fd, mode);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fchmod(fd = %d, mode = %#o) = %d; errno = %d %s\n",
	    fd,
	    mode,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("fchmod(fd = %d, mode = %#o) = %d\n", fd, mode, result);
    }
    return result;
}


int
plasticfs_filter_log::mkdir(const char *path, mode_t mode)
{
    int result = inherited::mkdir(path, mode);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "mkdir(path = \"%s\", mode = %#o) = %d; errno = %d %s\n",
	    path,
	    mode,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("mkdir(path = \"%s\", mode = %#o) = %d\n", path, mode, result);
    }
    return result;
}


int
plasticfs_filter_log::mkfifo(const char *path, mode_t mode)
{
    int result = inherited::mkfifo(path, mode);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "mkfifo(path = \"%s\", mode = %#o) = %d; errno = %d %s\n",
	    path,
	    mode,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("mkfifo(path = \"%s\", mode = %#o) = %d\n", path, mode, result);
    }
    return result;
}


int
plasticfs_filter_log::unlink(const char *path)
{
    int result = inherited::unlink(path);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "unlink(path = \"%s\") = %d; errno = %d %s\n",
	    path,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("unlink(path = \"%s\") = %d\n", path, result);
    }
    return result;
}


int
plasticfs_filter_log::rmdir(const char *path)
{
    int result = inherited::rmdir(path);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "rmdir(path = \"%s\") = %d; errno = %d %s\n",
	    path,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("rmdir(path = \"%s\") = %d\n", path, result);
    }
    return result;
}


int
plasticfs_filter_log::rename(const char *oldpath, const char *newpath)
{
    int result = inherited::rename(oldpath, newpath);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "rename(oldpath = \"%s\", newpath = \"%s\") = %d; errno = %d %s\n",
	    oldpath,
	    newpath,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "rename(oldpath = \"%s\", newpath = \"%s\") = %d\n",
	    oldpath,
	    newpath,
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::link(const char *oldpath, const char *newpath)
{
    int result = inherited::link(oldpath, newpath);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "link(oldpath = \"%s\", newpath = \"%s\") = %d; errno = %d %s\n",
	    oldpath,
	    newpath,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "link(oldpath = \"%s\", newpath = \"%s\") = %d\n",
	    oldpath,
	    newpath,
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::symlink(const char *oldpath, const char *newpath)
{
    int result = inherited::symlink(oldpath, newpath);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "symlink(oldpath = \"%s\", newpath = \"%s\") = %d; errno = %d %s\n",
	    oldpath,
	    newpath,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "symlink(oldpath = \"%s\", newpath = \"%s\") = %d\n",
	    oldpath,
	    newpath,
	    result
	);
    }
    return result;
}


static char *
pathconf_name_representation(int x)
{
    switch (x)
    {
    case _PC_LINK_MAX:
	return "_PC_LINK_MAX";

    case _PC_MAX_CANON:
	return "_PC_MAX_CANON";

    case _PC_MAX_INPUT:
	return "_PC_MAX_INPUT";

    case _PC_NAME_MAX:
	return "_PC_NAME_MAX";

    case _PC_PATH_MAX:
	return "_PC_PATH_MAX";

    case _PC_PIPE_BUF:
	return "_PC_PIPE_BUF";

    case _PC_CHOWN_RESTRICTED:
	return "_PC_CHOWN_RESTRICTED";

    case _PC_NO_TRUNC:
	return "_PC_NO_TRUNC";

    case _PC_VDISABLE:
	return "_PC_VDISABLE";
    }
    static char buffer[20];
    snprintf(buffer, sizeof(buffer), "%d", x);
    return buffer;
}


long
plasticfs_filter_log::pathconf(const char *path, int name)
{
    long result = inherited::pathconf(path, name);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "pathconf(path = \"%s\", name = %s) = %ld; errno = %d %s\n",
	    path,
	    pathconf_name_representation(name),
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "pathconf(path = \"%s\", name = %s) = %ld\n",
	    path,
	    pathconf_name_representation(name),
	    result
	);
    }
    return result;
}


long
plasticfs_filter_log::fpathconf(int fd, int name)
{
    long result = inherited::fpathconf(fd, name);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fpathconf(filedesc = %d, name = %s) = %ld; errno = %d %s\n",
	    fd,
	    pathconf_name_representation(name),
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "pathconf(filedesc = %d, name = %s) = %ld\n",
	    fd,
	    pathconf_name_representation(name),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::xmknod(int ver, const char *path, mode_t mode, dev_t dev)
{
    int result = inherited::xmknod(ver, path, mode, dev);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "mknod(path = \"%s\", mode = %s, dev = %#x) = %d; errno = %d %s\n",
	    path,
	    stat_mode_representation(mode),
	    (int)dev,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "mknod(path = \"%s\", mode = %s, dev = %#x) = %d\n",
	    path,
	    stat_mode_representation(mode),
	    (int)dev,
	    result
	);
    }
    return result;
}


static char *
fcntl_cmd_representation(int x)
{
    switch (x)
    {
    case F_DUPFD:
	return "F_DUPFD";

    case F_GETFD:
	return "F_GETFD";

    case F_SETFD:
	return "F_SETFD";

    case F_GETFL:
	return "F_GETFL";

    case F_SETFL:
	return "F_SETFL";

    case F_GETLK:
	return "F_GETLK";

    case F_SETLK:
	return "F_SETLK";

    case F_SETLKW:
	return "F_SETLKW";

#if defined(F_GETLK64) && (F_GETLK64 != F_GETLK)
    case F_GETLK64:
	return "F_GETLK64";

    case F_SETLK64:
	return "F_SETLK64";

    case F_SETLKW64:
	return "F_SETLKW64";
#endif

    case F_GETOWN:
	return "F_GETOWN";

    case F_SETOWN:
	return "F_SETOWN";

    case F_GETSIG:
	return "F_GETSIG";

    case F_SETSIG:
	return "F_SETSIG";
    }
    static char buffer[20];
    snprintf(buffer, sizeof(buffer), "%d", x);
    return buffer;
}


static char *
flock_type_representation(int x)
{
    switch (x)
    {
    case F_RDLCK:
	return "F_RDLCK";

    case F_WRLCK:
	return "F_WRLCK";

    case F_UNLCK:
	return "F_UNLCK";

    case F_EXLCK:
	return "F_EXLCK";

    case F_SHLCK:
       return "F_SHLCK";
    }
    static char buffer[20];
    snprintf(buffer, sizeof(buffer), "%d", x);
    return buffer;
}


static char *
flock_representation(flock *p)
{
    if (!p)
	return "NULL";
    static char buffer[100];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ l_type = %s, l_whence = %s, l_start = %ld, l_len = %lu, "
	    "l_pid = %d }",
	flock_type_representation(p->l_type),
	whence_representation(p->l_whence),
	(long)p->l_start,
	(unsigned long)p->l_len,
	(int)p->l_pid
    );
    return buffer;
}


#if defined(F_GETLK64) && (F_GETLK64 != F_GETLK)

static char *
flock64_representation(flock64 *p)
{
    static char buffer[100];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ l_type = %s, l_whence = %s, l_start = %ld, l_len = %lu, "
	    "l_pid = %d }",
	flock_type_representation(p->l_type),
	whence_representation(p->l_whence),
	(long)p->l_start,
	(unsigned long)p->l_len,
	(int)p->l_pid
    );
    return buffer;
}

#endif


int
plasticfs_filter_log::fcntl(int filedesc, int cmd, void *arg)
{
    int result = inherited::fcntl(filedesc, cmd, arg);
    errno_insulator guard;
    switch (cmd)
    {
    default:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		result
	    );
	}
	break;

    case F_DUPFD:
    case F_SETFD:
    case F_SETOWN:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %ld) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		(long)arg,
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %ld) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		(long)arg,
		result
	    );
	}
	break;

    case F_GETFL:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s) = %s\n",
		fd,
		fcntl_cmd_representation(cmd),
		open_flags_representation(result)
	    );
	}
	break;

    case F_SETFL:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		open_flags_representation((long)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		open_flags_representation((long)arg),
		result
	    );
	}
	break;

    case F_GETLK:
    case F_SETLK:
    case F_SETLKW:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		flock_representation((flock *)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		flock_representation((flock *)arg),
		result
	    );
	}
	break;

#if defined(F_GETLK64) && (F_GETLK64 != F_GETLK)
    case F_GETLK64:
    case F_SETLK64:
    case F_SETLKW64:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		flock64_representation((flock64 *)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		flock64_representation((flock64 *)arg),
		result
	    );
	}
	break;
#endif

    case F_GETSIG:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s) = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd),
		result,
		strsignal(result)
	    );
	}
	break;

    case F_SETSIG:
	if (result < 0)
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %ld %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		(long)arg,
		strsignal((long)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl(fd = %d, cmd = %s, arg = %ld %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		(long)arg,
		strsignal((long)arg),
		result
	    );
	}
	break;
    }

    return result;
}


int
plasticfs_filter_log::fcntl64(int filedesc, int cmd, void *arg)
{
    int result = inherited::fcntl64(filedesc, cmd, arg);
    errno_insulator guard;
    switch (cmd)
    {
    default:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		result
	    );
	}
	break;

    case F_DUPFD:
    case F_SETFD:
    case F_SETOWN:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %ld) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		(long)arg,
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %ld) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		(long)arg,
		result
	    );
	}
	break;

    case F_GETFL:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s) = %s\n",
		fd,
		fcntl_cmd_representation(cmd),
		open_flags_representation(result)
	    );
	}
	break;

    case F_SETFL:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		open_flags_representation((long)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		open_flags_representation((long)arg),
		result
	    );
	}
	break;

    case F_GETLK:
    case F_SETLK:
    case F_SETLKW:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		flock_representation((flock *)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		flock_representation((flock *)arg),
		result
	    );
	}
	break;

#if defined(F_GETLK64) && (F_GETLK64 != F_GETLK)
    case F_GETLK64:
    case F_SETLK64:
    case F_SETLKW64:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		flock64_representation((flock64 *)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		flock64_representation((flock64 *)arg),
		result
	    );
	}
	break;
#endif

    case F_GETSIG:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s) = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd),
		result,
		strsignal(result)
	    );
	}
	break;

    case F_SETSIG:
	if (result < 0)
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %ld %s) = %d; errno = %d %s\n",
		fd,
		fcntl_cmd_representation(cmd), 
		(long)arg,
		strsignal((long)arg),
		result,
		guard.get(),
		guard.gets()
	    );
	}
	else
	{
	    stash
	    (
		"fcntl64(fd = %d, cmd = %s, arg = %ld %s) = %d\n",
		fd,
		fcntl_cmd_representation(cmd),
		(long)arg,
		strsignal((long)arg),
		result
	    );
	}
	break;
    }

    return result;
}


int
plasticfs_filter_log::dup(int oldfd)
{
    int result = inherited::dup(oldfd);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "dup(oldfd = %d) = %d; errno = %d %s\n",
	    oldfd,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("dup(oldfd = %d) = %d\n", oldfd, result);
    }
    return result;
}


int
plasticfs_filter_log::dup2(int oldfd, int newfd)
{
    int result = inherited::dup2(oldfd, newfd);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "dup2(oldfd = %d, newfd = %d) = %d; errno = %d %s\n",
	    oldfd,
	    newfd,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("dup2(oldfd = %d, newfd = %d) = %d\n", oldfd, newfd, result);
    }
    return result;
}


int
plasticfs_filter_log::chdir(const char *path)
{
    int result = inherited::chdir(path);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "chdir(path = \"%s\") = %d; errno = %d %s\n",
	    path,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("chdir(path = \"%s\") = %d\n", path, result);
    }
    return result;
}


int
plasticfs_filter_log::fchdir(int fd)
{
    int result = inherited::fchdir(fd);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fchdir(filedesc = %d) = %d; errno = %d %s\n",
	    fd,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash("fchdir(filedesc = %d) = %d\n", fd, result);
    }
    return result;
}


#include <utime.h>


static char *
utimbuf_representation(const utimbuf *p)
{
    if (!p)
	return "NULL";
    static char buffer[40];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ actime = %ld, modtime = %ld }",
	(long)p->actime,
	(long)p->modtime
    );
    return buffer;
}


int
plasticfs_filter_log::utime(const char *path, utimbuf *buf)
{
    int result = inherited::utime(path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "utime(path = \"%s\", buf = %s) = %d; errno = %d %s\n",
	    path,
	    utimbuf_representation(buf),
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "utime(path = \"%s\", buf = %s) = %d\n",
	    path,
	    utimbuf_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::lutime(const char *path, utimbuf *buf)
{
    int result = inherited::lutime(path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "lutime(path = \"%s\", buf = %s) = %d; errno = %d %s\n",
	    path,
	    utimbuf_representation(buf),
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "lutime(path = \"%s\", buf = %s) = %d\n",
	    path,
	    utimbuf_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::futime(int fd, utimbuf *buf)
{
    int result = inherited::futime(fd, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "futime(filedesc = %d, buf = %s) = %d; errno = %d %s\n",
	    fd,
	    utimbuf_representation(buf),
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "futime(filedesc = %d, buf = %s) = %d\n",
	    fd,
	    utimbuf_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::execve(const char *path, char *const *argv,
    char *const *envp)
{
    // We have to log the command before we do anything,
    // because it doesn't return on success.
    stash("execve(path = \"%s\", argv = %p, envp = %p)", path, argv, envp);
    int result = inherited::execve(path, argv, envp);
    errno_insulator guard;
    stash(" = %d; errno = %d %s\n", result, guard.get(), guard.gets());
    return result;
}


int
plasticfs_filter_log::fexecve(int fd, char *const *argv,
    char *const *envp)
{
    // We have to log the command before we do anything,
    // because it doesn't return on success.
    stash("execve(filedesc = %d, argv = %p, envp = %p)", fd, argv, envp);
    int result = inherited::fexecve(fd, argv, envp);
    errno_insulator guard;
    stash(" = %d; errno = %d %s\n", result, guard.get(), guard.gets());
    return result;
}


static char *
access_type_representation(int x)
{
    if (x == F_OK)
	return "F_OK";
    static char buffer[20];
    buffer[0] = 0;
    if (x & R_OK)
    {
	strlcpy(buffer, "R_OK", sizeof(buffer));
	x &= ~R_OK;
    }
    if (x & W_OK)
    {
	if (buffer[0])
	    strlcat(buffer, "|", sizeof(buffer));
	strlcat(buffer, "W_OK", sizeof(buffer));
	x &= ~W_OK;
    }
    if (x & X_OK)
    {
	if (buffer[0])
	    strlcat(buffer, "|", sizeof(buffer));
	strlcat(buffer, "X_OK", sizeof(buffer));
	x &= ~X_OK;
    }
    if (x)
    {
	char tmp[20];
	snprintf(tmp, sizeof(tmp), "%#o", x);
	if (buffer[0])
	    strlcat(buffer, "|", sizeof(buffer));
	strlcat(buffer, tmp, sizeof(buffer));
    }
    return buffer;
}


int
plasticfs_filter_log::access(const char *path, int type)
{
    int result = inherited::access(path, type);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "access(path = \"%s\", type = %s) = %d; errno = %d %s\n",
	    path,
	    access_type_representation(type),
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "access(path = \"%s\", type = %s) = %d\n",
	    path,
	    access_type_representation(type),
	    result
	);
    }
    return result;
}


static char *
statfs_representation(const struct statfs *p)
{
    static char buffer[250];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ f_type=%#x, f_bsize=%d, f_blocks=%ld, f_bfree=%ld, f_bavail=%ld, "
	    "f_files=%ld, f_ffree=%ld, f_fsid={%d,%d}, f_namelen=%d }",
	p->f_type,
	p->f_bsize,
	(long)p->f_blocks,
	(long)p->f_bfree,
	(long)p->f_bavail,
	(long)p->f_files,
	(long)p->f_ffree,
	p->f_fsid.__val[0],
	p->f_fsid.__val[1],
	p->f_namelen
    );
    return buffer;
}


int
plasticfs_filter_log::statfs(const char *path, struct statfs *buf)
{
    int result = inherited::statfs(path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "statfs(path = \"%s\", buf = %p) = %d; errno = %d %s\n",
	    path,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "statfs(path = \"%s\", type = %s) = %d\n",
	    path,
	    statfs_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::fstatfs(int fd, struct statfs *buf)
{
    int result = inherited::fstatfs(fd, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fstatfs(fd = %d, buf = %p) = %d; errno = %d %s\n",
	    fd,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "fstatfs(fd = %d, type = %s) = %d\n",
	    fd,
	    statfs_representation(buf),
	    result
	);
    }
    return result;
}


static char *
statfs64_representation(const struct statfs64 *p)
{
    static char buffer[250];
    snprintf
    (
	buffer,
	sizeof(buffer),
	"{ f_type=%#x, f_bsize=%d, f_blocks=%ld, f_bfree=%ld, f_bavail=%ld, "
	    "f_files=%ld, f_ffree=%ld, f_fsid={%d,%d}, f_namelen=%d }",
	p->f_type,
	p->f_bsize,
	(long)p->f_blocks,
	(long)p->f_bfree,
	(long)p->f_bavail,
	(long)p->f_files,
	(long)p->f_ffree,
	p->f_fsid.__val[0],
	p->f_fsid.__val[1],
	p->f_namelen
    );
    return buffer;
}


int
plasticfs_filter_log::statfs64(const char *path, struct statfs64 *buf)
{
    int result = inherited::statfs64(path, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "statfs64(path = \"%s\", buf = %p) = %d; errno = %d %s\n",
	    path,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "statfs64(path = \"%s\", type = %s) = %d\n",
	    path,
	    statfs64_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::fstatfs64(int fd, struct statfs64 *buf)
{
    int result = inherited::fstatfs64(fd, buf);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "fstatfs64(fd = %d, buf = %p) = %d; errno = %d %s\n",
	    fd,
	    buf,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "fstatfs64(fd = %d, type = %s) = %d\n",
	    fd,
	    statfs64_representation(buf),
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::truncate(const char *path, off_t length)
{
    int result = inherited::truncate(path, length);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "truncate(path = \"%s\", length = %ld) = %d; errno = %d %s\n",
	    path,
	    (long)length,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "truncate(path = \"%s\", length = %ld) = %d\n",
	    path,
	    (long)length,
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::ftruncate(int fd, off_t length)
{
    int result = inherited::ftruncate(fd, length);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "ftruncate(fd = %d, length = %ld) = %d; errno = %d %s\n",
	    fd,
	    (long)length,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "ftruncate(fd = %d, length = %ld) = %d\n",
	    fd,
	    (long)length,
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::truncate64(const char *path, off64_t length)
{
    int result = inherited::truncate64(path, length);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "truncate64(path = \"%s\", length = %ld) = %d; errno = %d %s\n",
	    path,
	    (long)length,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "truncate64(path = \"%s\", length = %ld) = %d\n",
	    path,
	    (long)length,
	    result
	);
    }
    return result;
}


int
plasticfs_filter_log::ftruncate64(int fd, off64_t length)
{
    int result = inherited::ftruncate64(fd, length);
    errno_insulator guard;
    if (result < 0)
    {
	stash
	(
	    "ftruncate64(fd = %d, length = %ld) = %d; errno = %d %s\n",
	    fd,
	    (long)length,
	    result,
	    guard.get(),
	    guard.gets()
	);
    }
    else
    {
	stash
	(
	    "ftruncate64(fd = %d, length = %ld) = %d\n",
	    fd,
	    (long)length,
	    result
	);
    }
    return result;
}
