//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/plasticfs/filter/log.cc
//

#ifndef LIB_PLASTICFS_FILTER_LOG_H
#define LIB_PLASTICFS_FILTER_LOG_H

#pragma interface "libplasticfs::plasticfs_filter_log"

#include <plasticfs/filter.h>
#include <rcstring_list.h>

namespace libplasticfs {

/**
  * The plasticfs_filter_log class is used to represent a file access
  * logging filter.  It doesn't actually change anything, just stashes
  * a log of everything that goes through.  The results are very similar
  * to using strace, but only the file system calls are logged.
  */
class plasticfs_filter_log:
    public plasticfs_filter
{
public:
    /**
      * The destructor.
      */
    virtual ~plasticfs_filter_log();

    /**
      * The constructor.
      */
    plasticfs_filter_log(plasticfs *deeper, const rcstring_list &args);

    /**
      * The factory method is used to produce new plasticfs_filter_log
      * instances.
      */
    static plasticfs *factory(plasticfs *deeper, const rcstring_list &args);

    // See base class for documentation.
    int open(const char *, int, int);

    // See base class for documentation.
    int close(int);

    // See base class for documentation.
    ssize_t read(int, void *, size_t);

    // See base class for documentation.
    ssize_t write(int, const void *, size_t);

    // See base class for documentation.
    off_t lseek(int, off_t, int);

    // See base class for documentation.
    off64_t lseek64(int, off64_t, int);

    // See base class for documentation.
    int readlink(const char *, char *, size_t);

    // See base class for documentation.
    int xstat(int, const char *, struct stat *);

    // See base class for documentation.
    int xstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int lxstat(int, const char *, struct stat *);

    // See base class for documentation.
    int lxstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int fxstat(int, int, struct stat *);

    // See base class for documentation.
    int fxstat64(int, int, struct stat64 *);

#ifdef FILTER_OPENDIR
    // See base class for documentation.
    DIR *opendir(const char *);

    // See base class for documentation.
    int closedir(DIR *);

    // See base class for documentation.
    dirent *readdir(DIR *);

    // See base class for documentation.
    dirent64 *readdir64(DIR *);

    // See base class for documentation.
    void rewinddir(DIR *);

    // See base class for documentation.
    void seekdir(DIR *, off_t);

    // See base class for documentation.
    off_t telldir(DIR *);

    // See base class for documentation.
    int dirfd(DIR *);

    // See base class for documentation.
    char *getcwd(char *, long);
#endif // FILTER_OPENDIR

    // See base class for documentation.
    int getdents(int, dirent *, unsigned);

    // See base class for documentation.
    int getdents64(int, dirent64 *, unsigned);

    // See base class for documentation.
    ssize_t getdirentries(int, char *, size_t, off_t *);

    // See base class for documentation.
    ssize_t getdirentries64(int, char *, size_t, off64_t *);

    // See base class for documentation.
    int chmod(const char *, mode_t);

    // See base class for documentation.
    int fchmod(int, mode_t);

    // See base class for documentation.
    int mkdir(const char *, mode_t);

    // See base class for documentation.
    int mkfifo(const char *, mode_t);

    // See base class for documentation.
    int link(const char *, const char *);

    // See base class for documentation.
    int unlink(const char *);

    // See base class for documentation.
    int rename(const char *, const char *);

    // See base class for documentation.
    long pathconf(const char *, int);

    // See base class for documentation.
    long fpathconf(int, int);

    // See base class for documentation.
    int xmknod(int ver, const char *path, mode_t mode, dev_t dev);

    // See base class for documentation.
    int fcntl(int, int, void *);

    // See base class for documentation.
    int fcntl64(int, int, void *);

    // See base class for documentation.
    int dup(int);

    // See base class for documentation.
    int dup2(int, int);

    // See base class for documentation.
    int symlink(const char *, const char *);

    // See base class for documentation.
    int chdir(const char *);

    // See base class for documentation.
    int fchdir(int);

    // See base class for documentation.
    int rmdir(const char *);

    // See base class for documentation.
    int utime(const char *, utimbuf *);

    // See base class for documentation.
    int lutime(const char *, utimbuf *);

    // See base class for documentation.
    int futime(int, utimbuf *);

    // See base class for documentation.
    int execve(const char *, char *const *, char *const *);

    // See base class for documentation.
    int fexecve(int, char *const *, char *const *);

    // See base class for documentation.
    int access(const char *, int);

    // See base class for documentation.
    int statfs(const char *, struct statfs *);

    // See base class for documentation.
    int fstatfs(int, struct statfs *);

    // See base class for documentation.
    int statfs64(const char *, struct statfs64 *);

    // See base class for documentation.
    int fstatfs64(int, struct statfs64 *);

    // See base class for documentation.
    bool is_below_mount_point(const char *, size_t &);

    // See base class for documentation.
    int relative_readlink(const char *, char *, size_t);

    // See base class for documentation.
    int relative_lstat(const char *, struct stat *);

    // See base class for documentation.
    int truncate(const char *, off_t);

    // See base class for documentation.
    int truncate64(const char *, off64_t);

    // See base class for documentation.
    int ftruncate(int, off_t);

    // See base class for documentation.
    int ftruncate64(int, off64_t);

private:
    typedef plasticfs_filter inherited;

    /**
      * The fd instance variable is used to remember the file descriptor of
      * the log file.  This is set by the deferred_open method.
      */
    int fd;

    /**
      * The filename instance variable is used to remember the file name
      * of the log file.  This is used by the deferred_open method.
      */
    rcstring filename;

    /**
      * The deferred open method is used to open the log file once a
      * method is called which requires logging.
      */
    void deferred_open();

    /**
      * The stash method is used to write arbitrary text to the log file.
      * This is used by the various methods to record what happened.
      */
    void stash(const char *, ...)
					  __attribute__((format(printf, 2, 3)));

    /**
      * The default constructor.  Do not use.
      */
    plasticfs_filter_log();

    /**
      * The copy constructor.  Do not use.
      */
    plasticfs_filter_log(const plasticfs_filter_log &);

    /**
      * The assignment operator.  Do not use.
      */
    plasticfs_filter_log &operator=(const plasticfs_filter_log &);
};

} // namespace libplasticfs

#endif // LIB_PLASTICFS_FILTER_LOG_H
