//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_module class
//

#pragma implementation "libplasticfs::plasticfs_filter_module"

#include <plasticfs/filter/module.h>
#include <dlfcn.h>
#include <utils.h>

using namespace libplasticfs;


plasticfs_filter_module::~plasticfs_filter_module()
{
    //
    // We have to delete this first, because it uses code in the loadable
    // module referred to by 'handle'.  If we let the plasticfs_filter
    // destructor do it (it runs *after* this destructor) then we would
    // be trying to execute clde that isn't loaded any more.
    //
    delete_deeper();

    //
    // Now unload the code for the loadable module.
    //
    if (handle)
    {
	dlclose(handle);
	handle = 0;
    }
}


plasticfs_filter_module::plasticfs_filter_module(plasticfs *arg1, void *arg2) :
    plasticfs_filter(arg1),
    handle(arg2)
{
}


plasticfs *
plasticfs_filter_module::factory(plasticfs *deeper1, const rcstring_list &args)
{
    static char symbol_name[] = "factory";

    //
    // Build the file name of the shared library we are trying to open.
    // If the library doesn't exist (or can't be opened), fail.
    //
    if (args.size() < 1)
	return 0;
    char filename[100];
    strlcpy(filename, "plasticfs/", sizeof(filename));
    strlcat(filename, args[0].c_str(), sizeof(filename));
    strlcat(filename, ".so", sizeof(filename));
    void *handle = dlopen(filename, RTLD_NOW);
    if (!handle)
    {
	error_message
	(
	    "plasticfs: module \"%s\": %s\n",
	    args[0].c_str(),
	    dlerror()
	);
	return 0;
    }

    //
    // Look for the "factory" symbol.
    // If it doesn't exist, fail.
    //
    void *factory_p = dlsym(handle, symbol_name);
    char *err = dlerror();
    if (err)
    {
	error_message
	(
	    "plasticfs: module \"%s\": symbol \"%s\": %s\n",
	    args[0].c_str(),
	    symbol_name,
	    err
	);
	dlclose(handle);
	return 0;
    }
    if (factory_p == 0)
    {
	error_message
	(
	    "plasticfs: module \"%s\": symbol \"%s\": has NULL value\n",
	    args[0].c_str(),
	    symbol_name
	);
	dlclose(handle);
	return 0;
    }

    //
    // Run the factory function.  If it returns NULL, fail.
    //
    typedef plasticfs *(*factory_t)(plasticfs *, const rcstring_list &);
    plasticfs *deeper2 = ((factory_t)factory_p)(deeper1, args);
    if (!deeper2)
    {
	error_message
	(
	    "plasticfs: module \"%s\": %s returned NULL\n",
	    args[0].c_str(),
	    symbol_name
	);
	dlclose(handle);
	return 0;
    }

    //
    // Create a new module instance to remember the shared libraries
    // handle.  When it is deleted, the handle has to be closed.
    //
    plasticfs *deeper3 = new plasticfs_filter_module(deeper2, handle);
    if (!deeper3)
    {
	delete deeper2;
	dlclose(handle);
	return 0;
    }

    //
    // Report success.
    //
    return deeper3;
}


bool
plasticfs_filter_module::is_below_mount_point(const char *, size_t &)
{
    //
    // This is called by relative_realpath, available only to derived
    // classes of plasticfs_filter.  And since we (the derived class in
    // question) never call relative_realpath is isn't relevant what we
    // return at this point.  Be conservative.
    //
    return false;
}


int 
plasticfs_filter_module::relative_readlink(const char *, char *, size_t)
{
    // because is_below_mount_point is always false, this method never called.
    return -1;
}


int 
plasticfs_filter_module::relative_lstat(const char *, struct stat *)
{
    // because is_below_mount_point is always false, this method never called.
    return -1;
}
