//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_nocase class
//

#pragma implementation "libplasticfs::plasticfs_filter_nocase"

#include <plasticfs/filter/nocase.h>
#include <errno_insulator.h>
#include <cstring>
#include <cerrno>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/vfs.h>

//
// Define the NOCASE_DEBUG symbol when you need to see what methods
// are being called.
//
//#define NOCASE_DEBUG

using namespace libplasticfs;


plasticfs_filter_nocase::~plasticfs_filter_nocase()
{
}


plasticfs_filter_nocase::plasticfs_filter_nocase(plasticfs *deeper,
	const rcstring_list &args) :
    plasticfs_filter(deeper)
{
    switch (args.size())
    {
    case 0:
    case 1:
	mount_point = rcstring("/");
	break;

    default:
	error_message
	(
	    "plasticfs: %s filter: requires one argument\n",
	    args[0].c_str()
	);
	// fall through...

    case 2:
	mount_point = make_absolute(args[1]);
	break;
    }
}


plasticfs *
plasticfs_filter_nocase::factory(plasticfs *deeper,
    const rcstring_list &args)
{
    return new plasticfs_filter_nocase(deeper, args);
}


bool
plasticfs_filter_nocase::is_below_mount_point(const char *buf, size_t &size)
{
    return is_below_directory(mount_point, buf, size);
}


bool
plasticfs_filter_nocase::relative_name_mapping(const char *path,
    rcstring &result)
{
    rcstring dirname;
    rcstring filename;
    const char *cp = strrchr(path, '/');
    if (cp)
    {
	dirname = rcstring(path, cp - path);
	filename = cp + 1;
    }
    else
    {
	dirname = ".";
	filename = path;
    }

    plasticfs *fs = get_deeper();
    DIR *dirp = fs->opendir(dirname.c_str());
    if (!dirp)
    {
	#ifdef NOCASE_DEBUG
	error_message("%s: %d: relative_name_mapping \"%s\" -> false\n",
	    __FILE__, __LINE__, path);
	#endif
	return false;
    }
    for (;;)
    {
	dirent *d = fs->readdir(dirp);
	if (!d)
	    break;
	if (!d->d_ino)
	    continue;
	if
	(
	    d->d_name[0] == '.'
	&&
	    (d->d_name[1] == 0 || (d->d_name[1] == '.' && d->d_name[2] == 0))
	)
	    continue;
	if (0 == strcasecmp(filename.c_str(), d->d_name))
	{
	    result = d->d_name;
	    fs->closedir(dirp);
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: relative_name_mapping \"%s\" -> \"%s\"\n",
		__FILE__, __LINE__, path, result.c_str());
	    #endif
	    return (filename != result);
	}
    }
    fs->closedir(dirp);
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: relative_name_mapping \"%s\" -> false\n", __FILE__,
	__LINE__, path);
    #endif
    return false;
}


int
plasticfs_filter_nocase::pc_name_max(int n)
{
    return n;
}


int
plasticfs_filter_nocase::pc_no_trunc(int n)
{
    return n;
}


int
plasticfs_filter_nocase::open(const char *path, int flags, int mode)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: open(\"%s\", %#o, %#o)\n", __FILE__, __LINE__, path,
	flags, mode);
    #endif

    bool file_must_exist = ((flags & O_CREAT) == 0);

    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, file_must_exist, true))
	return -1;
    if (abspath == relpath)
	return inherited::open(abspath.c_str(), flags, mode);

    int fd = inherited::open(abspath.c_str(), flags, mode);
    if (fd >= 0)
    {
	//
	// Keep track of open files below our mount point.
	//
	errno_insulator guard;
	fdtab.open(fd, flags, relpath, get_deeper());
    }
    return fd;
}


int
plasticfs_filter_nocase::readlink(const char *path, char *buf, size_t nbytes)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: readlink(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    return inherited::readlink(abspath.c_str(), buf, nbytes);
}


int
plasticfs_filter_nocase::lxstat(int vers, const char *path, struct stat *st)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: lxstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    return inherited::lxstat(vers, abspath.c_str(), st);
}


int
plasticfs_filter_nocase::lxstat64(int vers, const char *path, struct stat64 *st)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: lxstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    return inherited::lxstat64(vers, abspath.c_str(), st);
}


int
plasticfs_filter_nocase::xstat(int vers, const char *path, struct stat *st)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: xstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::xstat(vers, abspath.c_str(), st);
}


int
plasticfs_filter_nocase::xstat64(int vers, const char *path, struct stat64 *st)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: xstat64(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::xstat64(vers, abspath.c_str(), st);
}


int
plasticfs_filter_nocase::relative_lstat(const char *path, struct stat *st)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: relative_lstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring relpath(path_join(mount_point, rcstring(path)));
    return inherited::lxstat(_STAT_VER, relpath.c_str(), st);
}


int
plasticfs_filter_nocase::relative_readlink(const char *path, char *buffer,
    size_t nbytes)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: relative_readlink(\"%s\")\n", __FILE__, __LINE__,
	path);
    #endif
    rcstring relpath(path_join(mount_point, rcstring(path)));
    return inherited::readlink(relpath.c_str(), buffer, nbytes);
}


int
plasticfs_filter_nocase::close(int fd)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: close(%d)\n", __FILE__, __LINE__, fd);
    #endif

    //
    // If the file descriptor is in our active list, get rid of it and
    // release its resources.
    //
    fdtab.close(fd);

    //
    // We still have to actually close it.
    //
    return inherited::close(fd);
}


void
plasticfs_filter_nocase::relative_name_unmapping(char *, size_t)
{
    // do nothing
}


int
plasticfs_filter_nocase::getdents(int fd, dirent *buf, unsigned nbytes)
{
    int save_errno = errno;
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: getdents(fd=%d, buf=%p, nbytes=%u) {\n", __FILE__,
	__LINE__, fd, buf, nbytes);
    #endif

    for (;;)
    {
	//
	// Grab the directory entries.
	//
	int result = inherited::getdents(fd, buf, nbytes);
	if (result <= 0)
	    return result;

	//
	// See if this file descriptor is in our list.
	// No special processing for file descriptors not in our list.
	//
	file_tracker::active_t *p = fdtab.find(fd);
	if (!p)
	{
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    errno = save_errno;
	    return result;
	}

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < result)
	{
	    dirent *dp = (dirent *)((char *)buf + pos);
	    if (dp->d_ino)
	    {
		//
		// Convert the filename, the reverse of relative_name_mapping
		//
		relative_name_unmapping(dp->d_name, sizeof(dp->d_name));

		#ifdef NOCASE_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
	        if (p->direntries.query(ent))
		{
		    #ifdef NOCASE_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    return result;
	}
    }
}


int
plasticfs_filter_nocase::getdents64(int fd, dirent64 *buf, unsigned nbytes)
{
    int save_errno = errno;
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: getdents64(fd=%d, buf=%p, nbytes=%u) {\n", __FILE__,
	__LINE__, fd, buf, nbytes);
    #endif

    for (;;)
    {
	//
	// Grab the directory entries.
	//
	int result = inherited::getdents64(fd, buf, nbytes);
	if (result <= 0)
	    return result;

	//
	// See if this file descriptor is in our list.
	// No special processing for file descriptors not in our list.
	//
	file_tracker::active_t *p = fdtab.find(fd);
	if (!p)
	{
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    errno = save_errno;
	    return result;
	}

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < result)
	{
	    dirent64 *dp = (dirent64 *)((char *)buf + pos);
	    if (dp->d_ino)
	    {
		//
		// Convert the filename, the reverse of relative_name_mapping
		//
		relative_name_unmapping(dp->d_name, sizeof(dp->d_name));

		#ifdef NOCASE_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
	        if (p->direntries.query(ent))
		{
		    #ifdef NOCASE_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    return result;
	}
    }
}


ssize_t
plasticfs_filter_nocase::getdirentries(int fd, char *buf, size_t nbytes,
    off_t *basep)
{
    int save_errno = errno;
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: getdirentries(fd=%d, buf=%p, nbytes=%ld, "
	"*basep=%ld) {\n", __FILE__, __LINE__, fd, buf, (long)nbytes,
	(long)*basep);
    #endif

    for (;;)
    {
	//
	// Grab the directory entries.
	//
	int result = inherited::getdirentries(fd, buf, nbytes, basep);
	if (result <= 0)
	    return result;

	//
	// See if this file descriptor is in our list.
	// No special processing for file descriptors not in our list.
	//
	file_tracker::active_t *p = fdtab.find(fd);
	if (!p)
	{
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    errno = save_errno;
	    return result;
	}

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < result)
	{
	    dirent *dp = (dirent *)(buf + pos);
	    if (dp->d_ino)
	    {
		//
		// Convert the filename, the reverse of relative_name_mapping
		//
		relative_name_unmapping(dp->d_name, sizeof(dp->d_name));

		#ifdef NOCASE_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
	        if (p->direntries.query(ent))
		{
		    #ifdef NOCASE_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    return result;
	}
    }
}


ssize_t
plasticfs_filter_nocase::getdirentries64(int fd, char *buf, size_t nbytes,
    off64_t *basep)
{
    int save_errno = errno;
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: getdirentries64(fd=%d, buf=%p, nbytes=%ld, "
	"*basep=%ld) {\n", __FILE__, __LINE__, fd, buf, (long)nbytes,
	(long)*basep);
    #endif

    for (;;)
    {
	//
	// Grab the directory entries.
	//
	#ifdef NOCASE_DEBUG
	error_message("%s: %d\n", __FILE__, __LINE__);
	#endif
	int result = inherited::getdirentries64(fd, buf, nbytes, basep);
	if (result <= 0)
	    return result;

	//
	// See if this file descriptor is in our list.
	// No special processing for file descriptors not in our list.
	//
	file_tracker::active_t *p = fdtab.find(fd);
	if (!p)
	{
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    errno = save_errno;
	    return result;
	}

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < result)
	{
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d\n", __FILE__, __LINE__);
	    #endif
	    dirent64 *dp = (dirent64 *)(buf + pos);
	    if (dp->d_ino)
	    {
		//
		// Convert the filename, the reverse of relative_name_mapping
		//
		relative_name_unmapping(dp->d_name, sizeof(dp->d_name));

		#ifdef NOCASE_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
	        if (p->direntries.query(ent))
		{
		    #ifdef NOCASE_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	    #endif
	    return result;
	}
    }
}


int
plasticfs_filter_nocase::chmod(const char *path, mode_t mode)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::chmod(abspath.c_str(), mode);
}


int
plasticfs_filter_nocase::mkdir(const char *path, mode_t mode)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;
    return inherited::mkdir(abspath.c_str(), mode);
}


int
plasticfs_filter_nocase::mkfifo(const char *path, mode_t mode)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;
    return inherited::mkfifo(abspath.c_str(), mode);
}


int
plasticfs_filter_nocase::link(const char *src_path, const char *dest_path)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring src_abspath;
    rcstring src_relpath;
    if (!relative_realpath(src_path, src_abspath, src_relpath, true, true))
	return -1;

    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;

    return inherited::link(src_abspath.c_str(), dest_abspath.c_str());
}


int
plasticfs_filter_nocase::symlink(const char *text, const char *dest_path)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;
    return inherited::symlink(text, dest_abspath.c_str());
}


int
plasticfs_filter_nocase::unlink(const char *path)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    return inherited::unlink(abspath.c_str());
}


int
plasticfs_filter_nocase::rmdir(const char *path)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::rmdir(abspath.c_str());
}


int
plasticfs_filter_nocase::rename(const char *src_path, const char *dest_path)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: rename(\"%s\", \"%s\")\n", __FILE__, __LINE__,
	src_path, dest_path);
    #endif
    rcstring src_abspath;
    rcstring src_relpath;
    if (!relative_realpath(src_path, src_abspath, src_relpath, true, true))
	return -1;

    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;

    return inherited::rename(src_abspath.c_str(), dest_abspath.c_str());
}


long
plasticfs_filter_nocase::pathconf(const char *path, int name)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: pathconf(\"%s\", %d)\n", __FILE__, __LINE__, path,
	name);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    long result = inherited::pathconf(abspath.c_str(), name);
    if (result >= 0)
    {
	switch (name)
	{
	case _PC_NAME_MAX:
	    result = pc_name_max(result);
	    break;

	case _PC_NO_TRUNC:
	    result = pc_no_trunc(result);
	    break;
	}
    }
    return result;
}


long
plasticfs_filter_nocase::fpathconf(int fd, int name)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: fpathconf(%d, %d)\n", __FILE__, __LINE__, fd, name);
    #endif
    long result = inherited::fpathconf(fd, name);
    if (result >= 0 && fdtab.find(fd))
    {
	switch (name)
	{
	case _PC_NAME_MAX:
	    result = pc_name_max(result);
	    break;

	case _PC_NO_TRUNC:
	    result = pc_no_trunc(result);
	    break;
	}
    }
    return result;
}


int
plasticfs_filter_nocase::xmknod(int ver, const char *path, mode_t mode,
    dev_t dev)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;
    return inherited::xmknod(ver, abspath.c_str(), mode, dev);
}


int
plasticfs_filter_nocase::fcntl(int fd, int cmd, void *arg)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: fcntl(%d, %d, %ld) {\n", __FILE__, __LINE__, fd, cmd,
	(long)arg);
    #endif
    int result = inherited::fcntl(fd, cmd, arg);
    if (cmd == F_DUPFD && result >= 0)
	fdtab.dup(fd, result);
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
    #endif
    return result;
}


int
plasticfs_filter_nocase::fcntl64(int fd, int cmd, void *arg)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: fcntl64(%d, %d) {\n", __FILE__, __LINE__, fd, cmd);
    #endif
    int result = inherited::fcntl64(fd, cmd, arg);
    if (cmd == F_DUPFD && result >= 0)
	fdtab.dup(fd, result);
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
    #endif
    return result;
}


int
plasticfs_filter_nocase::dup(int fd)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    int result = inherited::dup(fd);
    if (result >= 0)
	fdtab.dup(fd, result);
    return result;
}


int
plasticfs_filter_nocase::dup2(int oldfd, int newfd)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    int result = inherited::dup2(oldfd, newfd);
    if (result >= 0)
	fdtab.dup(oldfd, result);
    return result;
}


int
plasticfs_filter_nocase::chdir(const char *path)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: chdir(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::chdir(abspath.c_str());
}


int
plasticfs_filter_nocase::utime(const char *path, utimbuf *buf)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::utime(abspath.c_str(), buf);
}


int
plasticfs_filter_nocase::lutime(const char *path, utimbuf *buf)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    return inherited::lutime(abspath.c_str(), buf);
}


int
plasticfs_filter_nocase::execve(const char *path, char *const *argv,
    char *const *envp)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: execve(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::execve(abspath.c_str(), argv, envp);
}


int
plasticfs_filter_nocase::access(const char *path, int type)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: access(\"%s\", %d)\n", __FILE__, __LINE__, path,
	type);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::access(abspath.c_str(), type);
}


int
plasticfs_filter_nocase::statfs(const char *path, struct statfs *buf)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: statfs(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    int result = inherited::statfs(abspath.c_str(), buf);
    if (result >= 0 && abspath != relpath)
	buf->f_namelen = pc_name_max(buf->f_namelen);
    return result;
}


int
plasticfs_filter_nocase::statfs64(const char *path, struct statfs64 *buf)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: statfs64(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    int result = inherited::statfs64(abspath.c_str(), buf);
    if (result >= 0 && abspath != relpath)
	buf->f_namelen = pc_name_max(buf->f_namelen);
    return result;
}


int
plasticfs_filter_nocase::truncate(const char *path, off_t length)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: truncate(\"%s\", %ld)\n", __FILE__, __LINE__, path,
	(long)length);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::truncate(abspath.c_str(), length);
}


int
plasticfs_filter_nocase::truncate64(const char *path, off64_t length)
{
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: truncate64(\"%s\", %ld)\n", __FILE__, __LINE__, path,
	(long)length);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    return inherited::truncate64(abspath.c_str(), length);
}
