//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/plasticfs/filter/nocase.cc
//

#ifndef LIB_PLASTICFS_FILTER_NOCASE_H
#define LIB_PLASTICFS_FILTER_NOCASE_H

#pragma interface "libplasticfs::plasticfs_filter_nocase"

#include <plasticfs/filter.h>
#include <file_tracker.h>
#include <rcstring_list.h>

namespace libplasticfs {

/**
  * The plasticfs_filter_nocase class is used to represent
  * a file system where filenames are case insensitive.
  */
class plasticfs_filter_nocase:
    public plasticfs_filter
{
public:
    /**
      * The destructor.
      */
    virtual ~plasticfs_filter_nocase();

    /**
      * The constructor.
      */
    plasticfs_filter_nocase(plasticfs *deeper, const rcstring_list &args);

    /**
      * The factory method is used to create anew instance of this class.
      */
    static plasticfs *factory(plasticfs *, const rcstring_list &);

    // See base class for documentation.
    int open(const char *, int, int);

    // See base class for documentation.
    int close(int);

    // See base class for documentation.
    int getdents(int, dirent *, unsigned);

    // See base class for documentation.
    int getdents64(int, dirent64 *, unsigned);

    // See base class for documentation.
    ssize_t getdirentries(int, char *, size_t, off_t *);

    // See base class for documentation.
    ssize_t getdirentries64(int, char *, size_t, off64_t *);

    // See base class for documentation.
    int xstat(int, const char *, struct stat *);

    // See base class for documentation.
    int xstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int lxstat(int, const char *, struct stat *);

    // See base class for documentation.
    int lxstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int readlink(const char *, char *, size_t);

    // See base class for documentation.
    int chmod(const char *, mode_t);

    // See base class for documentation.
    int mkdir(const char *, mode_t);

    // See base class for documentation.
    int mkfifo(const char *, mode_t);

    // See base class for documentation.
    int unlink(const char *);

    // See base class for documentation.
    int rename(const char *, const char *);

    // See base class for documentation.
    int link(const char *, const char *);

    // See base class for documentation.
    long fpathconf(int, int);

    // See base class for documentation.
    long pathconf(const char *, int);

    // See base class for documentation.
    int xmknod(int, const char *, mode_t, dev_t);

    // See base class for documentation.
    int fcntl(int, int, void *);

    // See base class for documentation.
    int fcntl64(int, int, void *);

    // See base class for documentation.
    int dup(int);

    // See base class for documentation.
    int dup2(int, int);

    // See base class for documentation.
    int symlink(const char *, const char *);

    // See base class for documentation.
    int chdir(const char *);

    // See base class for documentation.
    int rmdir(const char *);

    // See base class for documentation.
    int utime(const char *, utimbuf *);

    // See base class for documentation.
    int lutime(const char *, utimbuf *);

    // See base class for documentation.
    int execve(const char *, char *const *, char *const *);

    // See base class for documentation.
    int access(const char *, int);

    // See base class for documentation.
    int statfs(const char *, struct statfs *);

    // See base class for documentation.
    int statfs64(const char *, struct statfs64 *);

    // See base class for documentation.
    int truncate(const char *, off_t);

    // See base class for documentation.
    int truncate64(const char *, off64_t);

private:
    typedef plasticfs_filter inherited;

    /**
      * The mount_point instance variable is used to remember the point
      * within the file system where filenames become case insensitive.
      */
    rcstring mount_point;

    /**
      * The fdtab instance variable is used to remember information
      * about open file descriptors.
      */
    file_tracker fdtab;

    // See base class for documentation.
    bool is_below_mount_point(const char *, size_t &);

    // See base class for documentation.
    int relative_readlink(const char *, char *, size_t);

    // See base class for documentation.
    int relative_lstat(const char *, struct stat *);

    // See base class for documentation.
    bool relative_name_mapping(const char *, rcstring &);

    /**
      * The relative_name_mapping method is used to perform the reverse
      * maping to the relative_name_mapping method.  The `name' is
      * NUL terminated.  The new name shall not exceed `size' including
      * the NUL.  The default implementation does nothing.
      */
    virtual void relative_name_unmapping(char *name, size_t size);

    /**
      * The pc_name_max method is used to adjust the results of fpathconf,
      * pathconf, statfs and statfs64 methods when the file name mapping
      * also shortens filenames to a known limit.  The default to to
      * leave the length unchanged.
      */
    virtual int pc_name_max(int);

    /**
      * The pc_no_trunc method is used to adjust the results of fpathconf
      * and pathconf methods when the file name mapping also shortens
      * filenames to a known limit.  The default to to leave the setting
      * unchanged.
      */
    virtual int pc_no_trunc(int);

    /**
      * The default constructor.  Do not use.
      */
    plasticfs_filter_nocase();

    /**
      * The copy constructor.  Do not use.
      */
    plasticfs_filter_nocase(const plasticfs_filter_nocase &);

    /**
      * The assignment operator.  Do not use.
      */
    plasticfs_filter_nocase &operator=(const plasticfs_filter_nocase &);
};

} // namespace libplasticfs

#endif // LIB_PLASTICFS_FILTER_NOCASE_H
