//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_nocase_dos class
//

#pragma implementation "libplasticfs::plasticfs_filter_nocase_dos"

#include <plasticfs/filter/nocase/dos.h>
#include <cctype>
#include <cstring>
#include <utils.h>

using namespace libplasticfs;


plasticfs_filter_nocase_dos::~plasticfs_filter_nocase_dos()
{
}


plasticfs_filter_nocase_dos::plasticfs_filter_nocase_dos(
	plasticfs *deeper, const rcstring_list &args) :
    plasticfs_filter_nocase(deeper, args)
{
}


plasticfs *
plasticfs_filter_nocase_dos::factory(plasticfs *deeper,
    const rcstring_list &args)
{
    return new plasticfs_filter_nocase_dos(deeper, args);
}


static inline int
mangler(unsigned char c)
{
    if (islower(c))
	return toupper(c);
    if (isalnum(c) || c == '-')
	return c;
    return 0;
}


void
plasticfs_filter_nocase_dos::relative_name_unmapping(char *name, size_t)
{
    //
    // leave "." and ".." alone
    //
    if (name[0] == '.' && (!name[1] || (name[1] == '.' && !name[2])))
	return;

    //
    // Get the left hand side, before the dot.
    //
    int len = 0;
    char *ip = name;
    char *op = name;
    if (name[0] == '.')
    {
	// Leave leading dots alone.
	// This isn't strictly dospel.
	++ip;
	++op;
	++len;
    }
    while (*ip && *ip != '.' && len < 8)
    {
	int c = mangler(*ip);
	if (c)
	{
	    *op++ = c;
	    ++len;
	}
	++ip;
    }
    while (*ip && *ip != '.')
	++ip;

    //
    // Watch out for names which contain no legal DOS characters at all.
    //
    if (op == name)
	*op++ = 'A';

    //
    // Get the right hand side, after the dot.
    //
    if (ip[0] == '.' && ip[1])
    {
	*op++ = *ip++;
	len = 0;
	while (*ip && len < 3)
	{
	    int c = mangler(*ip);
	    if (c)
	    {
		*op++ = c;
		++len;
	    }
	    ++ip;
	}
    }

    //
    // Terminate it.
    //
    *op = '\0';
}


bool
plasticfs_filter_nocase_dos::relative_name_mapping(const char *path,
    rcstring &result)
{
    //
    // Break the path into a directory and a filename.
    //
    rcstring dirname;
    rcstring filename;
    const char *cp = strrchr(path, '/');
    if (cp)
    {
	dirname = rcstring(path, cp - path);
	filename = cp + 1;
    }
    else
    {
	dirname = ".";
	filename = path;
    }

    //
    // Generate a DOS filename as the pattern we are looking for.
    //
    char pattern[256];
    strlcpy(pattern, filename.c_str(), sizeof(pattern));
    relative_name_unmapping(pattern, sizeof(pattern));

    //
    // Walk the directory, looking for the pattern.
    //
    plasticfs *fs = get_deeper();
    DIR *dirp = fs->opendir(dirname.c_str());
    if (!dirp)
    {
	#ifdef NOCASE_DEBUG
	error_message("%s: %d: relative_name_mapping \"%s\" -> false\n",
	    __FILE__, __LINE__, path);
	#endif
	return false;
    }
    for (;;)
    {
	dirent *d = fs->readdir(dirp);
	if (!d)
	    break;
	if (!d->d_ino)
	    continue;
	if
	(
	    d->d_name[0] == '.'
	&&
	    (d->d_name[1] == 0 || (d->d_name[1] == '.' && d->d_name[2] == 0))
	)
	    continue;

	//
	// We take the actual name and chew it over into an ugly DOS
	// name, and then do a simple string compare of the pattern
	// with the candidate.	It may not ber super fast, but we are
	// definutely comparing oranges with oranges.
	//
	char candidate[256];
	strlcpy(candidate, d->d_name, sizeof(candidate));
	relative_name_unmapping(candidate, sizeof(candidate));

	if (0 == strcmp(pattern, candidate))
	{
	    result = d->d_name;
	    fs->closedir(dirp);
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: relative_name_mapping \"%s\" -> \"%s\"\n",
		__FILE__, __LINE__, path, result.c_str());
	    #endif
	    return (filename != result);
	}
    }
    fs->closedir(dirp);
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: relative_name_mapping \"%s\" -> false\n", __FILE__,
	__LINE__, path);
    #endif
    return false;
}


int
plasticfs_filter_nocase_dos::pc_name_max(int n)
{
    return (n < 12 ? n : 12);
}


int
plasticfs_filter_nocase_dos::pc_no_trunc(int)
{
    return 0;
}
