//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_nocase_shorten class
//

#pragma implementation "libplasticfs::plasticfs_filter_nocase_shorten"

#include <plasticfs/filter/nocase/shorten.h>
#include <cstdlib>
#include <cstring>
#include <unistd.h>

using namespace libplasticfs;


plasticfs_filter_nocase_shorten::~plasticfs_filter_nocase_shorten()
{
}


static rcstring_list
extract_first_argument(const rcstring_list args)
{
    rcstring_list temp;
    if (args.size() > 0)
	temp.push_back(args[0]);
    if (args.size() > 1)
	temp.push_back(args[1]);
    return temp;
}


plasticfs_filter_nocase_shorten::plasticfs_filter_nocase_shorten(
	plasticfs *deeper, const rcstring_list &args) :
    plasticfs_filter_nocase(deeper, extract_first_argument(args)),
    name_max(14)
{
    rcstring_list temp(args);
    switch (args.size())
    {
    case 0:
    case 1:
    case 2:
	break;

    default:
	error_message
	(
	    "plasticfs: usage: %s <directory> [ <name_max> ]\n",
	    args[0].c_str()
	);
	// fall through...

    case 3:
	char *end = 0;
	long n = strtol(args[2].c_str(), &end, 0);
	if (!end || *end)
	{
	    error_message
	    (
		"plasticfs: %s filter: second argument invalid, "
		    "must be a number\n",
		args[0].c_str()
	    );
	    break;
	}
	if (n < 2 || n > 32767)
	{
	    error_message
	    (
		"plasticfs: %s filter: second argument (%ld) out of range\n",
		args[0].c_str(),
		n
	    );
	    break;
	}
	name_max = n;
	break;
    }
}


plasticfs *
plasticfs_filter_nocase_shorten::factory(plasticfs *deeper,
    const rcstring_list &args)
{
    return new plasticfs_filter_nocase_shorten(deeper, args);
}


void
plasticfs_filter_nocase_shorten::relative_name_unmapping(char *name, size_t)
{
    int len = strlen(name);
    if (len > name_max)
	name[name_max] = '\0';
}


bool
plasticfs_filter_nocase_shorten::relative_name_mapping(const char *path,
    rcstring &result)
{
    //
    // Break the path into a directory and a filename.
    //
    rcstring dirname;
    rcstring filename;
    const char *cp = strrchr(path, '/');
    if (cp)
    {
	dirname = rcstring(path, cp - path);
	filename = cp + 1;
    }
    else
    {
	dirname = ".";
	filename = path;
    }

    //
    // Open the directory and get the apropriate filename maximum.
    //
    plasticfs *fs = get_deeper();
    DIR *dirp = fs->opendir(dirname.c_str());
    if (!dirp)
    {
	#ifdef NOCASE_DEBUG
	error_message("%s: %d: relative_name_mapping \"%s\" -> false\n",
	    __FILE__, __LINE__, path);
	#endif
	return false;
    }
    long name_limit = fpathconf(dirfd(dirp), _PC_NAME_MAX);
    if (name_limit < 0)
	name_limit = name_max;
    else if (name_limit > name_max)
	name_limit = name_max;

    //
    // Walk the directory, looking for the (shortened) filename.
    //
    for (;;)
    {
	dirent *d = fs->readdir(dirp);
	if (!d)
	    break;
	if (!d->d_ino)
	    continue;
	if
	(
	    d->d_name[0] == '.'
	&&
	    (d->d_name[1] == 0 || (d->d_name[1] == '.' && d->d_name[2] == 0))
	)
	    continue;

	if (0 == strncmp(filename.c_str(), d->d_name, name_limit))
	{
	    result = d->d_name;
	    fs->closedir(dirp);
	    #ifdef NOCASE_DEBUG
	    error_message("%s: %d: relative_name_mapping \"%s\" -> \"%s\"\n",
		__FILE__, __LINE__, path, result.c_str());
	    #endif
	    return (filename != result);
	}
    }
    fs->closedir(dirp);
    #ifdef NOCASE_DEBUG
    error_message("%s: %d: relative_name_mapping \"%s\" -> false\n", __FILE__,
	__LINE__, path);
    #endif
    return false;
}


int
plasticfs_filter_nocase_shorten::pc_name_max(int n)
{
    return (n < name_max ? n : name_max);
}


int
plasticfs_filter_nocase_shorten::pc_no_trunc(int)
{
    return 0;
}
