//
//	plasticfs - the plastic file system
//	Copyright (C) 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_smartlink class
//

#pragma implementation "libplasticfs::plasticfs_filter_smartlink"

#include <plasticfs/filter/smartlink.h>
#include <cstring>
#include <cctype>
#include <cstdlib>
#include <cerrno>
#include <fcntl.h>
#include <sys/stat.h>

using namespace libplasticfs;


plasticfs_filter_smartlink::~plasticfs_filter_smartlink()
{
}


plasticfs_filter_smartlink::plasticfs_filter_smartlink(plasticfs *deeper,
       	const rcstring_list &args) :
    plasticfs_filter(deeper)
{
    switch (args.size())
    {
    case 0:
    case 1:
	mount_point = rcstring("/");
	break;

    default:
	error_message("plasticfs: smartlink filter: requires one argument");
	// fall through...

    case 2:
	mount_point = make_absolute(args[1]);
	break;
    }
}


plasticfs *
plasticfs_filter_smartlink::factory(plasticfs *deeper,
    const rcstring_list &args)
{
    return new plasticfs_filter_smartlink(deeper, args);
}


bool
plasticfs_filter_smartlink::is_below_mount_point(const char *buf, size_t &size)
{
    return is_below_directory(mount_point, buf, size);
}


int
plasticfs_filter_smartlink::relative_lstat(const char *path, struct stat *st)
{
    rcstring abspath = path_join(mount_point, rcstring(path));
    return inherited::lxstat(_STAT_VER, abspath.c_str(), st);
}


int
plasticfs_filter_smartlink::relative_readlink(const char *path, char *buffer1,
    size_t nbytes)
{
    //
    // Read the symbolic link into a temporary buffer.
    //
    rcstring abspath = path_join(mount_point, rcstring(path));
    char buffer2[3000];
    int result = inherited::readlink(abspath.c_str(), buffer2, sizeof(buffer2));
    if (result < 0)
	return result;

    //
    // Look for shell expansions in the value we have read.
    //
    const char *b2p = buffer2;
    const char *b2end = buffer2 + result;
    char *b1p = buffer1;
    char *b1end = buffer1 + nbytes;
    while (b2p < b2end)
    {
	size_t b2nbytes = b2end - b2p;
	const char *b2mid = (const char *)memchr(b2p, '$', b2nbytes);
	if (!b2mid)
	{
	    // nothing left
	    if (b2nbytes > (size_t)(b1end - b1p))
		b2nbytes = b1end - b1p;
	    memcpy(b1p, b2p, b2nbytes);
	    b1p += b2nbytes;
	    break;
	}

	//
	// Found a '$' character, copy everything before it into the
	// result buffer.
	//
	b2nbytes = b2mid - b2p;
	if (b2nbytes > (size_t)(b1end - b1p))
	{
	    b2nbytes = b1end - b1p;
	    memcpy(b1p, b2p, b2nbytes);
	    b1p += b2nbytes;
	    break;
	}
	memcpy(b1p, b2p, b2nbytes);
	b1p += b2nbytes;
	b2p += b2nbytes;

	//
	// Now extract the value of the environment variable.
	//
	++b2p;
	if (b2p >= b2end)
	    break;
	rcstring env_var_name;
	if (*b2p == '{')
	{
	    ++b2p;
	    const char *cp = (const char *)memchr(b2p, '}', b2end - b2p);
	    if (cp)
	    {
		env_var_name = rcstring(b2p, cp - b2p);
		b2p = cp + 1;
	    }
	    else
	    {
		env_var_name = rcstring(b2p, b2end - b2p);
		b2p = b2end;
	    }
	}
	else
	{
	    const char *cp = b2p;
	    while (cp < b2end)
	    {
		if (!isalnum((unsigned char)*cp) && *cp != '_')
		    break;
		++cp;
	    }
	    env_var_name = rcstring(b2p, cp - b2p);
	    b2p = cp;
	}

	//
	// Get the value of the environment variable.
	// Do nothing if it's not defined.
	//
	const char *env_value = getenv(env_var_name.c_str());
	if (!env_value || !*env_value)
	    continue;

	//
	// Transfer the value into the result buffer.
	//
	size_t env_nbytes = strlen(env_value);
	if (env_nbytes > (size_t)(b1end - b1p))
	{
	    env_nbytes = b1end - b1p;
	    memcpy(b1p, env_value, env_nbytes);
	    b1p += env_nbytes;
	    break;
	}
	memcpy(b1p, env_value, env_nbytes);
	b1p += env_nbytes;
    }

    //
    // The return value is the number of bytes in the result buffer.
    //
    return (b1p - buffer1);
}


int
plasticfs_filter_smartlink::open(const char *path, int flags, int mode)
{
    //
    // Try to open the file without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::open(path, flags, mode);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    bool last_file_must_exist = ((flags & O_CREAT) == 0);
    if (!relative_realpath(path, abspath, relpath, last_file_must_exist, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::open(abspath.c_str(), flags, mode);
}


int
plasticfs_filter_smartlink::readlink(const char *path, char *buf, size_t nbytes)
{
    //
    // Try to read the symlink without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::readlink(path, buf, nbytes);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::readlink(abspath.c_str(), buf, nbytes);
}


int
plasticfs_filter_smartlink::lxstat(int vers, const char *path, struct stat *st)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::lxstat(vers, path, st);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, false))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::lxstat(vers, abspath.c_str(), st);
}


int
plasticfs_filter_smartlink::lxstat64(int vers, const char *path,
    struct stat64 *st)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::lxstat64(vers, path, st);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, false))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::lxstat64(vers, abspath.c_str(), st);
}


int
plasticfs_filter_smartlink::xstat(int vers, const char *path, struct stat *st)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::xstat(vers, path, st);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::xstat(vers, abspath.c_str(), st);
}


int
plasticfs_filter_smartlink::xstat64(int vers, const char *path, struct stat64 *st)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::xstat64(vers, path, st);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::xstat64(vers, abspath.c_str(), st);
}


int
plasticfs_filter_smartlink::chmod(const char *path, mode_t mode)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::chmod(path, mode);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::chmod(abspath.c_str(), mode);
}


int
plasticfs_filter_smartlink::mkdir(const char *path, mode_t mode)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::mkdir(path, mode);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::mkdir(abspath.c_str(), mode);
}


int
plasticfs_filter_smartlink::mkfifo(const char *path, mode_t mode)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::mkfifo(path, mode);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::mkfifo(abspath.c_str(), mode);
}


int
plasticfs_filter_smartlink::link(const char *src_path, const char *dest_path)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::link(src_path, dest_path);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring src_abspath;
    rcstring src_relpath;
    if (!relative_realpath(src_path, src_abspath, src_relpath, true, true))
	return -1;
    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::link(src_abspath.c_str(), dest_abspath.c_str());
}


int
plasticfs_filter_smartlink::symlink(const char *text, const char *dest_path)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::symlink(text, dest_path);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::symlink(text, dest_abspath.c_str());
}


int
plasticfs_filter_smartlink::unlink(const char *path)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::unlink(path);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::unlink(abspath.c_str());
}


int
plasticfs_filter_smartlink::rmdir(const char *path)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::rmdir(path);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::rmdir(abspath.c_str());
}


int
plasticfs_filter_smartlink::rename(const char *old_path, const char *new_path)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::rename(old_path, new_path);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring old_abspath;
    rcstring old_relpath;
    if (!relative_realpath(old_path, old_abspath, old_relpath, true, true))
	return -1;
    rcstring new_abspath;
    rcstring new_relpath;
    if (!relative_realpath(new_path, new_abspath, new_relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::rename(old_abspath.c_str(), new_abspath.c_str());
}


long
plasticfs_filter_smartlink::pathconf(const char *path, int name)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::pathconf(path, name);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::pathconf(abspath.c_str(), name);
}


int
plasticfs_filter_smartlink::xmknod(int ver, const char *path, mode_t mode,
    dev_t dev)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::xmknod(ver, path, mode, dev);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::xmknod(ver, abspath.c_str(), mode, dev);
}


int
plasticfs_filter_smartlink::chdir(const char *path)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::chdir(path);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::chdir(abspath.c_str());
}


int
plasticfs_filter_smartlink::utime(const char *path, utimbuf *buf)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::utime(path, buf);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::utime(abspath.c_str(), buf);
}


int
plasticfs_filter_smartlink::lutime(const char *path, utimbuf *buf)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::lutime(path, buf);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::lutime(abspath.c_str(), buf);
}


int
plasticfs_filter_smartlink::execve(const char *path, char *const *argv,
    char *const *envp)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::execve(path, argv, envp);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::execve(abspath.c_str(), argv, envp);
}


int
plasticfs_filter_smartlink::access(const char *path, int type)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::access(path, type);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::access(abspath.c_str(), type);
}


int
plasticfs_filter_smartlink::statfs(const char *path, struct statfs *buf)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::statfs(path, buf);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::statfs(abspath.c_str(), buf);
}


int
plasticfs_filter_smartlink::statfs64(const char *path, struct statfs64 *buf)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::statfs64(path, buf);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::statfs64(abspath.c_str(), buf);
}


int
plasticfs_filter_smartlink::truncate(const char *path, off_t length)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::truncate(path, length);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::truncate(abspath.c_str(), length);
}


int
plasticfs_filter_smartlink::truncate64(const char *path, off64_t length)
{
    //
    // Try the function without doing anything else,
    // because most files aren't going to be smart links.
    //
    int errno_save = errno;
    int result = inherited::truncate64(path, length);
    if (result >= 0 || errno != ENOENT)
	return result;
    errno = errno_save;

    //
    // Resolve all of the symlinks in the path;
    // this does the symlink shell expansions.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;

    //
    // Use the resolved path to try again.
    //
    return inherited::truncate64(abspath.c_str(), length);
}
