//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_filter_viewpath class
//

#pragma implementation "libplasticfs::plasticfs_filter_viewpath"

#include <plasticfs/filter/viewpath.h>
#include <cerrno>
#include <cstring>
#include <errno_insulator.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <input/file.h>
#include <output/file.h>
#include <utime.h>

//
// Define the VIEWPATH_DEBUG symbol when you need to see what methods
// are being called.
//
//#define VIEWPATH_DEBUG

using namespace libplasticfs;


plasticfs_filter_viewpath::~plasticfs_filter_viewpath()
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: destructor\n", __FILE__, __LINE__);
    #endif
}


plasticfs_filter_viewpath::plasticfs_filter_viewpath(plasticfs *deeper,
	const rcstring_list &args) :
    plasticfs_filter(deeper)
{
    if (args.size() < 3)
    {
	error_message
	(
	    "plasticfs: viewpath filter: requires at least two arguments"
	);
    }
    if (args.size() < 2)
	viewpath.push_back(cwd());
    else
    {
	for (size_t j = 1; j < args.size(); ++j)
	    viewpath.push_back(make_absolute(args[j]));
    }
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: constructor \"%s\"\n", __FILE__, __LINE__,
	viewpath[0].c_str());
    #endif
}


plasticfs *
plasticfs_filter_viewpath::factory(plasticfs *deeper, const rcstring_list &args)
{
    return new plasticfs_filter_viewpath(deeper, args);
}


rcstring
plasticfs_filter_viewpath::find(const rcstring &relpath)
{
    struct stat statbuf;
    return find(relpath, statbuf);
}


rcstring
plasticfs_filter_viewpath::find(const rcstring &relpath, struct stat &statbuf)
{
    // make sure we leave errno untouched
    errno_insulator guard;

    //
    // Now locate the file on the viewpath.
    //
    for (size_t j = 0; j < viewpath.size(); ++j)
    {
	rcstring abspath(path_join(viewpath[j], relpath));
	if (inherited::lxstat(_STAT_VER, abspath.c_str(), &statbuf) == 0)
	    return abspath;
    }

    //
    // The file was not found, behave as if it is in the first directory
    // of the viewpath.
    //
    memset(&statbuf, 0, sizeof(statbuf));
    return path_join(viewpath[0], relpath);
}


bool
plasticfs_filter_viewpath::copy(const rcstring &to_fn, const rcstring &from_fn)
{
    long long the_end = ~(unsigned long long)0 >> 1;
    return copy_some(to_fn, from_fn, the_end);
}


bool
plasticfs_filter_viewpath::copy_some(const rcstring &to_fn,
    const rcstring &from_fn, long long length)
{
    int err = errno;

    //
    // open the files
    //
    int from_fd = inherited::open(from_fn.c_str(), O_RDONLY, 0);
    if (from_fd < 0)
	return false;
    struct stat64 st;
    if (inherited::fxstat64(_STAT_VER, from_fd, &st) != 0)
    {
	err = errno;
	inherited::close(from_fd);
	errno = err;
	return false;
    }
    if (st.st_size < length)
	length = st.st_size;

    int to_flags = O_WRONLY | O_CREAT | O_TRUNC;
    int to_mode = (st.st_mode & 07777) | 0666;
    if (to_mode & 0111)
	to_mode |= 0100;
    int to_fd = inherited::open(to_fn.c_str(), to_flags, to_mode);
    if (to_fd < 0)
    {
	err = errno;
	inherited::close(from_fd);
	errno = err;
	return false;
    }

    //
    // copy the data
    //
    long long pos = 0;
    while (pos < length)
    {
	char buffer[4096];
	long long nbytes = length - pos;
	if (nbytes > (long long)sizeof(buffer))
	    nbytes = sizeof(buffer);
	ssize_t rn = inherited::read(from_fd, buffer, size_t(nbytes));
	if (rn == (ssize_t)(-1))
	{
	    err = errno;
	    inherited::close(to_fd);
	    inherited::close(from_fd);
	    inherited::unlink(to_fn.c_str());
	    errno = err;
	    return false;
	}
	if (rn == 0)
	    break;
	pos += rn;
	ssize_t wn = inherited::write(to_fd, buffer, rn);
	if (rn != wn)
	{
	    err = (wn == ((ssize_t)(-1))) ? errno : EINVAL;
	    inherited::close(to_fd);
	    inherited::close(from_fd);
	    inherited::unlink(to_fn.c_str());
	    errno = err;
	    return false;
	}
    }

    //
    // close the files
    //
    if (inherited::close(to_fd) < 0)
    {
	err = errno;
	inherited::close(from_fd);
	inherited::unlink(to_fn.c_str());
	errno = err;
	return false;
    }
    if (inherited::close(from_fd) < 0)
	return false;

    //
    // Update the file mod time and file access time.
    // Ignore any error.
    //
    utimbuf when;
    when.modtime = st.st_mtime;
    when.actime = st.st_atime;
    inherited::utime(to_fn.c_str(), &when);

    //
    // report success
    //
    errno = err;
    return true;
}

#define WHITEOUT_FILENAME ".whiteout"
#define WHITEOUT_NEW WHITEOUT_FILENAME ".tmp"


void
plasticfs_filter_viewpath::whiteout_remove(const rcstring &relpath)
{
    errno_insulator guard;

    rcstring dirname;
    rcstring subject;
    const char *cp = strrchr(relpath.c_str(), '/');
    if (cp)
    {
	dirname = rcstring(relpath.c_str(), cp - relpath.c_str());
	subject = rcstring(cp + 1);
    }
    else
	subject = relpath;
    rcstring absdirname(path_join(viewpath[0], dirname));
    rcstring ifn(path_join(absdirname, rcstring(WHITEOUT_FILENAME)));
    rcstring ofn(path_join(absdirname, rcstring(WHITEOUT_NEW)));

    {
	// Scope the files, so that they are closed by the time we do
	// the rename.
	input_file ip(get_deeper(), ifn);
	output_file op(get_deeper(), ofn);
	for (;;)
	{
	    rcstring temp;
	    if (!ip.readline_c(temp))
		break;
	    if (temp != subject)
	    {
		op.puts_c(temp);
		op.putc('\n');
	    }
	}
    }

    //
    // Now move the new file on top of the old file.
    // Ignore any errors.
    //
    inherited::rename(ofn.c_str(), ifn.c_str());
}


void
plasticfs_filter_viewpath::whiteout_add(const rcstring &relpath)
{
    rcstring dirname;
    rcstring subject;
    const char *cp = strrchr(relpath.c_str(), '/');
    if (cp)
    {
	dirname = rcstring(relpath.c_str(), cp - relpath.c_str());
	subject = rcstring(cp + 1);
    }
    else
	subject = relpath;
    rcstring absdirname(path_join(viewpath[0], dirname));
    rcstring ifn(path_join(absdirname, rcstring(WHITEOUT_FILENAME)));
    rcstring ofn(path_join(absdirname, rcstring(WHITEOUT_NEW)));

    {
	// Scope the files, so that they are closed by the time we do
	// the rename.
	input_file ip(get_deeper(), ifn);
	output_file op(get_deeper(), ofn);
	for (;;)
	{
	    rcstring temp;
	    if (!ip.readline_c(temp))
		break;
	    if (temp != subject)
	    {
		op.puts_c(temp);
		op.putc('\n');
	    }
	}
	op.puts_c(subject);
	op.putc('\n');
    }

    //
    // Now move the new file on top of the old file.
    // Ignore any errors.
    //
    inherited::rename(ofn.c_str(), ifn.c_str());
}


void
plasticfs_filter_viewpath::whiteout_read(file_tracker::active_t *p)
{
    rcstring absdirname(path_join(viewpath[0], p->relpath));
    rcstring ifn(path_join(absdirname, rcstring(WHITEOUT_FILENAME)));

    input_file ip(get_deeper(), ifn);
    for (;;)
    {
	rcstring temp;
	if (!ip.readline_c(temp))
	    break;
	static int dummy = 42;
	p->direntries.assign(temp, &dummy);
    }
}


bool
plasticfs_filter_viewpath::whiteout_present(const rcstring &relpath)
{
    rcstring dirname;
    rcstring subject;
    const char *cp = strrchr(relpath.c_str(), '/');
    if (cp)
    {
	dirname = rcstring(relpath.c_str(), cp - relpath.c_str());
	subject = rcstring(cp + 1);
    }
    else
	subject = relpath;
    rcstring whiteout_filename(path_join(path_join(viewpath[0], dirname),
	rcstring(WHITEOUT_FILENAME)));

    input_file ip(get_deeper(), whiteout_filename);
    for (;;)
    {
	rcstring temp;
	if (!ip.readline_c(temp))
	    return false;
	if (temp == subject)
	    return true;
    }
}


int
plasticfs_filter_viewpath::open(const char *path, int flags, int mode)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: open(\"%s\", %#o, %#o)\n", __FILE__, __LINE__, path,
	flags, mode);
    #endif
    //
    // We only ever modify files in the first directory of the viewpath,
    // so that's where we create them, too.
    //
    if (flags & O_CREAT)
    {
	rcstring abspath;
	rcstring relpath;
	if (!relative_realpath(path, abspath, relpath, false, true))
	    return -1;
	if (abspath == relpath)
	    return inherited::open(abspath.c_str(), flags, mode);
	if (!mkdir_between(relpath))
	    return -1;
	rcstring newpath(path_join(viewpath[0], relpath));
	int result = inherited::open(newpath.c_str(), flags, mode);
	if (result >= 0)
	    whiteout_remove(relpath);
	return result;
    }

    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::open(abspath.c_str(), flags, mode);

    struct stat statbuf;
    rcstring newpath(find(relpath, statbuf));
    switch (flags & O_ACCMODE)
    {
    case O_RDWR:
    case O_WRONLY:
	if (abspath != newpath)
	{
	    // Only copy regular files.
	    if (!S_ISREG(statbuf.st_mode))
		break;

	    if (!mkdir_between(relpath))
		return -1;
	    if (!(flags & O_TRUNC))
	    {
		//
		// need copy to copy deeper files to shallow
		//
		if (!copy(abspath, newpath))
		    return -1;
		newpath = abspath;
	    }
	    else
	    {
		// If we have to shallow the file, we are allowed to create it
		mode |= 0666;
		flags |= O_CREAT;
	    }
	}
	break;

    case O_RDONLY:
	//
	// Make directories shallow, otherwise GNU find 4.1.7 gets upset.
	//
	if (S_ISDIR(statbuf.st_mode))
	{
	    if (!mkdir_between(relpath + "/."))
		return -1;
	    newpath = abspath;
	}
	break;
    }

    int fd = inherited::open(newpath.c_str(), flags, mode);
    if (fd >= 0)
    {
	if (flags & O_CREAT)
	    whiteout_remove(relpath);

	//
	// Keep track of open files below our mount point.
	//
	errno_insulator guard;
	file_tracker::active_t *p =
	    fdtab.open(fd, flags, relpath, get_deeper());
	if (p)
	{
	    //
	    // Read the white-out files, so that deleted files are
	    // invisable.  Do this by pre-loading the direntries table.
	    //
	    whiteout_read(p);
	}
    }
    return fd;
}


bool
plasticfs_filter_viewpath::mkdir_between(const rcstring &relpath)
{
    int err = errno;
    const char *begin = relpath.c_str();
    const char *cp = begin;
    for (;;)
    {
	while (*cp == '/')
	    ++cp;
	const char *ep = strchr(cp, '/');
	if (!ep)
	{
	    errno = err;
	    return true;
	}
	rcstring dir(begin, ep - begin);
	rcstring path(path_join(viewpath[0], dir));
	int n = inherited::mkdir(path.c_str(), 0777);
	if (n >= 0)
	{
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: mkdir(\"%s\")\n", __FILE__, __LINE__,
		path.c_str());
	    #endif
	}
	else if (errno != EEXIST)
	    return false;
	cp = ep + 1;
    }
}


int
plasticfs_filter_viewpath::readlink(const char *path, char *buf, size_t nbytes)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: readlink(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    if (abspath == relpath)
	return inherited::readlink(abspath.c_str(), buf, nbytes);
    return inherited::readlink(find(relpath).c_str(), buf, nbytes);
}


int
plasticfs_filter_viewpath::correct_nlinks(const rcstring &relpath)
{
    // make sure we leave errno untouched
    errno_insulator guard;

    //
    // Now locate each instance of the directory on the viewpath and
    // add the links together.  This may be too many, but it won't be
    // too few, and too few is what makes GNU find 4.1.7 gets upset.
    //
    int nlinks = 2;
    for (size_t j = 0; j < viewpath.size(); ++j)
    {
	rcstring abspath(path_join(viewpath[j], relpath));
	struct stat st;
	if
	(
	    inherited::lxstat(_STAT_VER, abspath.c_str(), &st) == 0
	&&
	    S_ISDIR(st.st_mode)
	)
	    nlinks += st.st_nlink - 2;
    }

    // Return the number of links we found.
    return nlinks;
}


int
plasticfs_filter_viewpath::fxstat(int vers, int fd, struct stat *st)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: fxstat(%d)\n", __FILE__, __LINE__, fd);
    #endif

    if (inherited::fxstat(vers, fd, st) < 0)
	return -1;

    // Directories need special treatment
    // otherwise GNU find 4.1.7 shortcircuits too early
    if (S_ISDIR(st->st_mode))
    {
	file_tracker::active_t *p = fdtab.find(fd);
	if (p)
	    st->st_nlink = correct_nlinks(p->relpath);
    }

    // report success
    return 0;
}


int
plasticfs_filter_viewpath::lxstat(int vers, const char *path, struct stat *st)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: lxstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    if (abspath == relpath)
	return inherited::lxstat(vers, abspath.c_str(), st);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    // Directories need special treatment
    // otherwise GNU find 4.1.7 shortcircuits too early
    rcstring true_path(find(relpath, *st));
    if (S_ISDIR(st->st_mode))
    {
	if (true_path != abspath)
	{
	    if (!mkdir_between(relpath + "/."))
		return -1;
	    if (inherited::lxstat(vers, abspath.c_str(), st) < 0)
		return -1;
	}
	st->st_nlink = correct_nlinks(relpath);
	return 0;
    }

    return inherited::lxstat(vers, true_path.c_str(), st);
}


int
plasticfs_filter_viewpath::lxstat64(int vs, const char *path, struct stat64 *st)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: lxstat64(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    if (abspath == relpath)
	return inherited::lxstat64(vs, abspath.c_str(), st);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    // Directories need special treatment
    // otherwise GNU find 4.1.7 shortcircuits too early
    struct stat st32;
    rcstring true_path(find(relpath, st32));
    if (S_ISDIR(st32.st_mode))
    {
	if (true_path != abspath)
	{
	    if (!mkdir_between(relpath + "/."))
		return -1;
	}
	int result = inherited::lxstat64(vs, abspath.c_str(), st);
	if (result >= 0)
	    st->st_nlink = correct_nlinks(relpath);
	return result;
    }

    return inherited::lxstat64(vs, true_path.c_str(), st);
}


int
plasticfs_filter_viewpath::xstat(int vers, const char *path, struct stat *st)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: xstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    // relative_realpath gets rid of the symlinks, so can use lxstat from now
    if (abspath == relpath)
	return inherited::lxstat(vers, abspath.c_str(), st);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    // Directories need special treatment
    // otherwise GNU find 4.1.7 shortcircuits too early
    rcstring true_path(find(relpath, *st));
    if (S_ISDIR(st->st_mode))
    {
	if (true_path != abspath)
	{
	    if (!mkdir_between(relpath + "/."))
		return -1;
	    if (inherited::lxstat(vers, abspath.c_str(), st) < 0)
		return -1;
	}
	st->st_nlink = correct_nlinks(relpath);
	return 0;
    }

    return inherited::lxstat(vers, find(relpath).c_str(), st);
}


int
plasticfs_filter_viewpath::xstat64(int vers, const char *path, struct stat64 *st)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: xstat64(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    // relative_realpath gets rid of the symlinks, so can use lxstat64 from now
    if (abspath == relpath)
	return inherited::lxstat64(vers, abspath.c_str(), st);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    // Directories need special treatment
    // otherwise GNU find 4.1.7 shortcircuits too early
    struct stat st32;
    rcstring true_path(find(relpath, st32));
    if (S_ISDIR(st32.st_mode))
    {
	if (true_path != abspath)
	{
	    if (!mkdir_between(relpath + "/."))
		return -1;
	}
	int result = inherited::lxstat64(vers, abspath.c_str(), st);
	if (result >= 0)
	    st->st_nlink = correct_nlinks(relpath);
	return result;
    }

    return inherited::lxstat64(vers, true_path.c_str(), st);
}


int
plasticfs_filter_viewpath::relative_lstat(const char *path, struct stat *st)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: relative_lstat(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    int old_errno = errno;
    rcstring relpath(path);

    // make sure the file hasn't been whited out.
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    // look down the view path
    for (size_t j = 0; j < viewpath.size(); ++j)
    {
	rcstring path = path_join(viewpath[j], relpath);
	if (inherited::lxstat(_STAT_VER, path.c_str(), st) == 0)
	{
	    errno = old_errno;
	    return 0;
	}
	if (errno != ENOENT)
	    return -1;
    }
    errno = ENOENT;
    return -1;
}


int
plasticfs_filter_viewpath::relative_readlink(const char *path, char *buffer,
    size_t nbytes)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: relative_readlink(\"%s\")\n", __FILE__, __LINE__,
	path);
    #endif
    int old_errno = errno;
    rcstring relpath(path);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    for (size_t j = 0; j < viewpath.size(); ++j)
    {
	rcstring path = path_join(viewpath[j], relpath);
	int result = inherited::readlink(path.c_str(), buffer, nbytes);
	if (result >= 0)
	{
	    errno = old_errno;
	    return result;
	}
	if (errno != ENOENT)
	    return -1;
    }
    errno = ENOENT;
    return -1;
}


bool
plasticfs_filter_viewpath::is_below_mount_point(const char *dest_buf,
    size_t &nbytes)
{
    return is_below_directory(viewpath[0], dest_buf, nbytes);
}


int
plasticfs_filter_viewpath::close(int fd)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: close(%d)\n", __FILE__, __LINE__, fd);
    #endif

    //
    // If the file descriptor is in our active list, get rid of it and
    // release its resources.
    //
    fdtab.close(fd);

    //
    // We still have to actually close it.
    //
    return inherited::close(fd);
}


int
plasticfs_filter_viewpath::getdents(int fd, dirent *buf, unsigned nbytes)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: getdents(fd=%d, buf=%p, nbytes=%u) {\n", __FILE__,
	__LINE__, fd, buf, nbytes);
    #endif
    int save_errno = errno;

    //
    // See if this file descriptor is in our list.
    //
    file_tracker::active_t *p = fdtab.find(fd);

    //
    // No special processing for file descriptors not in our list.
    //
    if (!p)
    {
	int result = inherited::getdents(fd, buf, nbytes);
	#ifdef VIEWPATH_DEBUG
	error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	#endif
	return result;
    }

    //
    // Scan all of the sub-directories.
    //
    for (;;)
    {
	//
	// Open the sub-directory if necessary.
	// (Report end-of-file if we have run out of sub-directories.)
	//
	if (p->sub_fd < 0)
	{
	    if (p->sub_idx >= viewpath.size())
	    {
		errno = save_errno;
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: return 0 }\n", __FILE__, __LINE__);
		#endif
		return 0;
	    }
	    rcstring abspath(path_join(viewpath[p->sub_idx], p->relpath));
	    p->sub_fd = inherited::open(abspath.c_str(), p->flags, 0);
	    if (p->sub_fd < 0)
	    {
		if (errno != ENOENT)
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: return -1 }\n", __FILE__, __LINE__);
		    #endif
		    return -1;
		}
		p->sub_idx++;
		continue;
	    }
	    p->sub_filepos = 0;
	}

	//
	// Read the directory contents.
	// If we see end-of-file, go on to the next directory.
	//
	int n = inherited::getdents(p->sub_fd, buf, nbytes);
	if (n < 0)
	{
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
	if (n == 0)
	{
	    if (p->sub_fd >= 0)
	    {
		// ignore any errors
		inherited::close(p->sub_fd);
		p->sub_fd = -1;
	    }
	    p->sub_idx++;
	    continue;
	}

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < n)
	{
	    dirent *dp = (dirent *)((char *)buf + pos);
	    if (dp->d_ino)
	    {
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
                rcstring ent(dp->d_name);
                if (p->direntries.query(ent))
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}
	p->filepos += n;

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
    }
}


int
plasticfs_filter_viewpath::getdents64(int fd, dirent64 *buf, unsigned nbytes)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: getdents(fd=%d, buf=%p, nbytes=%u) {\n", __FILE__,
	__LINE__, fd, buf, nbytes);
    #endif
    int save_errno = errno;

    //
    // See if this file descriptor is in our list.
    //
    file_tracker::active_t *p = fdtab.find(fd);

    //
    // No special processing for file descriptors not in our list.
    //
    if (!p)
    {
	int result = inherited::getdents64(fd, buf, nbytes);
	#ifdef VIEWPATH_DEBUG
	error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	#endif
	return result;
    }

    //
    // Scan all of the sub-directories.
    //
    for (;;)
    {
	//
	// Open the sub-directory if necessary.
	// (Report end-of-file if we have run out of sub-directories.)
	//
	if (p->sub_fd < 0)
	{
	    if (p->sub_idx >= viewpath.size())
	    {
		errno = save_errno;
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: return 0 }\n", __FILE__, __LINE__);
		#endif
		return 0;
	    }
	    rcstring abspath(path_join(viewpath[p->sub_idx], p->relpath));
	    p->sub_fd = inherited::open(abspath.c_str(), p->flags, 0);
	    if (p->sub_fd < 0)
	    {
		if (errno != ENOENT)
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: return -1 }\n", __FILE__, __LINE__);
		    #endif
		    return -1;
		}
		p->sub_idx++;
		continue;
	    }
	    p->sub_filepos = 0;
	}

	//
	// Read the directory contents.
	// If we see end-of-file, go on to the next directory.
	//
	int n = inherited::getdents64(p->sub_fd, buf, nbytes);
	if (n < 0)
	{
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
	if (n == 0)
	{
	    if (p->sub_fd >= 0)
	    {
		// ignore any errors
		inherited::close(p->sub_fd);
		p->sub_fd = -1;
	    }
	    p->sub_idx++;
	    continue;
	}

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < n)
	{
	    dirent64 *dp = (dirent64 *)((char *)buf + pos);
	    if (dp->d_ino)
	    {
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
	        if (p->direntries.query(ent))
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}
	p->filepos += n;

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
    }
}


ssize_t
plasticfs_filter_viewpath::getdirentries(int fd, char *buf, size_t nbytes,
    off_t *basep)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: getdirentries(fd=%d, buf=%p, nbytes=%ld, "
	"*basep=%ld) {\n", __FILE__, __LINE__, fd, buf, (long)nbytes,
	(long)*basep);
    #endif
    int save_errno = errno;

    //
    // See if this file descriptor is in our list.
    //
    file_tracker::active_t *p = fdtab.find(fd);

    //
    // No special processing for file descriptors not in our list.
    //
    if (!p)
    {
	int result = inherited::getdirentries(fd, buf, nbytes, basep);
	#ifdef VIEWPATH_DEBUG
	error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	#endif
	return result;
    }

    //
    // Watch out for them rewinding the directory to the start.
    //
    if (*basep != p->filepos)
    {
	if (*basep)
	{
	    errno = EINVAL;
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: (*basep=%ld) != (p->filepos=%ld), so return "
		"EINVAL }\n", __FILE__, __LINE__, (long)*basep,
		(long)p->filepos);
	    #endif
	    return -1;
	}
	if (p->sub_fd >= 0)
	{
	    inherited::close(p->sub_fd);
	    p->sub_fd = -1;
	}
	p->sub_idx = 0;
	p->filepos = 0;
    }

    //
    // Scan all of the sub-directories.
    //
    for (;;)
    {
	//
	// Open the sub-directory if necessary.
	// (Report end-of-file if we have run out of sub-directories.)
	//
	if (p->sub_fd < 0)
	{
	    if (p->sub_idx >= viewpath.size())
	    {
		errno = save_errno;
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: return 0 }\n", __FILE__, __LINE__);
		#endif
		return 0;
	    }
	    rcstring abspath(path_join(viewpath[p->sub_idx], p->relpath));
	    p->sub_fd = inherited::open(abspath.c_str(), p->flags, 0);
	    if (p->sub_fd < 0)
	    {
		if (errno != ENOENT)
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: return -1 }\n", __FILE__, __LINE__);
		    #endif
		    return -1;
		}
		p->sub_idx++;
		continue;
	    }
	    p->sub_filepos = 0;
	}

	//
	// Read the directory contents.
	// If we see end-of-file, go on to the next directory.
	//
	off_t base = p->sub_filepos;
	int n = inherited::getdirentries(p->sub_fd, buf, nbytes, &base);
	if (n < 0)
	{
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
	if (n == 0)
	{
	    if (p->sub_fd >= 0)
	    {
		// ignore any errors
		inherited::close(p->sub_fd);
		p->sub_fd = -1;
	    }
	    p->sub_idx++;
	    continue;
	}
	p->sub_filepos = base;

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < n)
	{
	    dirent *dp = (dirent *)(buf + pos);
	    if (dp->d_ino)
	    {
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
                if (p->direntries.query(ent))
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: duplicate\n", __FILE__, __LINE__);
		    #endif
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}
	p->filepos += n;

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    *basep = p->filepos;
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
    }
}


ssize_t
plasticfs_filter_viewpath::getdirentries64(int fd, char *buf, size_t nbytes,
    off64_t *basep)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: getdirentries64(fd=%d, buf=%p, nbytes=%ld, "
	"*basep=%ld) {\n", __FILE__, __LINE__, fd, buf, (long)nbytes,
	(long)*basep);
    #endif
    int save_errno = errno;

    //
    // See if this file descriptor is in out list.
    //
    file_tracker::active_t *p = fdtab.find(fd);

    //
    // No special processing for file descriptors not in our list.
    //
    if (!p)
    {
	int result = inherited::getdirentries64(fd, buf, nbytes, basep);
	#ifdef VIEWPATH_DEBUG
	error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
	#endif
	return result;
    }

    //
    // Watch out for them rewinding the directory to the start.
    //
    if (*basep != p->filepos)
    {
	if (*basep)
	{
	    errno = EINVAL;
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return EINVAL }\n", __FILE__, __LINE__);
	    #endif
	    return -1;
	}
	if (p->sub_fd >= 0)
	{
	    inherited::close(p->sub_fd);
	    p->sub_fd = -1;
	}
	p->sub_idx = 0;
	p->filepos = 0;
    }

    //
    // Scan all of the sub-directories.
    //
    for (;;)
    {
	//
	// Open the sub-directory if necessary.
	// (Report end-of-file if we have run out of sub-directories.)
	//
	if (p->sub_fd < 0)
	{
	    if (p->sub_idx >= viewpath.size())
	    {
		errno = save_errno;
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: return 0 }\n", __FILE__, __LINE__);
		#endif
		return 0;
	    }
	    rcstring abspath(path_join(viewpath[p->sub_idx], p->relpath));
	    p->sub_fd = inherited::open(abspath.c_str(), p->flags, 0);
	    if (p->sub_fd < 0)
	    {
		if (errno != ENOENT)
		{
		    #ifdef VIEWPATH_DEBUG
		    error_message("%s: %d: return -1 }\n", __FILE__, __LINE__);
		    #endif
		    return -1;
		}
		p->sub_idx++;
		continue;
	    }
	    p->sub_filepos = 0;
	}

	//
	// Read the directory contents.
	// If we see end-of-file, go on to the next directory.
	//
	off64_t base = p->sub_filepos;
	int n = inherited::getdirentries64(p->sub_fd, buf, nbytes, &base);
	if (n < 0)
	{
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
	if (n == 0)
	{
	    if (p->sub_fd >= 0)
	    {
		// ignore any errors
		inherited::close(p->sub_fd);
		p->sub_fd = -1;
	    }
	    p->sub_idx++;
	    continue;
	}
	p->sub_filepos = base;

	//
	// Stash information about the existence of each directory entry,
	// and use this existence information to get rid of duplicates.
	//
	bool something = false;
	int pos = 0;
	while (pos < n)
	{
	    dirent64 *dp = (dirent64 *)(buf + pos);
	    if (dp->d_ino)
	    {
		#ifdef VIEWPATH_DEBUG
		error_message("%s: %d: d_name=\"%s\"\n", __FILE__, __LINE__,
		    dp->d_name);
		#endif
		rcstring ent(dp->d_name);
                if (p->direntries.query(ent))
		{
		    // suppress duplicate names
		    dp->d_type = DT_UNKNOWN;
		    dp->d_ino = 0;
		}
		else
		{
		    static int bogus;
		    p->direntries.assign(ent, &bogus);
		    something = true;
		}
	    }
	    pos += dp->d_reclen;
	}
	p->filepos += n;

	if (something)
	{
	    //
	    // Report success.
	    //
	    errno = save_errno;
	    *basep = p->filepos;
	    #ifdef VIEWPATH_DEBUG
	    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, n);
	    #endif
	    return n;
	}
    }
}


int
plasticfs_filter_viewpath::chmod(const char *path, mode_t mode)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::chmod(abspath.c_str(), mode);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    rcstring old_path(find(relpath));
    if (abspath != old_path)
    {
	if (!mkdir_between(relpath))
	    return -1;
	if (!copy(abspath, old_path))
	    return -1;
    }
    return inherited::chmod(abspath.c_str(), mode);
}


int
plasticfs_filter_viewpath::mkdir(const char *path, mode_t mode)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;
    if (abspath == relpath)
	return inherited::mkdir(abspath.c_str(), mode);

    if (!mkdir_between(relpath))
	return -1;
    int result = inherited::mkdir(abspath.c_str(), mode);
    if (result >= 0)
	whiteout_remove(relpath);
    return result;
}


int
plasticfs_filter_viewpath::mkfifo(const char *path, mode_t mode)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;
    if (abspath == relpath)
	return inherited::mkfifo(abspath.c_str(), mode);

    if (!mkdir_between(relpath))
	return -1;
    int result = inherited::mkfifo(abspath.c_str(), mode);
    if (result >= 0)
	whiteout_remove(relpath);
    return result;
}


int
plasticfs_filter_viewpath::link(const char *src_path, const char *dest_path)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring src_abspath;
    rcstring src_relpath;
    if (!relative_realpath(src_path, src_abspath, src_relpath, true, true))
	return -1;

    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;

    if (src_abspath == src_relpath)
    {
	//
	// The source file is outside the mount mount.
	//
	if (dest_abspath == dest_relpath)
	{
	    //
	    // The source file is outside the mount mount.
	    // The destination file is outside the mount point.
	    //
	    return inherited::link(src_abspath.c_str(), dest_abspath.c_str());
	}
	else
	{
	    //
	    // The source file is outside the mount mount.
	    // The destination file is below the mount mount.
	    //
	    // In this case, link the file into the first directory in
	    // the view path, where we create all files.
	    //
	    if (!mkdir_between(dest_relpath))
		return -1;
	    int result =
		inherited::link(src_abspath.c_str(), dest_abspath.c_str());
	    if (result >= 0)
		whiteout_remove(dest_relpath);
	    return result;
	}
    }
    else
    {
	//
	// The source file is below the mount mount.
	//
	if (whiteout_present(src_relpath))
	{
	    errno = ENOENT;
	    return -1;
	}

	if (dest_abspath == dest_relpath)
	{
	    //
	    // The source file is below the mount mount.
	    // The destination file is outside the mount point.
	    //
	    // In this case, find the real file within the viewpath and
	    // do the link that way.
	    //
	    return
		inherited::link
		(
		    find(src_relpath).c_str(),
		    dest_abspath.c_str()
		);
	}
	else
	{
	    //
	    // The source file is below the mount mount.
	    // The destination file is below the mount mount.
	    //
	    // In this case, find the real file within the viewpath and
	    // do the link that way.
	    //
	    // In this case, link the file into the first directory in
	    // the view path, where we create all files.
	    //
	    if (!mkdir_between(dest_relpath))
		return -1;
	    int result =
		inherited::link
		(
		    find(src_relpath).c_str(),
		    dest_abspath.c_str()
		);
	    if (result >= 0)
		whiteout_remove(dest_relpath);
	    return result;
	}
    }
}


int
plasticfs_filter_viewpath::symlink(const char *text, const char *dest_path)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring dest_abspath;
    rcstring dest_relpath;
    if (!relative_realpath(dest_path, dest_abspath, dest_relpath, false, true))
	return -1;

    if (dest_abspath == dest_relpath)
    {
	//
	// The destination file is outside the mount point.
	//
	return inherited::symlink(text, dest_abspath.c_str());
    }
    else
    {
	//
	// The destination file is below the mount mount.
	//
	// In this case, link the file into the first directory in
	// the view path, where we create all files.
	//
	if (!mkdir_between(dest_relpath))
	    return -1;
	int result = inherited::symlink(text, dest_abspath.c_str());
	if (result >= 0)
	    whiteout_remove(dest_relpath);
	return result;
    }
}


int
plasticfs_filter_viewpath::unlink(const char *path)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    if (abspath == relpath)
	return inherited::unlink(abspath.c_str());
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    int result = 0;
    if (find(relpath) == abspath)
	result = inherited::unlink(abspath.c_str());
    else
    {
	// Check that it is not a directory.
	struct stat st;
	if (relative_lstat(relpath.c_str(), &st) == 0 && S_ISDIR(st.st_mode))
	{
	    errno = EISDIR;
	    return -1;
	}
    }
    if (result >= 0)
    {
	if (!mkdir_between(relpath))
	    return -1;
	whiteout_add(relpath);
    }
    return result;
}

dirent64*
plasticfs_filter_viewpath::readdir64(DIR *dir)
{
    dirent64 *dp = NULL;

    if (!dir)
        return NULL;

    //
    // Skip already removed entries, skip also .whiteout
    //
    while ((dp = inherited::readdir64(dir)))
    {
        if (0 == strcmp(dp->d_name, WHITEOUT_FILENAME))
            continue;
        if (0 == strcmp(dp->d_name, WHITEOUT_NEW))
            continue;
        if (!whiteout_present(rcstring(dp->d_name)))
            break;
    }
    return dp;

}

dirent*
plasticfs_filter_viewpath::readdir(DIR *dir)
{
    dirent *dp = NULL;

    if (!dir)
        return NULL;

    //
    // Skip already removed entries, skip also .whiteout
    //
    while ((dp = inherited::readdir(dir)))
    {
        if (0 == strcmp(dp->d_name, WHITEOUT_FILENAME))
            continue;
        if (0 == strcmp(dp->d_name, WHITEOUT_NEW))
            continue;
        if (!whiteout_present(rcstring(dp->d_name)))
            break;
    }
    return dp;
}

int
plasticfs_filter_viewpath::rmdir(const char *path)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
        return inherited::rmdir(abspath.c_str());
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    //
    // Check that each of the corresponding directories in each
    // element of the view path is empty.
    //
    int number_of_directory_entries = 0;
    DIR *dirp = opendir(abspath.c_str());
    if (dirp)
    {
	for (;;)
	{
	    dirent *dp = inherited::readdir(dirp);
	    if (!dp)
		break;
	    if
	    (
		dp->d_name[0] == '.'
	    &&
		(
		    dp->d_name[1] == '\0'
		||
		    (dp->d_name[1] == '.' && dp->d_name[2] == '\0')
		)
	    )
		continue;
	    if (0 == strcmp(dp->d_name, WHITEOUT_FILENAME))
		continue;
	    if (0 == strcmp(dp->d_name, WHITEOUT_NEW))
		continue;
            ++number_of_directory_entries;
	}
	closedir(dirp);
	dirp = 0;
    }
    if (number_of_directory_entries != 0)
    {
	errno = ENOTEMPTY;
	return -1;
    }

    int result = 0;
    if (find(relpath) == abspath)
    {
	//
	// There could be a .whiteout file left in the directory,
	// so remove it.  Ignore any error.
	//
	inherited::unlink(path_join(abspath, rcstring(WHITEOUT_FILENAME)).c_str());
	inherited::unlink(path_join(abspath, rcstring(WHITEOUT_NEW)).c_str());

	//
	// Now remove the top-level directory.
	//
	result = inherited::rmdir(abspath.c_str());
    }
    else
    {
	// Check that it is a directory.
	struct stat st;
	if (relative_lstat(relpath.c_str(), &st) == 0 && !S_ISDIR(st.st_mode))
	{
	    errno = ENOTDIR;
	    return -1;
	}
    }
    if (result >= 0)
    {
	if (!mkdir_between(relpath))
	    return -1;
	whiteout_add(relpath);
    }
    return result;
}


int
plasticfs_filter_viewpath::rename(const char *old_path, const char *new_path)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    //
    // We implement rename as link followed by unlink.  This has race
    // conditions, but it makes the semantics WAY more consistent.
    //
    // Also, some of the four cases can only be implemented this way
    // (and the exponential explosion of error cases got very messy).
    //
    // We use unqualified "link" and "unlink" method names, so they
    // redirect through this class' methods.
    //
    // THIS DOES NOT WORK FOR DIRECTORIES, but then, nothing ever was,
    // without a huge amount of machinery to support it.  So the link()
    // call will return a suitable error and we can stop worrying.
    //
    int result = unlink(new_path);
    if (result < 0 && errno == ENOENT)
	result = 0;
    if (result < 0)
	return result;
    result = link(old_path, new_path);
    if (result < 0)
	return result;
    result = unlink(old_path);
    if (result < 0)
    {
	errno_insulator guard;
	unlink(new_path);
    }
    return result;
}


long
plasticfs_filter_viewpath::pathconf(const char *path, int name)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::pathconf(abspath.c_str(), name);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    return inherited::pathconf(find(relpath).c_str(), name);
}


int
plasticfs_filter_viewpath::xmknod(int ver, const char *path, mode_t mode,
    dev_t dev)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, false, true))
	return -1;
    if (abspath == relpath)
	return inherited::xmknod(ver, abspath.c_str(), mode, dev);
    int result = inherited::xmknod(ver, abspath.c_str(), mode, dev);
    if (result >= 0)
	whiteout_remove(relpath);
    return result;
}


int
plasticfs_filter_viewpath::fcntl(int fd, int cmd, void *arg)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: fcntl(%d, %d, %ld) {\n", __FILE__, __LINE__, fd, cmd,
	(long)arg);
    #endif
    int result = inherited::fcntl(fd, cmd, arg);
    if (cmd == F_DUPFD && result >= 0)
	fdtab.dup(fd, result);
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
    #endif
    return result;
}


int
plasticfs_filter_viewpath::fcntl64(int fd, int cmd, void *arg)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: fcntl64(%d, %d) {\n", __FILE__, __LINE__, fd, cmd);
    #endif
    int result = inherited::fcntl64(fd, cmd, arg);
    if (cmd == F_DUPFD && result >= 0)
	fdtab.dup(fd, result);
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: return %d }\n", __FILE__, __LINE__, result);
    #endif
    return result;
}


int
plasticfs_filter_viewpath::dup(int fd)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    int result = inherited::dup(fd);
    if (result >= 0)
	fdtab.dup(fd, result);
    return result;
}


int
plasticfs_filter_viewpath::dup2(int oldfd, int newfd)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    int result = inherited::dup2(oldfd, newfd);
    if (result >= 0)
	fdtab.dup(oldfd, result);
    return result;
}


int
plasticfs_filter_viewpath::chdir(const char *path)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: chdir(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::chdir(abspath.c_str());

    //
    // The directory may only exist in a deeper directory, so we need
    // to make sure there is a directory to go to.
    //
    if (!mkdir_between(relpath + rcstring("/.")))
	return -1;

    //
    // Now we can go there.
    //
    return inherited::chdir(abspath.c_str());
}


int
plasticfs_filter_viewpath::fchdir(int fd)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: fchdir(%d)\n", __FILE__, __LINE__, fd);
    #endif
    file_tracker::active_t *p = fdtab.find(fd);
    if (!p)
	return inherited::fchdir(fd);

    //
    // The directory may only exist in a deeper directory, so we need
    // to make sure there is a directory to go to.
    //
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: this is \"%s\"\n", __FILE__, __LINE__, p->relpath.c_str());
    #endif
    if (!mkdir_between(p->relpath + rcstring("/.")))
	return -1;

    //
    // Now we can go there.
    //
    return inherited::chdir(path_join(viewpath[0], p->relpath).c_str());
}


int
plasticfs_filter_viewpath::utime(const char *path, utimbuf *buf)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::utime(abspath.c_str(), buf);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    rcstring old_path(find(relpath));
    if (abspath != old_path)
    {
	if (!mkdir_between(relpath))
	    return -1;
	if (!copy(abspath, old_path))
	    return -1;
    }
    return inherited::utime(abspath.c_str(), buf);
}


int
plasticfs_filter_viewpath::lutime(const char *path, utimbuf *buf)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: %s\n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, false))
	return -1;
    if (abspath == relpath)
	return inherited::lutime(abspath.c_str(), buf);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    rcstring old_path(find(relpath));
    if (abspath != old_path)
    {
	if (!mkdir_between(relpath))
	    return -1;
	if (!copy(abspath, old_path))
	    return -1;
    }
    return inherited::lutime(abspath.c_str(), buf);
}


int
plasticfs_filter_viewpath::execve(const char *path, char *const *argv,
    char *const *envp)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: execve(\"%s\")\n", __FILE__, __LINE__, path);
    #endif
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::execve(abspath.c_str(), argv, envp);
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }
    return inherited::execve(find(relpath).c_str(), argv, envp);
}


int
plasticfs_filter_viewpath::access(const char *path, int type)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: access(\"%s\", %d)\n", __FILE__, __LINE__, path,
	type);
    #endif
    //
    // If the file isn't below our mount point, call the deeper filter.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::access(abspath.c_str(), type);

    //
    // If whiteout is present, say the file is gone.
    //
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    //
    // If they aren't asking to be able to write the file,
    // check the real file.
    //
    rcstring old_path(find(relpath));
    if (!(type & W_OK))
	return inherited::access(old_path.c_str(), type);

    //
    // If the file is not in the first directory of the viewpath
    // we need to copy the file so that it is writable.
    //
    if (abspath != old_path)
    {
	if (!mkdir_between(relpath))
	    return -1;
	if (!copy(abspath, old_path))
	    return -1;
    }

    //
    // Now check the file in the first directory of the view path.
    //
    return inherited::access(abspath.c_str(), type);
}


int
plasticfs_filter_viewpath::statfs(const char *path, struct statfs *buf)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: statfs(\"%s\")\n", __FILE__, __LINE__, path);
    #endif

    //
    // If the file isn't below our mount point, call the deeper filter.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::statfs(abspath.c_str(), buf);

    //
    // If whiteout is present, say the file is gone.
    //
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    //
    // Now check the real file.
    //
    return inherited::statfs(find(relpath).c_str(), buf);
}


int
plasticfs_filter_viewpath::statfs64(const char *path, struct statfs64 *buf)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: statfs64(\"%s\")\n", __FILE__, __LINE__, path);
    #endif

    //
    // If the file isn't below our mount point, call the deeper filter.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::statfs64(abspath.c_str(), buf);

    //
    // If whiteout is present, say the file is gone.
    //
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    //
    // Now check the real file.
    //
    return inherited::statfs64(find(relpath).c_str(), buf);
}


int
plasticfs_filter_viewpath::truncate(const char *path, off_t length)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: truncate(\"%s\", %ld)\n", __FILE__, __LINE__, path,
	(long)length);
    #endif

    //
    // If the file isn't below our mount point, call the deeper filter.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::truncate(abspath.c_str(), length);

    //
    // If whiteout is present, say the file is gone.
    //
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    //
    // If the file is deeper, copy it short.
    //
    rcstring truepath(find(relpath));
    if (truepath != abspath)
    {
	if (!copy_some(abspath, truepath, length))
	    return -1;
	return 0;
    }

    //
    // Now check the real file.
    //
    return inherited::truncate(abspath.c_str(), length);
}


int
plasticfs_filter_viewpath::truncate64(const char *path, off64_t length)
{
    #ifdef VIEWPATH_DEBUG
    error_message("%s: %d: truncate64(\"%s\", %ld)\n", __FILE__, __LINE__, path,
	(long)length);
    #endif

    //
    // If the file isn't below our mount point, call the deeper filter.
    //
    rcstring abspath;
    rcstring relpath;
    if (!relative_realpath(path, abspath, relpath, true, true))
	return -1;
    if (abspath == relpath)
	return inherited::truncate64(abspath.c_str(), length);

    //
    // If whiteout is present, say the file is gone.
    //
    if (whiteout_present(relpath))
    {
	errno = ENOENT;
	return -1;
    }

    //
    // If the file is deeper, copy it short.
    //
    rcstring truepath(find(relpath));
    if (truepath != abspath)
    {
	if (!copy_some(abspath, truepath, length))
	    return -1;
	return 0;
    }

    //
    // Now check the real file.
    //
    return inherited::truncate64(abspath.c_str(), length);
}
