//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/plasticfs/filter/viewpath.cc
//

#ifndef LIB_PLASTICFS_FILTER_VIEWPATH_H
#define LIB_PLASTICFS_FILTER_VIEWPATH_H

#pragma interface "libplasticfs::plasticfs_filter_viewpath"

#include <plasticfs/filter.h>
#include <file_tracker.h>
#include <rcstring_list.h>

namespace libplasticfs {

/**
  * The plasticfs_filter_viewpath class is used to represent a filesystem
  * where a list of directories form a view path.  Files outside this
  * view path are unaltered.
  */
class plasticfs_filter_viewpath:
    public plasticfs_filter
{
public:
    /**
      * The destructor.
      */
    virtual ~plasticfs_filter_viewpath();

    /**
      * The constructor.
      */
    plasticfs_filter_viewpath(plasticfs *deeper, const rcstring_list &args);

    // See base class for documentation.
    int open(const char *path, int flags, int mode);

    // See base class for documentation.
    int close(int);

    // See base class for documentation.
    int getdents(int, dirent *, unsigned);

    // See base class for documentation.
    int getdents64(int, dirent64 *, unsigned);

    // See base class for documentation.
    ssize_t getdirentries(int fd, char *buf, size_t nbytes, off_t *basep);

    // See base class for documentation.
    ssize_t getdirentries64(int fd, char *buf, size_t nbytes, off64_t *basep);

    // See base class for documentation.
    int fxstat(int, int, struct stat *);

    // See base class for documentation.
    int xstat(int, const char *, struct stat *);

    // See base class for documentation.
    int xstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int lxstat(int, const char *, struct stat *);

    // See base class for documentation.
    int lxstat64(int, const char *, struct stat64 *);

    // See base class for documentation.
    int readlink(const char *, char *, size_t);

    // See base class for documentation.
    int chmod(const char *, mode_t);

    // See base class for documentation.
    int mkdir(const char *, mode_t);

    // See base class for documentation.
    int mkfifo(const char *, mode_t);

    // See base class for documentation.
    int unlink(const char *);

    // See base class for documentation.
    int rename(const char *, const char *);

    // See base class for documentation.
    int link(const char *, const char *);

    // See base class for documentation.
    long pathconf(const char *, int);

    // See base class for documentation.
    int xmknod(int ver, const char *path, mode_t mode, dev_t dev);

    // See base class for documentation.
    int fcntl(int, int, void *);

    // See base class for documentation.
    int fcntl64(int, int, void *);

    // See base class for documentation.
    int dup(int);

    // See base class for documentation.
    int dup2(int, int);

    // See base class for documentation.
    int symlink(const char *, const char *);

    // See base class for documentation.
    int chdir(const char *);

    // See base class for documentation.
    int fchdir(int);

    // See base class for documentation.
    dirent64 *readdir64(DIR *);

    // See base class for documentation.
    dirent *readdir(DIR *);

    // See base class for documentation.
    int rmdir(const char *);

    // See base class for documentation.
    int utime(const char *, utimbuf *);

    // See base class for documentation.
    int lutime(const char *, utimbuf *);

    // See base class for documentation.
    int execve(const char *, char *const *, char *const *);

    // See base class for documentation.
    int access(const char *, int);

    // See base class for documentation.
    int statfs(const char *, struct statfs *);

    // See base class for documentation.
    int statfs64(const char *, struct statfs64 *);

    // See base class for documentation.
    int truncate(const char *, off_t);

    // See base class for documentation.
    int truncate64(const char *, off64_t);

    /**
      * The factory method is used to produce new
      * plasticfs_filter_viewpath instances.
      */
    static plasticfs *factory(plasticfs *deeper, const rcstring_list &args);

private:
    typedef plasticfs_filter inherited;

    /**
      * The fdtab instance variable is used to remember open file
      * descriptors and associated data.
      */
    file_tracker fdtab;

    /**
      * The active table is indexed by the file descriptor.  The O(1)
      * lookup means that there isn't a significant performance penalty
      * for doing this.
      */
    itab active;

    // See base class for doumentation.
    bool is_below_mount_point(const char *, size_t &);

    // See base class for documentation.
    int relative_readlink(const char *, char *, size_t);

    /**
      * The mkdir_between method takes a relative path (no leading slash)
      * and creates directories for all leading path portions (if the
      * don't exist already).
      */
    bool mkdir_between(const rcstring &relpath);

    /**
      * The find method is used to find a file along the viewpath.
      * A relative path is given, and an absolute path is returned.
      */
    rcstring find(const rcstring &);

    /**
      * The find method is used to find a file along the viewpath.
      * A relative path is given, and an absolute path is returned,
      * and also information about the file itself.
      */
    rcstring find(const rcstring &, struct stat &);

    /**
      * The copy method is used to copy the `from' to to the `to' file.
      */
    bool copy(const rcstring &to, const rcstring &from);

    /**
      * The copy method is used to copy the `from' to to the `to' file,
      * limited to the first `length' bytes.
      */
    bool copy_some(const rcstring &to, const rcstring &from, long long length);

    /**
      * The viewpath instance variable is used to remember the list of
      * directories to be searched.
      */
    rcstring_list viewpath;

    // See base class for documentation.
    int relative_lstat(const char *, struct stat *);

    /**
      * The whiteout_add method is use to add white space for the given
      * relative path.  It isn't an error if there is whiteout already
      * present.
      */
    void whiteout_add(const rcstring &relpath);

    /**
      * The whiteout_remove method is use to remove white space from
      * the given relative path.  It isn't an error if there was no
      * whiteout to remove.
      */
    void whiteout_remove(const rcstring &relpath);

    /**
      * The whiteout_read method is used to read the whiteout names
      * into the symbol table for the given directory, in preparation
      * for getdirentries.
      */
    void whiteout_read(file_tracker::active_t *);

    /**
      * The whiteout_present method is used to determine if whiteout is
      * present for the given relative path.
      */
    bool whiteout_present(const rcstring &relpath);

    /**
      * The correct_nlinks method is used to generate a (more) correct
      * value for st_nlinks for directories.
      */
    int correct_nlinks(const rcstring &relpath);

    /**
      * The default constructor.  Do not use.
      */
    plasticfs_filter_viewpath();

    /**
      * The copy constructor.  Do not use.
      */
    plasticfs_filter_viewpath(const plasticfs_filter_viewpath &);

    /**
      * The assignment operator.  Do not use.
      */
    plasticfs_filter_viewpath &operator=(const plasticfs_filter_viewpath &);
};

} // namespace libplasticfs

#endif // LIB_PLASTICFS_FILTER_VIEWPATH_H
