//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_getcwd class
//

#pragma implementation "libplasticfs::plasticfs_getcwd"

#include <plasticfs.h>
#include <cstring>
#include <cstdlib>

using namespace libplasticfs;

//
// This file is derived from
// GNU Glibc 2.3  sysdeps/posix/getcwd.c
// Copyright (C) 1991-1999 Free Software Foundation, Inc.
//

#include <cerrno>
#include <cstdlib>
#include <sys/types.h>
#include <sys/stat.h>

//#define GETCWD_DEBUG

#ifndef MAX
# define MAX(a, b) ((a) < (b) ? (b) : (a))
#endif

#include <limits.h>

#ifndef PATH_MAX
# ifdef	MAXPATHLEN
#  define PATH_MAX MAXPATHLEN
# else
#  define PATH_MAX 1024
# endif
#endif

//
// Get the pathname of the current working directory, and put it in SIZE
// bytes of BUF.  Returns NULL if the directory couldn't be determined or
// SIZE was too small.  If successful, returns BUF.  In GNU, if BUF is
// NULL, an array is allocated with `malloc'; the array is SIZE bytes long,
// unless SIZE == 0, in which case it is as big as necessary.
//
// This virtual method calls the open, getdents and close virtual methods.
// Thus, if you have correct implementations of these methods, this
// getcwd implementation will report a filtered representation of the
// current directory name.
//
char *
plasticfs::getcwd(char *buf, long size)
{
    #ifdef GETCWD_DEBUG
    error_message("%s: %d: getcwd(buf = %p, size = %ld) {\n", __FILE__,
	__LINE__, buf, size);
    error_message("%s: %d: this=%p\n", __FILE__, __LINE__, this);
    #endif

    static const char dots[] =
	"../../../../../../../../../../../../../../../../../../../../../../../"
        "../../../../../../../../../../../../../../../../../../../../../../../"
	"../../../../../../../../../../../../../../../../../../../../../../../"
	"../../../../../..";
    const char *dotp = &dots[sizeof(dots)];
    const char *dotlist = dots;

    // No matter how this function exits, always free dotlist (if it
    // needs it) on the way out.  It's even exception safe.
    class if_dotlist_ne_dots_free_dotlist
    {
    private:
	const char *dots;
	const char *volatile &dotlist;
    public:
	if_dotlist_ne_dots_free_dotlist(const char *&a2):
		dots(a2), dotlist(a2) { }
	~if_dotlist_ne_dots_free_dotlist()
	    { if (dotlist != dots && dotlist) free((void *)dotlist); }
    };
    if_dotlist_ne_dots_free_dotlist dotlist_guard(dotlist);


    size_t dotsize = sizeof(dots) - 1;
    int prev_errno = errno;
    size_t allocated = size;

    if (size == 0)
    {
	if (buf)
	{
	    errno = EINVAL;
	    #ifdef GETCWD_DEBUG
	    error_message("%s: %d: return NULL }\n", __FILE__, __LINE__);
	    #endif
	    return 0;
	}

	allocated = PATH_MAX + 1;
    }

    char *path = buf;
    if (!path)
    {
	path = (char *)malloc(allocated);
	if (!path)
	{
	    #ifdef GETCWD_DEBUG
	    error_message("%s: %d: return NULL }\n", __FILE__, __LINE__);
	    #endif
	    return 0;
	}
    }

    // No matter how this function exits, always free path (if it needs
    // it) on the way out.  It's even exception safe.
    class if_not_buf_free_path
    {
    private:
	const char *buf;
	char *volatile &path;
    public:
	if_not_buf_free_path(const char *a1, char *&a2):
	    buf(a1), path(a2) { }
	~if_not_buf_free_path()
	    { if (!buf && path) free(path); }
    };
    if_not_buf_free_path path_guard(buf, path);

    char *pathp = path + allocated;
    *--pathp = '\0';

    struct stat st;
    if (lstat(".", &st) < 0)
    {
	#ifdef GETCWD_DEBUG
	error_message("%s: %d: return NULL }\n", __FILE__, __LINE__);
	#endif
	return 0;
    }
    dev_t thisdev = st.st_dev;
    ino_t thisino = st.st_ino;
    #ifdef GETCWD_DEBUG
    error_message("%s: %d: thisdev=%#x, thisino = %d\n", __FILE__, __LINE__,
	(int)thisdev, (int)thisino);
    #endif

    if (lstat("/", &st) < 0)
    {
	#ifdef GETCWD_DEBUG
	error_message("%s: %d: return NULL }\n", __FILE__, __LINE__);
	#endif
	return 0;
    }
    dev_t rootdev = st.st_dev;
    ino_t rootino = st.st_ino;
    #ifdef GETCWD_DEBUG
    error_message("%s: %d: rootdev=%#x, rootino = %d\n", __FILE__, __LINE__,
	(int)rootdev, (int)rootino);
    #endif

    //
    // Walk backwards up the parent directory links until we get to root.
    // As we go, we look for the name of the directory in the directory
    // above, and construct the directory name.
    //
    for (;;)
    {
	if  (thisdev == rootdev && thisino == rootino)
	    break;

	//
	// Build the path to the parent directory.
	//
	if (dotp == dotlist)
	{
	    // My, what a deep directory tree you have, Grandma.
	    char *new_dot_list = 0;
	    if (dotlist == dots)
	    {
		new_dot_list = (char *)malloc(dotsize * 2 + 1);
		if (!new_dot_list)
		{
		    #ifdef GETCWD_DEBUG
		    error_message("%s: %d: return NULL }\n", __FILE__,
			__LINE__);
		    #endif
		    return 0;
		}
		memcpy(new_dot_list, dots, dotsize);
		dotp = &new_dot_list[dotsize];
	    }
	    else
	    {
		new_dot_list = (char *)realloc((void *)dotlist, dotsize * 2 + 1);
		if (!new_dot_list)
		{
		    #ifdef GETCWD_DEBUG
		    error_message("%s: %d: return NULL }\n", __FILE__,
			__LINE__);
    #endif
		    return 0;
		}
		dotp = &new_dot_list[dotsize];
	    }
	    memcpy((char *)dotp, new_dot_list, dotsize);
	    dotsize *= 2;
	    new_dot_list[dotsize] = '\0';
	    dotlist = new_dot_list;
	}
	dotp -= 3;

	// Look at the parentdirectory.
	if (lstat(dotp, &st) < 0)
	{
	    #ifdef GETCWD_DEBUG
	    error_message("%s: %d: return NULL }\n", __FILE__, __LINE__);
	    #endif
	    return 0;
	}
	dev_t dotdev = st.st_dev;
	ino_t dotino = st.st_ino;
	#ifdef GETCWD_DEBUG
	error_message("%s: %d: dotdev=%#x, dotino = %d\n", __FILE__, __LINE__,
	    (int)dotdev, (int)dotino);
	#endif

	// Stop if we have reached a root point outside the root tree,
	// probably a path outside our present chrrot jail.
	if (dotdev == thisdev && dotino == thisino)
	    break;

	// Figure out if this directory is a mount point.
	bool mount_point = (dotdev != thisdev);

	// Search for the last directory.
	DIR *dirstream = opendir(dotp);
	if (!dirstream)
	{
	    #ifdef GETCWD_DEBUG
	    error_message("%s: %d: return NULL }\n", __FILE__, __LINE__);
	    #endif
	    return 0;
	}
	dirent *d = 0;
	for (;;)
	{
	    // Clear errno to distinguish EOF from error if readdir
	    // returns NULL.
	    errno = 0;
	    d = readdir(dirstream);
	    if (!d)
		break;

	    if
	    (
		d->d_name[0] == '.'
	    &&
		(
		    d->d_name[1] == '\0'
		||
		    (d->d_name[1] == '.' && d->d_name[2] == '\0')
		)
	    )
		continue;

	    #ifdef GETCWD_DEBUG
	    error_message("%s: %d: name=\"%s\", ino=%d\n", __FILE__, __LINE__,
		d->d_name, (int)d->d_ino);
	    #endif
	    if (mount_point || (ino_t)d->d_ino == thisino)
	    {
		char name[dotlist + dotsize - dotp + 1 + _D_ALLOC_NAMLEN(d)];
		memcpy(name, dotp, dotlist + dotsize - dotp);
		name[dotlist + dotsize - dotp] = '/';
		strcpy(&name[dotlist + dotsize - dotp + 1], d->d_name);

		// We don't fail here if we cannot stat() a directory entry.
		// This can happen when (network) filesystems fail.  If this
		// entry is in fact the one we are looking for we will find
		// out soon as we reach the end of the directory without
		// having found anything.
		if
		(
		    lstat(name, &st) >= 0
		&&
		    st.st_dev == thisdev
		&&
		    st.st_ino == thisino
		)
		    break;
	    }
	}
	if (!d)
	{
	    int save = errno;
	    closedir(dirstream);
	    if (save == 0)
	    {
		// EOF on dirstream, which means that the current directory
		// has been removed.
		save = ENOENT;
	    }
	    errno = save;
	    #ifdef GETCWD_DEBUG
	    error_message("%s: %d: return NULL (%d) }\n", __FILE__, __LINE__,
		save);
	    #endif
	    return 0;
	}

	size_t namlen = _D_EXACT_NAMLEN(d);
	if ((size_t)(pathp - path) <= namlen)
	{
	    if (size != 0)
	    {
		closedir(dirstream);
		errno = ERANGE;
	       	#ifdef GETCWD_DEBUG
		error_message("%s: %d: return ERANGE }\n", __FILE__, __LINE__);
		#endif
		return 0;
	    }
	    else
	    {
		size_t oldsize = allocated;
		allocated = 2 * MAX(allocated, namlen);
		char *tmp = (char *)realloc(path, allocated);
		if (!tmp)
		{
		    closedir(dirstream);
		    errno = ENOMEM; // closedir might have changed it.
		    #ifdef GETCWD_DEBUG
		    error_message("%s: %d: return ENOMEM }\n", __FILE__,
			__LINE__);
		    #endif
		    return 0;
		}

		// Move current contents up to the end of the buffer.
		// This is guaranteed to be non-overlapping.
		char *new_pathp = tmp + allocated - (path + oldsize - pathp);
		memcpy(new_pathp, tmp + (pathp - path), path + oldsize - pathp);
		pathp = new_pathp;
		path = tmp;
	    }
	}
	pathp -= namlen;
	memcpy(pathp, d->d_name, namlen);
	*--pathp = '/';
	closedir(dirstream);

	thisdev = dotdev;
	thisino = dotino;
    }

    //
    // Prepend a leading slash to make the path absolute.
    //
    if (pathp == &path[allocated - 1])
	*--pathp = '/';

    //
    // Move the path to the front of the buffer.
    // (This may overlap.)
    //
    memmove(path, pathp, path + allocated - pathp);

    // Restore errno on successful return.
    errno = prev_errno;

    #ifdef GETCWD_DEBUG
    error_message("%s: %d: return \"%s\" }\n", __FILE__, __LINE__, path);
    #endif
    return path;
}
