//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_opendir class
//

#include <plasticfs.h>

extern "C" {
#define _LIBC // so we get __libc_lock_t typed correctly
#define NOT_IN_libc  // so we don't have linux threads include problems
#include <bits/libc-lock.h>
}

#include <sys/stat.h>
#include <cerrno>
#include <fcntl.h>
#include <stdlib.h>

using namespace libplasticfs;


#define BUFSIZ 1024


struct __dirstream
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/dirstream.h
    // Copyright (C) 1993-1996 Free Software Foundation, Inc.
    int fd;                     // File descriptor.
    char *data;                 // Directory block.
    size_t allocation;          // Space allocated for the block.
    size_t size;                // Total valid data in the block.
    size_t offset;              // Current offset into the block.
    off_t filepos;              // Position of next entry to read.
    __libc_lock_define(, lock)  // Mutex lock for this structure.
};


class libc_lock_insulator
{
public:
    libc_lock_insulator(__libc_lock_t &arg):
	ref(&arg)
    {
	__libc_lock_lock(*ref);
    }

    ~libc_lock_insulator()
    {
	__libc_lock_unlock(*ref);
    }

private:
    __libc_lock_t *ref;
};


DIR *
plasticfs::opendir(const char *path)
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/opendir.c
    // Copyright (C) 1991-2002 Free Software Foundation, Inc.
    DIR             *dirp;
    struct stat64   statbuf;
    size_t          allocation;
    int             save_errno;

    if (path[0] == '\0')
    {
	// POSIX.1-1990 says an empty name gets ENOENT;
	// but `open' might like it fine.
	errno = ENOENT;
	return NULL;
    }

    int fd = open(path, O_RDONLY | O_NDELAY | O_DIRECTORY | O_LARGEFILE, 0);
    if (fd < 0)
	return NULL;

    // Now make sure this really is a directory and nothing changed
    // since the `stat' call.  We do not have to perform the test for
    // the descriptor being associated with a directory if we know the
    // O_DIRECTORY flag is honored by the kernel.  */
    if (fstat64(fd, &statbuf) < 0)
	goto lose;
    if (!S_ISDIR(statbuf.st_mode))
    {
	save_errno = ENOTDIR;
	goto lose;
    }

    if (fcntl(fd, F_SETFD, (void *)(long)FD_CLOEXEC) < 0)
	goto lose;

    if ((size_t)statbuf.st_blksize >= sizeof(dirent))
	allocation = statbuf.st_blksize;
    else
	allocation = (BUFSIZ < sizeof(dirent) ? sizeof(dirent) : BUFSIZ);

    dirp = (DIR *)malloc(sizeof(DIR) + allocation);
    if (dirp == NULL)
    {
        lose:
	save_errno = errno;
	close(fd);
	errno = save_errno;
	return NULL;
    }
    dirp->data = (char *)(dirp + 1);
    dirp->allocation = allocation;
    dirp->fd = fd;
    dirp->size = 0;
    dirp->offset = 0;
    dirp->filepos = 0;

    __libc_lock_init(dirp->lock);

    return dirp;
}


int
plasticfs::closedir(DIR *dirp)
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/closedir.c
    // Copyright (C) 1991-1998 Free Software Foundation, Inc.
    if (dirp == NULL)
    {
	errno = EINVAL;
	return -1;
    }
    __libc_lock_lock(dirp->lock);
    int fd = dirp->fd;
    __libc_lock_fini(dirp->lock);
    free((void *)dirp);
    return close(fd);
}


dirent *
plasticfs::readdir(DIR *dirp)
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/readdir.c
    // Copyright (C) 1991-2002 Free Software Foundation, Inc.
    int saved_errno = errno;
    libc_lock_insulator(dirp->lock);
    dirent *dp = 0;
    for (;;)
    {
	if (dirp->offset >= dirp->size)
	{
	    // We've emptied out our buffer.  Refill it.
	    unsigned maxread = dirp->allocation;
	    int bytes = getdents(dirp->fd, (dirent *)dirp->data, maxread);
	    if (bytes <= 0)
	    {
		// On some systems getdents fails with ENOENT when
		// the open directory has been rmdir'd already.
		// POSIX.1 requires that we treat this condition like
		// normal EOF.
		if (bytes < 0 && errno == ENOENT)
		    bytes = 0;

		// Don't modifiy errno when reaching EOF.
		if (bytes == 0)
		    errno = saved_errno;
		dp = NULL;
		break;
	    }
	    dirp->size = (size_t)bytes;

	    // Reset the offset into the buffer.
	    dirp->offset = 0;
	}

	dp = (dirent *)(dirp->data + dirp->offset);
	size_t reclen = dp->d_reclen;
	dirp->offset += reclen;
	dirp->filepos += reclen;

	// Skip deleted files.
	if (dp->d_ino)
	    break;
    }
    return dp;
}


dirent64 *
plasticfs::readdir64(DIR *dirp)
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/readdir.c
    // Copyright (C) 1991-2002 Free Software Foundation, Inc.
    int saved_errno = errno;
    libc_lock_insulator(dirp->lock);
    dirent64 *dp = 0;
    for (;;)
    {
	if (dirp->offset >= dirp->size)
	{
	    // We've emptied out our buffer.  Refill it.
	    unsigned maxread = dirp->allocation;
	    int bytes = getdents64(dirp->fd, (dirent64 *)dirp->data, maxread);
	    if (bytes <= 0)
	    {
		// On some systems getdents fails with ENOENT when
		// the open directory has been rmdir'd already.
		// POSIX.1 requires that we treat this condition like
		// normal EOF.
		if (bytes < 0 && errno == ENOENT)
		    bytes = 0;

		// Don't modifiy errno when reaching EOF.
		if (bytes == 0)
		    errno = saved_errno;
		dp = NULL;
		break;
	    }
	    dirp->size = (size_t)bytes;

	    // Reset the offset into the buffer.
	    dirp->offset = 0;
	}

	dp = (dirent64 *)(dirp->data + dirp->offset);
	size_t reclen = dp->d_reclen;
	dirp->offset += reclen;
	dirp->filepos += reclen;

	// Skip deleted files.
	if (dp->d_ino)
	    break;
    }
    return dp;
}


void
plasticfs::rewinddir(DIR *dirp)
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/rewinddir.c
    // Copyright (C) 1991-1998 Free Software Foundation, Inc.
    libc_lock_insulator guard(dirp->lock);
    lseek(dirp->fd, (off_t)0, SEEK_SET);
    dirp->offset = 0;
    dirp->size = 0;
}


void
plasticfs::seekdir(DIR *dirp, long pos)
{
    // derived from
    // GNU glibc 2.3 sysdeps/unix/seekdir.c
    // Copyright (C) 1991-2001 Free Software Foundation, Inc.
    libc_lock_insulator guard(dirp->lock);
    lseek(dirp->fd, pos, SEEK_SET);
    dirp->size = 0;
    dirp->offset = 0;
    dirp->filepos = pos;
}


long
plasticfs::telldir(DIR *dirp)
{
    return dirp->filepos;
}


int
plasticfs::dirfd(DIR *dirp)
{
    return dirp->fd;
}
