//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_parse class
//

#include <plasticfs.h>
#include <plasticfs/dlsym_rtld_next.h>
#include <plasticfs/filter/chroot.h>
#include <plasticfs/filter/log.h>
#include <plasticfs/filter/module.h>
#include <plasticfs/filter/nocase/dos.h>
#include <plasticfs/filter/nocase/shorten.h>
#include <plasticfs/filter/nocase/title.h>
#include <plasticfs/filter/nocase/toupper.h>
#include <plasticfs/filter/nocase/tolower.h>
#include <plasticfs/filter/smartlink.h>
#include <plasticfs/filter/viewpath.h>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <rcstring_list.h>

using namespace libplasticfs;


static const char *lex_buf;
static size_t lex_buf_pos;


static void
lex_open()
{
    lex_buf_pos = 0;
    lex_buf = getenv("PLASTICFS");
    if (!lex_buf)
	lex_buf = "";
}


static int
lex_getch()
{
    int c = (unsigned char)lex_buf[lex_buf_pos];
    if (c)
	++lex_buf_pos;
    return c;
}


static void
lex_ungetch(int c)
{
    if (lex_buf_pos > 0 && lex_buf[lex_buf_pos - 1] == c)
	--lex_buf_pos;

}


enum token_t
{
    token_end,
    token_string,
    token_pipe
};


static token_t token_current;
static char *token_value;
static size_t token_value_max;


static void
token_next()
{
    for (;;)
    {
	int c = lex_getch();
	switch (c)
	{
	case 0:
	    if (token_value)
	    {
		delete token_value;
		token_value = 0;
		token_value_max = 0;
	    }
	    token_current = token_end;
	    return;

	case ' ':
	case '\t':
	case '\r':
	case '\f':
	case '\n':
	    continue;

	case '^':
	case '|':
	    token_current = token_pipe;
	    return;

	default:
	    size_t pos = 0;
	    for (;;)
	    {
		if (pos >= token_value_max)
		{
		    size_t newmax = token_value_max * 2 + 16;
		    char *newbuf = new char[newmax + 1];
		    if (token_value_max)
		    {
			memcpy(newbuf, token_value, token_value_max);
			delete token_value;
		    }
		    token_value = newbuf;
		    token_value_max = newmax;
		}
		token_value[pos++] = c;
		c = lex_getch();
		switch (c)
		{
		default:
		    continue;

		case 0:
		    break;

		case ' ':
		case '\t':
		case '\r':
		case '\f':
		case '\n':
		case '^':
		case '|':
		    lex_ungetch(c);
		    break;
		}
		break;
	    }
	    token_value[pos] = '\0';
	    token_current = token_string;
	    return;
	}
    }
}


typedef plasticfs *(*factory_t)(plasticfs *, const rcstring_list &);


factory_t
find_filter(const char *name)
{
    struct table_t
    {
	const char *name;
	factory_t value;
    };

    static table_t table[] =
    {
	{ "chroot", plasticfs_filter_chroot::factory, },
	{ "dos", plasticfs_filter_nocase_dos::factory, },
	{ "downcase", plasticfs_filter_nocase_tolower::factory, },
	{ "log", plasticfs_filter_log::factory, },
	{ "nocase", plasticfs_filter_nocase::factory, },
	{ "shortname", plasticfs_filter_nocase_shorten::factory, },
	{ "smartlink", plasticfs_filter_smartlink::factory, },
	{ "titlecase", plasticfs_filter_nocase_title::factory, },
	{ "tolower", plasticfs_filter_nocase_tolower::factory, },
	{ "toupper", plasticfs_filter_nocase_toupper::factory, },
	{ "upcase", plasticfs_filter_nocase_toupper::factory, },
	{ "viewpath", plasticfs_filter_viewpath::factory, },
    };

    #define SIZEOF(a) (sizeof(a) / sizeof(a[0]))
    #define ENDOF(a) ((a) + SIZEOF(a))

    for (table_t *tp = table; tp < ENDOF(table); ++tp)
    {
	if (0 == strcmp(name, tp->name))
	    return tp->value;
    }
    return 0;
}


static plasticfs *
get_filter(plasticfs *deeper)
{
    //
    // This shouldn't happen, just return the deeper file
    // system without any filtering.
    //
    while (token_current == token_pipe)
	token_next();
    if (token_current != token_string)
	return deeper;

    //
    // Collect the filter arguments.
    //
    rcstring_list arguments;
    for (;;)
    {
	arguments.push_back(rcstring(token_value));
	token_next();

	if (token_current != token_string)
	    break;
    }

    //
    // The filters are applied in reverse order, deepest to shallowest.
    // Recursion is easier than malloc()ing stuff and processing it later.
    //
    deeper = get_filter(deeper);

    //
    // Now build this filter, and apply it to the deeper filter(s)
    // constructed thus far.
    //
    factory_t factory = find_filter(arguments[0].c_str());
    if (!factory)
    {
	plasticfs *deeper2 =
	    plasticfs_filter_module::factory(deeper, arguments);
	if (!deeper2)
	{
	    plasticfs::error_message
	    (
		"plasticfs: filter \"%s\" unknown\n",
		arguments[0].c_str()
	    );
	}
	else
	    deeper = deeper2;
    }
    else
    {
	deeper = factory(deeper, arguments);
    }

    //
    // Return the combined filter pipe.
    //
    return deeper;
}


plasticfs *
plasticfs::parse()
{
    //
    // At the end of the pipe is the class instance which passes our
    // system callbacks through to the next module in the preload list,
    // or libc if we are the last on the list.
    //
    plasticfs *result = new plasticfs_dlsym_rtld_next();
    plasticfs *old_master = master;
    master = result; // for error messages

    //
    // Security paranoia: if this program is doing something interesting,
    // show it the real file system, not the plastic one.
    //
    if (geteuid() == 0 || getuid() != geteuid() || getgid() != getegid())
	return master;

    //
    // Read the PLASTICFS environment variable, and parse it into a
    // filter list.
    //
    lex_open();
    token_next();
    result = get_filter(result);

    //
    // Return the combined filter pipe.
    //
    master = old_master;
    return result;
}
