//
//	plasticfs - the plastic file system
//	Copyright (C) 2002 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: functions to impliment the plasticfs_realpath class
//

#include <plasticfs.h>
#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <sys/stat.h>
#include <sys/param.h> // for MAXSYMLINKS

using namespace libplasticfs;


size_t
plasticfs::get_path_max()
{
#ifdef PATH_MAX
    return PATH_MAX;
#else
    long path_max = pathconf(name, _PC_PATH_MAX);
    if (path_max <= 0)
	path_max = 1024;
    return path_max;
#endif
}


int
plasticfs::maxsymlinks()
{
    return MAXSYMLINKS;
}


char *
plasticfs::realpath(const char *path, char *result)
{
    //
    // If they gave us a result buffer,
    // POSIX says it's MAX_PATH bytes long.
    //
    size_t path_max = get_path_max();
    if (result)
    {
	if (!realpath(path, result, path_max, true))
	    return 0;
	return result;
    }

    //
    // If they didn't give us a result buffer, POSIX says it's an error,
    // but a GNU libc extension malloc()s a buffer, so we do too.
    //
    // Because we are faking a C interface, we have to use malloc,
    // we can't use operator new, even though we are writing in C++.
    //
    char *result1 = (char *)malloc(path_max);
    if (!result1)
	return 0;
    if (!realpath(path, result1, path_max, true))
    {
	free(result1);
	return 0;
    }

    //
    // Now shorten the buffer to exactly
    // the number of bytes actually used.
    //
    size_t len = strlen(result1) + 1;
    char *result2 = (char *)realloc(result1, len);
    if (!result2)
    {
	free(result1);
	return 0;
    }
    result1 = 0;
    return result2;
}


//
// Return the canonical absolute name of file NAME.  A canonical name
// does not contain any `.', `..' components nor any repeated path
// separators ('/') or symlinks.  All path components must exist.  If
// RESOLVED is null, the result is malloc'd; otherwise, if the
// canonical name is nbytes chars or more, returns null with `errno'
// set to ENAMETOOLONG; if the name fits in fewer than PATH_MAX chars,
// returns the name in RESOLVED.  If the name cannot be resolved and
// RESOLVED is non-NULL, it contains the path of the first component
// that cannot be resolved.  If the path can be resolved, RESOLVED
// holds the same value as the value returned.  */
//
bool
plasticfs::realpath(const char *name, char *rpath, size_t nbytes,
    bool file_must_exist)
{
    // Derived from
    // GNU libc 2.3 stdlib/canonicalize.c
    // Copyright (C) 1996-2002 Free Software Foundation, Inc.
    if (!name || !rpath)
    {
	//
	// As per Single Unix Specification V2 we must return an error if
	// either parameter is a null pointer.  We extend this to allow
	// the RESOLVED parameter to be NULL in case the we are expected to
	// allocate the room for the return value.
	//
	errno = EINVAL;
	return false;
    }
    if (name[0] == '\0')
    {
	//
	// As per Single Unix Specification V2 we must return an error
	// if the name argument points to an empty string.
	//
	errno = ENOENT;
	return false;
    }
    if (nbytes < 2)
    {
	// Need room for at least '/' and '\0'
	errno = ENAMETOOLONG;
	return false;
    }

    char *rpath_limit = rpath + nbytes;

    // The free_string_guard_t class may be used to ensure that the
    // subject string is guaranteed to be free()ed when it goes out
    // of scope.  It's even exception safe.
    class free_string_guard_t
    {
    private:
	char *volatile &subject;
    public:
	free_string_guard_t(char *&arg): subject(arg) { }
	~free_string_guard_t() { if (subject) free(subject); }
    };

    char *name_buf = 0;
    free_string_guard_t name_buf_guard(name_buf);

    char *readlink_buf = 0;
    free_string_guard_t readlink_buf_guard(readlink_buf);

    char *dest = 0;
    if (name[0] != '/')
    {
	if (!getcwd(rpath, nbytes))
	    return false;
	dest = (char *)memchr(rpath, '\0', nbytes);
	if (!dest)
	{
	    errno = ENAMETOOLONG;
	    return false;
	}
    }
    else
    {
	rpath[0] = '/';
	dest = rpath + 1;
    }

    size_t path_max = get_path_max();

    //
    // The link counter is to count the number of symlinks we have
    //followed, in case we get caught in a symbolic link loop.
    //
    int num_links = maxsymlinks();

    //
    // The 'start' points at the start of the next path element, and
    // 'end' eventually points at the end of the next path element.
    // While we initially use 'name', we could redirect it into
    // 'name_buf' when we chase symbolic links, so avoid assumptions
    // about which we are pointing into.
    //
    const char *start;
    const char *end;
    for (start = end = name; *start; start = end)
    {
	// Skip all path separator characters.
	while (*start == '/')
	    ++start;

	// Find the end of the next path element.
	for (end = start; *end && *end != '/'; ++end)
	    ;
	if (end == start)
	    break;

	size_t element_length = end - start;
	if (element_length == 1 && start[0] == '.')
	{
	    // Ignore "." in the path.
	}
	else if (element_length == 2 && start[0] == '.' && start[1] == '.')
	{
	    // Back up to previous path element, ignore if at root already.
	    if (dest > rpath + 1)
	    {
		for (;;)
		{
		    --dest;
		    if (dest[-1] == '/')
			break;
		}
	    }
	}
	else
	{
	    //
	    // Make sure there is enough room in rpath for the next path
	    // element (including the '/' and the NUL).
	    //
	    if (dest + element_length + 2 > rpath_limit)
	    {
		errno = ENAMETOOLONG;
		return false;
	    }

	    //
	    // Copy the next element to the end of the result buffer.
	    //
	    if (dest[-1] != '/')
		*dest++ = '/';
	    memcpy(dest, start, element_length);
	    dest += element_length;
	    *dest = '\0';

	    //
	    // See what the file is.
	    //
	    struct stat64 st;
	    if (lstat64(rpath, &st) < 0)
	    {
		if (!file_must_exist && (errno == ENOTDIR || errno == ENOENT))
		{
		    // We've walked off the end of reality, so tack the
		    // end onto rpath and stop.  The end is either empty,
		    // or starts with a slash.
		    size_t len = strlen(end);
		    if (dest + len + 1 > rpath_limit)
		    {
			errno = ENAMETOOLONG;
			return false;
		    }
		    memcpy(dest, end, len);
		    dest += len;
		    break;
		}
		return false;
	    }

	    //
	    // If it's not a symbolic link,
	    // move on to the next path element.
	    //
	    if (!S_ISLNK(st.st_mode))
		continue;

	    //
	    // Make sure we aren't caught in a symbolic link loop.
	    //
	    if (--num_links <= 0)
	    {
		errno = ELOOP;
		return false;
	    }

	    //
	    // Allocate a buffer to read the symbolic link into.
	    //
	    if (!readlink_buf)
	    {
		readlink_buf = (char *)malloc(path_max);
		if (!readlink_buf)
		    return false;
	    }

	    //
	    // Read the symbolic link.
	    //
	    int n = readlink(rpath, readlink_buf, path_max - 1);
	    if (n < 0)
		return false;
	    if (n == 0)
	    {
		readlink_buf[0] = '.';
		n = 1;
	    }

	    //
	    // Look at what's left to scan.  It's either empty, or starts
	    // with a slash.  If it isn't empty, append it to the end
	    // of the symlink value.
	    //
	    // Make sure it will fit on the
	    // end of 'readlink_buf', including the NUL.
	    //
	    size_t len = strlen(end);
	    if (n + len + 1 > path_max)
	    {
		errno = ENAMETOOLONG;
		return false;
	    }
	    // Copy it to the end of 'readlink_buf' and pretend it was
	    // always that long.  This end part could be coming out
	    // of name_buf, so we can't simply copy it straight there.
	    if (len)
	    {
		memcpy(readlink_buf + n, end, len);
		n += len;
	    }

	    //
	    // Allocate a buffer to hold the diverted name
	    // constructed from the symbolic link.
	    //
	    if (!name_buf)
	    {
		name_buf = (char *)malloc(path_max);
		if (!name_buf)
		    return false;
	    }

	    //
	    // Copy the symlink into name_buf.  Don't worry about the
	    // leading path, we just continue from the point we were at.
	    //
	    memcpy(name_buf, readlink_buf, n);
	    name_buf[n] = 0;
	    end = name_buf;

	    //
	    // What we do with the result buffer depends on whether it's
	    // an absolute or relative symbolic link.
	    //
	    if (readlink_buf[0] == '/')
	    {
		// It's an absolute symlink, so restart the result.
		dest = rpath + 1;
	    }
	    else
	    {
		// It's a relative symlink, back up one path element.
		dest -= element_length;
	    }
	}
    }
    if (dest > rpath + 1 && dest[-1] == '/')
	--dest;
    *dest++ = '\0';
    return true;
}
