//
//	plasticfs - the plastic file system
//	Copyright (C) 2002, 2003 Peter Miller;
//	All rights reserved.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
//
// MANIFEST: interface definition for lib/rcstring.cc
//

#ifndef LIB_RCSTRING_H
#define LIB_RCSTRING_H

#pragma interface "libplasticfs::rcstring"

#include <stddef.h>
#include <stdarg.h>

namespace libplasticfs {

/**
  * The rcstring class is used to represent a reference counted string.
  */
class rcstring
{
public:
    /**
      * The destructor.
      */
    virtual ~rcstring()
    {
	value->one_less();
    }

    /**
      * The default constructor.
      */
    rcstring()
    {
	value = string_ty::from_c("");
    }

    /**
      * A constructor.
      */
    rcstring(const char *s)
    {
	value = string_ty::from_c(s);
    }

    /**
      * A constructor.
      */
    rcstring(const char *s, size_t len)
    {
	value = string_ty::n_from_c(s, len);
    }

    /**
      * The copy constructor.
      */
    rcstring(const rcstring &arg)
    {
	value = arg.value;
	value->one_more();
    }

    /**
      * The assignment operator.
      */
    rcstring &
    operator=(const rcstring &arg)
    {
	if (this != &arg)
	{
    	    value->one_less();
    	    value = arg.value;
    	    value->one_more();
	}
	return *this;
    }

    /**
      * The c_str method is used to obtain a pointer to a C string.
      */
    const char *
    c_str()
	const
    {
	return value->c_str();
    }

    typedef unsigned long hash_ty;

    hash_ty
    get_hash()
	const
    {
	return value->get_hash();
    }

private:
    /**
      * The string_ty inner class is used to hold the reference
      * counted value.  Its interface is very restricted.
      */
    class string_ty
    {
    public:
	/**
	  * A constructor.
	  */
	string_ty(const char *);

	/**
	  * A constrctor.
	  */
	string_ty(const char *, size_t);

	/**
	  * We have one more reference.
	  */
	void one_more();

	/**
	  * We hasve one fewer reference.
	  * Could result in this being deleted.
	  */
	void one_less();

	const char *
	c_str()
	    const
	{
	    return text;
	}

	size_t
	length()
	    const
	{
	    return str_length;
	}

	rcstring::hash_ty
	get_hash()
	    const
	{
	    return str_hash;
	}

    private:
	/**
	  * The destructor.
	  */
	~string_ty();

	string_ty(); // do not use
	string_ty(const string_ty &); // do not use
	string_ty &operator=(const string_ty &); // do not use

	rcstring::hash_ty str_hash;
	string_ty	*next;
	long		references;
	size_t		str_length;
	char		*text;

	static void initialize();
	static void split();

	static string_ty *from_c(const char *);
	static string_ty *n_from_c(const char *, size_t);
	void construct(const char *, size_t);

	static string_ty **hash_table;
	static rcstring::hash_ty hash_modulus;
	static rcstring::hash_ty hash_mask;
	static rcstring::hash_ty hash_load;

	friend class rcstring;
    };

    /**
      * The value of the rcstring.
      * (Should never be NULL.)
      */
    string_ty *value;

public:
    /**
      * Turn the string into upper case.
      */
    rcstring upcase() const;

    /**
      * Turn the string into lower case.
      */
    rcstring downcase() const;

    /**
      * The equality operator.
      */
    bool
    operator==(const rcstring &rhs)
	const
    {
	return (this->value == rhs.value);
    }

    /**
      * The inequality operator.
      */
    bool
    operator!=(const rcstring &rhs)
	const
    {
	return (this->value != rhs.value);
    }

    /**
      * The length method is used to determine the length of the string,
      * not including the terminating NUL.
      */
    size_t
    length()
	const
    {
	return value->length();
    }

    /**
      * Extract a substring.
      */
    rcstring substr(size_t start, size_t len) const;

    /**
      * The to_long method to used to conver the string into an integer.
      * Returns true on success (and sets n) or false on failure.
      */
    bool to_long(long &n, int base = 10) const;

    /**
      * The catenate method is used to join two strings end-to-end
      * to produce a new string.
      */
    static rcstring catenate(const rcstring &, const rcstring &);

    /**
      * The add and replace operator.
      */
    rcstring &
    operator+=(const rcstring &arg)
    {
	return operator=(catenate(*this, arg));
    }

    /**
      * The operator[] method is used to index to a specific character
      * within a string.  Note that you may not alter a string this way.
      */
    const char &
    operator[](size_t n)
	const
    {
	if (n >= length())
	{
	    static char dummy;
	    return dummy;
	}
	return c_str()[n];
    }
};

inline rcstring
operator+(const rcstring &lhs, const rcstring &rhs)
{
    return rcstring::catenate(lhs, rhs);
}

bool operator<(const rcstring &lhs, const rcstring &rhs);
bool operator<=(const rcstring &lhs, const rcstring &rhs);
bool operator>(const rcstring &lhs, const rcstring &rhs);
bool operator>=(const rcstring &lhs, const rcstring &rhs);

} // namespace libplasticfs

#endif // LIB_RCSTRING_H
